//
// Copyright (C) 2007 Institut fuer Telematik, Universitaet Karlsruhe (TH)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//

/**
 * @file DHT.cc
 * @author Ingmar Baumgart
 */

#include <IPAddressResolver.h>

#include "DHT.h"


Define_Module(DHT);

DHT::DHT()
{
    dataStorage = NULL;
}

DHT::~DHT()
{
    std::map<unsigned int, BaseCallMessage*>::iterator it = rpcIdMap.begin();
    while(it != rpcIdMap.end()) {
        cancelAndDelete(it->second);
      it++;
    }
    
    rpcIdMap.clear();
    getMap.clear();
    
    if (dataStorage != NULL) {
        dataStorage->clear();
    }
}

void DHT::initializeApp(int stage)
{
    if(stage != MIN_STAGE_APP)
        return;

    dataStorage = check_and_cast<DHTDataStorage*>
                  (parentModule()->submodule("dhtDataStorage"));

    numReplica = par("numReplica");
    numGetRequests = par("numGetRequests");
    ratioIdentical = par("ratioIdentical");

    if (numReplica > overlay->getMaxNumSiblings()) {
        opp_error("DHT::initialize(): numReplica bigger than what this overlay can handle (%d)", overlay->getMaxNumSiblings());
    }
    numStored = 0;
    maintenanceMessages = 0;
    normalMessages = 0;
    numBytesMaintenance = 0;
    numBytesNormal = 0;
    WATCH(numStored);
    WATCH(maintenanceMessages);
    WATCH(normalMessages);
    WATCH(numBytesNormal);
    WATCH(numBytesMaintenance);
    WATCH_MAP(rpcIdMap);
}

void DHT::handleTimerEvent(cMessage* msg)
{
    DHTTtlTimer* msg_timer;
    if (msg->isName("ttl_timer")) {
	msg_timer = check_and_cast<DHTTtlTimer*>(msg);

        EV << "(DHT) received timer ttl, key: "
	   << msg_timer->getKey().toString(16)
	   << "\n (thisNode.key = "
	   << thisNode.key.toString(16) << ")"
	   << endl;
	
	dataStorage->removeData(msg_timer->getKey());
	delete msg_timer; 
    }
}

bool DHT::handleRpc(BaseCallMessage* msg)
{
    // delegate messages
    RPC_SWITCH_START( msg )
    // RPC_DELEGATE( <messageName>[Call|Response], <methodToCall> )
    RPC_DELEGATE( DHTPut, handlePutRequest );
    RPC_DELEGATE( DHTGet, handleGetRequest );
    RPC_DELEGATE( DHTputCAPI, handlePutCAPIRequest ); //requests coming from an upper tier
    RPC_DELEGATE( DHTgetCAPI, handleGetCAPIRequest );
    RPC_SWITCH_END( )

    return RPC_HANDLED;
}

void DHT::handleRpcResponse(BaseResponseMessage* msg, int rpcId,
                              simtime_t rtt)
{
    RPC_SWITCH_START(msg)
    RPC_ON_RESPONSE( DHTPut ) {
        handlePutResponse(_DHTPutResponse);
        EV << "DHT Put RPC Response received: id=" << rpcId
        << " msg=" << *_DHTPutResponse << " rtt=" << rtt << endl;
        break;
    }
    RPC_ON_RESPONSE( DHTGet ) {
        handleGetResponse(_DHTGetResponse);
        EV << "DHT Get RPC Response received: id=" << rpcId
        << " msg=" << *_DHTGetResponse << " rtt=" << rtt << endl;
        break;
    }
    RPC_ON_RESPONSE( Lookup ) {
        replicate(_LookupResponse);
        EV << "Replica Set RPC Response received: id=" << rpcId
        << " msg=" << *_LookupResponse << " rtt=" << rtt << endl;
        break;
    }
    RPC_SWITCH_END( )
}

void DHT::handleRpcTimeout(BaseCallMessage* msg, const TransportAddress& dest, int rpcId, const OverlayKey& destKey)
{

    RPC_SWITCH_START( msg )
    RPC_ON_CALL( DHTPut ) {
        EV << "DHTPut Timeout" << endl;
        std::map<unsigned int, BaseCallMessage*>::iterator it =
	    rpcIdMap.find(msg->getNonce());
        if (it == rpcIdMap.end() || it->second == NULL)
          return;

        DHTputCAPIResponse* capiPutRespMsg = new DHTputCAPIResponse();
        capiPutRespMsg->setKey(_DHTPutCall->getKey());
        capiPutRespMsg->setIsSuccess(false);
        sendRpcResponse(it->second, capiPutRespMsg);
        rpcIdMap.erase(msg->getNonce());
        break;
    }
    RPC_ON_CALL( DHTGet ) {
        EV << "DHTGet Timeout" << endl;
        std::map<unsigned int, BaseCallMessage*>::iterator it =
	    rpcIdMap.find(_DHTGetCall->getNonce());
        std::map<OverlayKey, GetMapEntry>::iterator it2 =
	    getMap.find(_DHTGetCall->getKey());

        if (it2 == getMap.end()) { //unknown request
          return;
        }
          
        if (!(it == rpcIdMap.end() || it->second == NULL)) {
        //we have sent a 'real' get request
          rpcIdMap.erase(_DHTGetCall->getNonce());
          //ask anyone else, if possible
          if ((it2->second.hashVector != NULL)
	      && (it2->second.hashVector->size() > 0)) {

            DHTGetCall* getCall = new DHTGetCall();
            getCall->setKey(_DHTGetCall->getKey());
            getCall->setIsHash(false);
            getCall->setLength(GETCALL_L(getCall));
            RECORD_STATS(normalMessages++;
			             numBytesNormal += getCall->byteLength());
            int nonce = sendRouteRpcCall(TIER1_COMP,
                                         it2->second.hashVector->back(),
                                         getCall);
            rpcIdMap.insert(make_pair(nonce, it2->second.callMsg));
            it2->second.hashVector->pop_back();
          } else {
            //no one else
            DHTgetCAPIResponse* capiGetRespMsg = new DHTgetCAPIResponse();
            capiGetRespMsg->setKey(_DHTGetCall->getKey());
            capiGetRespMsg->setValue(BinaryValue::UNSPECIFIED_VALUE);
            capiGetRespMsg->setIsSuccess(false);
            sendRpcResponse(it2->second.callMsg,
			    capiGetRespMsg);
            getMap.erase(_DHTGetCall->getKey());
            return;
          }
        } else {
        //try to ask another one of the replica list for the hash
          if (it2->second.replica.size() > 0) {
            DHTGetCall* getCall = new DHTGetCall();
            getCall->setKey(_DHTGetCall->getKey());
            getCall->setIsHash(true);
            getCall->setLength(GETCALL_L(getCall));
            RECORD_STATS(normalMessages++;
                         numBytesNormal += getCall->byteLength());
            sendRouteRpcCall(TIER1_COMP, it2->second.replica.back(), getCall);
            it2->second.replica.pop_back();

          }
          else {
            //no one else to ask, see what we can do with what we have
            if (it2->second.numResponses > 0){
              unsigned int maxCount = 0;
              ReplicaVector* hashVector = NULL;
              std::map<BinaryValue, ReplicaVector>::iterator itHashes;
              for (itHashes =  it2->second.hashes.begin(); itHashes != it2->second.hashes.end(); itHashes++) {
                if (itHashes->second.size() > maxCount) {
                  maxCount = itHashes->second.size();
                  hashVector = &(itHashes->second);
                }
              }
              if ((double)maxCount/(double)it2->second.numResponses >= ratioIdentical) {
                it2->second.hashVector = hashVector;
              }
            }

            if ((it2->second.hashVector != NULL) && (it2->second.hashVector->size() > 0))
            {
              DHTGetCall* getCall = new DHTGetCall();
              getCall->setKey(_DHTGetCall->getKey());
              getCall->setIsHash(false);
              getCall->setLength(GETCALL_L(getCall));
              RECORD_STATS(normalMessages++;
                           numBytesNormal += getCall->byteLength());
              int nonce = sendRouteRpcCall(TIER1_COMP,
                                           it2->second.hashVector->back(), getCall);
              rpcIdMap.insert(make_pair(nonce, it2->second.callMsg));
              it2->second.hashVector->pop_back();
            }
            else {
              DHTgetCAPIResponse* capiGetRespMsg = new DHTgetCAPIResponse();
              capiGetRespMsg->setKey(_DHTGetCall->getKey());
              capiGetRespMsg->setIsSuccess(false);
              sendRpcResponse(it2->second.callMsg, capiGetRespMsg);
              getMap.erase(_DHTGetCall->getKey());
            }
          }
        }
        

        break;
    }
    RPC_SWITCH_END( )

}


void DHT::handleUpperMessage(cMessage* msg)
{
    error("DHT::handleUpperMessage(): Received message with unknown type!");

    delete msg;
}

void DHT::handlePutRequest(DHTPutCall* dhtMsg)
{
    std::string tempString = "PUT_REQUEST received: " +
        std::string(dhtMsg->getKey().toString(16));
    parentModule()->parentModule()->bubble(tempString.c_str());

    if (!(dataStorage->isModifiable(dhtMsg->getKey()))) {
        //check if the put request came from the right node
        NodeHandle sourceNode = dataStorage->getSourceNode(dhtMsg->getKey());
        if (((!sourceNode.isUnspecified()) && (!dhtMsg->getSrcNode().isUnspecified()) && (sourceNode != dhtMsg->getSrcNode()))
        || ((dhtMsg->getMaintenance()) && (dhtMsg->getOwnerNode() == sourceNode))) {
            // TODO: Set Owner!!!
            DHTPutResponse* responseMsg = new DHTPutResponse();
            responseMsg->setKey(dhtMsg->getKey());
            tempString = "Error, not allowed to modify this key";
            responseMsg->setValue(BinaryValue(tempString));
            responseMsg->setLength(PUTRESPONSE_L(responseMsg));
            RECORD_STATS(normalMessages++;
                         numBytesNormal += responseMsg->byteLength());
            sendRpcResponse(dhtMsg, responseMsg);
            return;
        }

    }

    // remove data item from local data storage
    cancelAndDelete(dataStorage->getTtlMessage(dhtMsg->getKey()));
    dataStorage->removeData(dhtMsg->getKey());
    if (dhtMsg->getValue().size() > 0) {
      //add ttl timer
      DHTTtlTimer *timerMsg = new DHTTtlTimer("ttl_timer");
      timerMsg->setKey(dhtMsg->getKey());
      scheduleAt(simTime()+dhtMsg->getTtl(), timerMsg);
      // storage data item in local data storage
      bool err;
      dataStorage->addData(dhtMsg->getKey(), dhtMsg->getValue(), timerMsg,
			   dhtMsg->getIsModifiable(), dhtMsg->getSrcNode(),
			   overlay->isSiblingFor(thisNode, dhtMsg->getKey(),
						 1, &err));
    }

    // send back
    DHTPutResponse* responseMsg = new DHTPutResponse();
    responseMsg->setKey(dhtMsg->getKey());

    responseMsg->setValue(dhtMsg->getValue());
    responseMsg->setLength(PUTRESPONSE_L(responseMsg));
    RECORD_STATS(normalMessages++; numBytesNormal += responseMsg->byteLength());

    sendRpcResponse(dhtMsg, responseMsg);
}

void DHT::handleGetRequest(DHTGetCall* dhtMsg)
{
    std::string tempString = "GET_REQUEST received: " +
	std::string(dhtMsg->getKey().toString(16));
    parentModule()->parentModule()->bubble(tempString.c_str());

    BinaryValue storedValue = dataStorage->getData(dhtMsg->getKey());

   
    // send back
    DHTGetResponse* responseMsg = new DHTGetResponse();
    responseMsg->setKey(dhtMsg->getKey());
    responseMsg->setIsHash(dhtMsg->getIsHash());
    if (storedValue == BinaryValue::UNSPECIFIED_VALUE) {
        responseMsg->setValue(BinaryValue::UNSPECIFIED_VALUE);
    }
    else {
        if (dhtMsg->getIsHash()) {
            CSHA1 sha1;
            char temp[41];
            temp[0] = '\0';
            sha1.Reset();
            sha1.Update((uint8_t*)(&(*storedValue.begin())),
			storedValue.size());
            sha1.Final();
            sha1.ReportHash(temp, CSHA1::REPORT_HEX);

            responseMsg->setValue(BinaryValue((char*)temp));
        }
        else {
            responseMsg->setValue(storedValue);
        }
    }
    responseMsg->setLength(GETRESPONSE_L(responseMsg));
    RECORD_STATS(normalMessages++; numBytesNormal += responseMsg->byteLength());
    sendRpcResponse(dhtMsg, responseMsg);
}

void DHT::handlePutCAPIRequest(DHTputCAPICall* capiPutMsg)
{
    //asks the replica list
    LookupCall* replicaMsg = new LookupCall();
    replicaMsg->setKey(capiPutMsg->getKey());
    replicaMsg->setNumSiblings(numReplica);
    int nonce = sendInternalRpcCall(OVERLAY_COMP, replicaMsg);
    rpcIdMap.insert(make_pair(nonce, capiPutMsg));
}

void DHT::handleGetCAPIRequest(DHTgetCAPICall* capiGetMsg)
{
    LookupCall* replicaMsg = new LookupCall();
    replicaMsg->setKey(capiGetMsg->getKey());
    replicaMsg->setNumSiblings(numReplica);
    int nonce = sendInternalRpcCall(OVERLAY_COMP, replicaMsg);
    rpcIdMap.insert(make_pair(nonce, capiGetMsg));
    lastGetCall = simTime();
}

void DHT::handlePutResponse(DHTPutResponse* dhtMsg)
{
    std::map<unsigned int, BaseCallMessage*>::iterator it =
	rpcIdMap.find(dhtMsg->getNonce());
    if (it == rpcIdMap.end() || it->second == NULL)
      return;
    DHTputCAPIResponse* capiPutRespMsg = new DHTputCAPIResponse();
    capiPutRespMsg->setKey(dhtMsg->getKey());
    capiPutRespMsg->setValue(dhtMsg->getValue());
    capiPutRespMsg->setIsSuccess(true);
    sendRpcResponse(it->second, capiPutRespMsg);
    rpcIdMap.erase(dhtMsg->getNonce());
}

void DHT::handleGetResponse(DHTGetResponse* dhtMsg)
{
    std::map<unsigned int, BaseCallMessage*>::iterator it =
	rpcIdMap.find(dhtMsg->getNonce());
    std::map<OverlayKey, GetMapEntry>::iterator it2 =
	getMap.find(dhtMsg->getKey());
    if (it2 == getMap.end()) //unknown request
      return;
      
    if ((it != rpcIdMap.end()) && (it->second != NULL)) {
    //we have sent a 'real' get request
        rpcIdMap.erase(dhtMsg->getNonce());
        if (!dhtMsg->getIsHash()) {
            DHTgetCAPIResponse* capiGetRespMsg = new DHTgetCAPIResponse();
            capiGetRespMsg->setKey(dhtMsg->getKey());
            capiGetRespMsg->setValue(dhtMsg->getValue());
            capiGetRespMsg->setIsSuccess(true);
            sendRpcResponse(it2->second.callMsg,
                            capiGetRespMsg);
            getMap.erase(dhtMsg->getKey());
            return;
        }
    }
    else {
        if (dhtMsg->getIsHash()) {
            std::map<BinaryValue, ReplicaVector>::iterator itHashes =  it2->second.hashes.find(dhtMsg->getValue());
            if (itHashes == it2->second.hashes.end()) //new hash
            {
                ReplicaVector vect = ReplicaVector();
                vect.push_back(dhtMsg->getSrcNode());
                it2->second.hashes.insert(make_pair(dhtMsg->getValue(), vect));
            }
            else {
                itHashes->second.push_back(dhtMsg->getSrcNode());
            }
            it2->second.numResponses++;
            if (it2->second.numResponses >= numGetRequests) {
                //we've got all the answers we wanted
                unsigned int maxCount = 0;
                ReplicaVector* hashVector = NULL;
                for (itHashes =  it2->second.hashes.begin(); itHashes != it2->second.hashes.end(); itHashes++) {
                    if (itHashes->second.size() > maxCount) {
                        maxCount = itHashes->second.size();
                        hashVector = &(itHashes->second);
                    }
                }
                if ((double)maxCount/(double)it2->second.numResponses >= ratioIdentical) {
                    it2->second.hashVector = hashVector;
                }
                else {
                    //we'll try to ask some other nodes
                    if (it2->second.replica.size() > 0) {
                        DHTGetCall* getCall = new DHTGetCall();
                        getCall->setKey(dhtMsg->getKey());
                        getCall->setIsHash(true);
                        getCall->setLength(GETCALL_L(getCall));
                        RECORD_STATS(normalMessages++;
                                     numBytesNormal += getCall->byteLength());
                        sendRouteRpcCall(TIER1_COMP,
                                         it2->second.replica.back(), getCall);
                        it2->second.replica.pop_back();
                    }
                    else { //we don't have anyone else to ask
                        DHTgetCAPIResponse* capiGetRespMsg = new DHTgetCAPIResponse();
                        capiGetRespMsg->setKey(dhtMsg->getKey());
                        capiGetRespMsg->setValue(BinaryValue::UNSPECIFIED_VALUE);
                        capiGetRespMsg->setIsSuccess(false);
                        sendRpcResponse(it2->second.callMsg,
                                        capiGetRespMsg);
                        getMap.erase(dhtMsg->getKey());
                    }


                }
            }
        }
    }
    
    if (it2->second.hashVector != NULL) {
    //we have already received all the response and chosen a hash
        if (it2->second.hashVector->size() > 0) {
            DHTGetCall* getCall = new DHTGetCall();
            getCall->setKey(dhtMsg->getKey());
            getCall->setIsHash(false);
            getCall->setLength(GETCALL_L(getCall));
            RECORD_STATS(normalMessages++;
			             numBytesNormal += getCall->byteLength());
            int nonce = sendRouteRpcCall(TIER1_COMP,
                                         it2->second.hashVector->back(),
                                         getCall);
            rpcIdMap.insert(make_pair(nonce, it2->second.callMsg));
            it2->second.hashVector->pop_back();
        }
        else { //we don't have anyone else to ask
            DHTgetCAPIResponse* capiGetRespMsg = new DHTgetCAPIResponse();
            capiGetRespMsg->setKey(dhtMsg->getKey());
            capiGetRespMsg->setValue(BinaryValue::UNSPECIFIED_VALUE);
            capiGetRespMsg->setIsSuccess(false);
            sendRpcResponse(it2->second.callMsg,
                            capiGetRespMsg);
            rpcIdMap.erase(dhtMsg->getNonce());
            getMap.erase(dhtMsg->getKey());
        }
    }
}

void DHT::update(const NodeHandle& node, bool joined)
{
    OverlayKey key;
    DHTPutCall* dhtMsg;
    bool err=false;
    DHTData entry;
    std::map<OverlayKey, DHTData>::iterator it = dataStorage->begin();
    for (unsigned int i=0; i< dataStorage->getSize() ; i++) {
        key = it->first;
        entry = it->second;
        if (joined) {
            if (entry.responsible && (overlay->isSiblingFor(node, key, numReplica, &err)
                    || err)) { // hack for Chord, if we've got a new predecessor

                dhtMsg = new DHTPutCall();
                dhtMsg->setKey(key);
                dhtMsg->setValue(entry.value);
                dhtMsg->setTtl((int)(entry.ttlMessage->arrivalTime()
                        - simulation.simTime()));
                dhtMsg->setIsModifiable(entry.is_modifiable);
                dhtMsg->setMaintenance(true);
                dhtMsg->setLength(PUTCALL_L(dhtMsg));
                RECORD_STATS(maintenanceMessages++;
                             numBytesMaintenance += dhtMsg->byteLength());
                sendRouteRpcCall(TIER1_COMP, node, dhtMsg);
            }
            if (err) {
                EV << "Unable to know if key: " << key
                << " is in range of node: " << node << endl;
            }
            //}
        } else {
#if 0
            //the update concerns a node who has left
            //replicate 
            LookupCall* replicaMsg = new LookupCall();
            replicaMsg->setKey(key);
            replicaMsg->setNumSiblings(numReplica);
            int nonce = sendInternalRpcCall(OVERLAY_COMP,
                                            replicaMsg);
            dhtMsg = new DHTPutCall();
            dhtMsg->setKey(key);
            dhtMsg->setValue(entry.value);
            dhtMsg->setTtl((int)(entry.ttlMessage->arrivalTime()
				 - simulation.simTime()));
            dhtMsg->setIsModifiable(entry.is_modifiable);
            dhtMsg->setMaintenance(true);
            dhtMsg->setLength(PUTCALL_L(dhtMsg));

            rpcIdMap.insert(make_pair(nonce, dhtMsg));
#endif
        }
        entry.responsible = overlay->isSiblingFor(thisNode, key, 1, &err);
        it++;
    }
}

void DHT::replicate(LookupResponse* lookupMsg)
{
    std::map<unsigned int, BaseCallMessage*>::iterator it =
        rpcIdMap.find(lookupMsg->getNonce());
    if (it == rpcIdMap.end() || it->second == NULL)
      return;

    if (dynamic_cast<DHTputCAPICall*>(it->second)) {
        DHTputCAPICall* capiPutMsg = dynamic_cast<DHTputCAPICall*>(it->second);
        rpcIdMap.erase(lookupMsg->getNonce());

        if ((lookupMsg->getIsValid() == false)
	    || (lookupMsg->getSiblingsArraySize() == 0)) {

            EV << "Unable to get replica list : invalid lookup" << endl;
            DHTputCAPIResponse* capiPutRespMsg = new DHTputCAPIResponse();
            capiPutRespMsg->setKey(lookupMsg->getKey());
            capiPutRespMsg->setIsSuccess(false);
            sendRpcResponse(capiPutMsg, capiPutRespMsg);
            return;
        }

        for( unsigned int i = 0; i < lookupMsg->getSiblingsArraySize(); i++ ) {
            DHTPutCall* dhtMsg = new DHTPutCall();
            dhtMsg->setKey(capiPutMsg->getKey());
            dhtMsg->setValue(capiPutMsg->getValue());
            dhtMsg->setTtl(capiPutMsg->getTtl());
            dhtMsg->setIsModifiable(capiPutMsg->getIsModifiable());
            dhtMsg->setMaintenance(false);
            dhtMsg->setLength(PUTCALL_L(dhtMsg));
            RECORD_STATS(normalMessages++;
                        numBytesNormal += dhtMsg->byteLength());
            int nonce = sendRouteRpcCall(TIER1_COMP,
                                         lookupMsg->getSiblings(i),
                                         dhtMsg);
            if (i == 0) {
                rpcIdMap.insert(make_pair(nonce, capiPutMsg));
            }
        }
    }
    else if (dynamic_cast<DHTgetCAPICall*>(it->second)) {
        DHTgetCAPICall* capiGetMsg = dynamic_cast<DHTgetCAPICall*>(it->second);
        rpcIdMap.erase(lookupMsg->getNonce());

        if ((lookupMsg->getIsValid() == false)
	    || (lookupMsg->getSiblingsArraySize() == 0)) {

            EV << "Unable to get replica list : invalid lookup" << endl;
            DHTgetCAPIResponse* capiGetRespMsg = new DHTgetCAPIResponse();
            capiGetRespMsg->setKey(lookupMsg->getKey());
            capiGetRespMsg->setValue(BinaryValue::UNSPECIFIED_VALUE);
            capiGetRespMsg->setIsSuccess(false);
            sendRpcResponse(capiGetMsg, capiGetRespMsg);
            return;
        }

        GetMapEntry mapEntry;

        for( unsigned int i = 0; i < lookupMsg->getSiblingsArraySize(); i++) {
            if (i < (unsigned int)numGetRequests) {
              DHTGetCall* dhtMsg = new DHTGetCall();
              dhtMsg->setKey(capiGetMsg->getKey());
              dhtMsg->setIsHash(true);
              dhtMsg->setLength(GETCALL_L(dhtMsg));
              RECORD_STATS(normalMessages++;
			               numBytesNormal += dhtMsg->byteLength());
              sendRouteRpcCall(TIER1_COMP, lookupMsg->getSiblings(i), dhtMsg);
            }
            else {
                //We don't send, we just store the remaining keys
                mapEntry.replica.push_back(lookupMsg->getSiblings(i));
            }
        }

        mapEntry.callMsg = capiGetMsg;
        mapEntry.numResponses = 0;
        mapEntry.hashVector = NULL;

        getMap.insert(make_pair(capiGetMsg->getKey(), mapEntry)); 
    }
    else if (dynamic_cast<DHTPutCall*>(it->second)) {
        DHTPutCall* putMsg = dynamic_cast<DHTPutCall*>(it->second);
        rpcIdMap.erase(lookupMsg->getNonce());

        if ((lookupMsg->getIsValid() == false)
	    || (lookupMsg->getSiblingsArraySize() == 0)) {

            EV << "Unable to get replica list : invalid lookup" << endl;
            delete putMsg;
            return;
        }

        for( unsigned int i = 0; i < lookupMsg->getSiblingsArraySize(); i++ ) {
            RECORD_STATS(maintenanceMessages++;
			 numBytesMaintenance += putMsg->byteLength());

            sendRouteRpcCall(TIER1_COMP, lookupMsg->getSiblings(i),
                             new DHTPutCall(*putMsg));
        }
        delete putMsg;
    }
}

void DHT::finishApp()
{
    double time = simTime() - creationTime;
    globalStatistics->addStdDev("DHT: Stored Messages", numStored);
    globalStatistics->addStdDev("DHT: Maintenance Messages",
				maintenanceMessages);
    globalStatistics->addStdDev("DHT: Normal Messages", normalMessages);
    globalStatistics->addStdDev("DHT: Bytes/s Maintenance Sent",
				numBytesMaintenance / time);
    globalStatistics->addStdDev("DHT: Bytes/s Normal Sent",
				numBytesNormal / time);
}
