//
// Copyright (C) 2006 Institut fuer Telematik, Universitaet Karlsruhe (TH)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//

/**
* @file SimpleGameClient.cc
* @author Helge Backhaus
*/

#include "SimpleGameClient.h"

Define_Module(SimpleGameClient);

void SimpleGameClient::initializeApp(int stage)
{
    // all initialization is done in the first stage
    if (stage != MIN_STAGE_APP)
        return;

    // fetch parameters
    areaDimension = par("areaDimension");
    movementRate = par("movementRate");
    movementSpeed = par("movementSpeed");
    movementDelay= 1.0 / movementRate;
    AOIWidth = par("AOIWidth");
    int groupSize = (int)par("groupSize");
    GeneratorType = (std::string)par("movementGenerator");

    WATCH_MAP(Neighbors);
    WATCH(position);
    WATCH(GeneratorType);
    WATCH(overlayReady);

    doRealworld = false;
    coordinator = NULL;
    scheduler = NULL;
    packetNotification = NULL;



    double speed = movementDelay * movementSpeed;
    if(strcmp(GeneratorType.c_str(), "greatGathering") == 0) {
        if(debugOutput) ev << "SIMPLECLIENT: Movement generator: greatGathering" << endl;
        coordinator = check_and_cast<GlobalCoordinator*>(simulation.moduleByPath("globalObserver.globalFunctions[0].coordinator"));
        Generator = new greatGathering(areaDimension, speed, &Neighbors, coordinator);
    }
    else if(strcmp(GeneratorType.c_str(), "groupRoaming") == 0) {
        if(debugOutput) ev << "SIMPLECLIENT: Movement generator: groupRoaming" << endl;
        coordinator = check_and_cast<GlobalCoordinator*>(simulation.moduleByPath("globalObserver.globalFunctions[0].coordinator"));
        Generator = new groupRoaming(areaDimension, speed, &Neighbors, coordinator, groupSize);
    }
    else if(strcmp(GeneratorType.c_str(), "realWorldRoaming") == 0) {
        if(debugOutput) ev << "SIMPLECLIENT: Movement generator: realWorldRoaming" << endl;
        Generator = new realWorldRoaming(areaDimension, movementSpeed, &Neighbors);
        mtu = par("mtu");
        packetNotification = new cMessage("packetNotification");
        scheduler = check_and_cast<RealtimeScheduler*>(simulation.scheduler());
        scheduler->setInterfaceModule(this, packetNotification, &packetBuffer, mtu, true);
        appFd = -1;
    }
    else {
        // debug output
        if(debugOutput) {
            if(strcmp(GeneratorType.c_str(), "randomRoaming") == 0) ev << "SIMPLECLIENT: Movement generator: randomRoaming" << endl;
            else ev << "SIMPLECLIENT: Movement generator: <unknown>. Defaulting to: randomRoaming" << endl;
        }
        Generator = new randomRoaming(areaDimension, speed, &Neighbors);
    }

    // self-messages
    move_timer = new cMessage("move_timer");
    overlayReady = false;
}

void SimpleGameClient::handleTimerEvent(cMessage* msg)
{
    if(msg->isName("move_timer")) {
        //reset timer
        cancelEvent(move_timer);
        if(overlayReady)
            scheduleAt(simulation.simTime() + movementDelay, msg);
        // handle event
        updatePosition();
    }
    if(msg->isName("packetNotification")) {
        while(packetBuffer.size() > 0) {
            // get packet from buffer and parse it
            RealtimeScheduler::PacketBufferEntry packet = *(packetBuffer.begin());
            packetBuffer.pop_front();
            switch (packet.func) {
                case RealtimeScheduler::PacketBufferEntry::DATA: {
                    if(packet.fd != appFd) {
                        error("SimpleClient::handleMessage(): Received packet from unknown socket!");
                    }
                    handleRealworldPacket(packet.data, packet.length);
                } break;
                case RealtimeScheduler::PacketBufferEntry::FD_NEW: {
                    if(appFd != -1) {
                        error("SimpleClient::handleMessage(): Multiple connections not supported yet!");
                    }
                    appFd = packet.fd;
                } break;
                case RealtimeScheduler::PacketBufferEntry::FD_CLOSE: {
                    if(packet.fd != appFd) {
                        error("SimpleClient::handleMessage(): Trying to close unknown connection!");
                    }
                    appFd = -1;
                }
            }
            delete packet.data;
        }
    }
}

void SimpleGameClient::handleLowerMessage(cMessage* msg)
{
    if(dynamic_cast<GameAPIMessage*>(msg)) {
        GameAPIMessage* gameAPIMsg = check_and_cast<GameAPIMessage*>(msg);
        switch(gameAPIMsg->getCommand()) {
            case MOVEMENT_REQUEST: {
                updatePosition();
            } break;
            case NEIGHBOR_UPDATE: {
                GameAPIListMessage* gameAPIListMsg = check_and_cast<GameAPIListMessage*>(msg);
                updateNeighbors(gameAPIListMsg);
            }
        }
    }
    delete msg;
}

void SimpleGameClient::handleReadyMessage(TierReadyMessage* msg)
{
    if(msg->getReady()) {
        overlayReady = true;
        scheduleAt(simulation.simTime() + movementDelay, move_timer);
    }
    else
        overlayReady = false;

    delete msg;
}

void SimpleGameClient::handleRealworldPacket(char *buf, uint32_t len)
{
    SCBasePacket *type = (SCBasePacket*)buf;
    switch(type->packetType) {
        case SC_PARAM_REQUEST: {
            SCParamPacket packet;
            packet.packetType = SC_PARAM_RESPONSE;
            packet.speed = movementSpeed;
            packet.dimension = areaDimension;
            packet.AOI = AOIWidth;
            packet.delay = movementDelay;
            packet.startX = position.x;
            packet.startY = position.y;
            int packetSize = sizeof(packet);
            scheduler->sendBytes((const char*)&packetSize, sizeof(int), 0, 0, true, appFd);
            scheduler->sendBytes((const char*)&packet, packetSize, 0, 0, true, appFd);
            doRealworld = true;
        } break;
        case SC_MOVE_INDICATION: {
            SCMovePacket *packet = (SCMovePacket*)type;
            Vector2D temp;
            temp.x = packet->posX;
            temp.y = packet->posY;
            dynamic_cast<realWorldRoaming*>(Generator)->setPosition(temp);
        }
    }
}

void SimpleGameClient::updateNeighbors(GameAPIListMessage* sgcMsg)
{
    unsigned int i;

    if(doRealworld) {
        int packetSize;

        SCRemovePacket removePacket;
        removePacket.packetType = SC_REMOVE_NEIGHBOR;
        packetSize = sizeof(removePacket);

        for(i=0; i<sgcMsg->getRemoveNeighborArraySize(); ++i) {
            removePacket.ip = sgcMsg->getRemoveNeighbor(i).ip.get4().getInt();
            scheduler->sendBytes((const char*)&packetSize, sizeof(int), 0, 0, true, appFd);
            scheduler->sendBytes((const char*)&removePacket, packetSize, 0, 0, true, appFd);
        }

        SCAddPacket addPacket;
        addPacket.packetType = SC_ADD_NEIGHBOR;
        packetSize = sizeof(addPacket);

        for(i=0; i<sgcMsg->getAddNeighborArraySize(); ++i) {
            addPacket.ip = sgcMsg->getAddNeighbor(i).ip.get4().getInt();
            addPacket.posX = sgcMsg->getNeighborPosition(i).x;
            addPacket.posY = sgcMsg->getNeighborPosition(i).y;
            scheduler->sendBytes((const char*)&packetSize, sizeof(int), 0, 0, true, appFd);
            scheduler->sendBytes((const char*)&addPacket, packetSize, 0, 0, true, appFd);
        }
    }
    else {
        for(i=0; i<sgcMsg->getRemoveNeighborArraySize(); ++i)
            Neighbors.erase(sgcMsg->getRemoveNeighbor(i));

        for(i=0; i<sgcMsg->getAddNeighborArraySize(); ++i) {
            Vector2D newPos;
            newPos = sgcMsg->getNeighborPosition(i);
            itNeighbors = Neighbors.find(sgcMsg->getAddNeighbor(i));
            if(itNeighbors != Neighbors.end()) {
                Vector2D temp = newPos - itNeighbors->second.position;
                temp.normalize();
                itNeighbors->second.direction = temp;
                itNeighbors->second.position = newPos;
            }
            else {
                NeighborMapEntry entry;
                entry.position = newPos;
                Neighbors.insert(std::make_pair(sgcMsg->getAddNeighbor(i), entry));
            }
        }
    }
}

void SimpleGameClient::updatePosition()
{
    Generator->move();
    position = Generator->getPosition();
    GameAPIPositionMessage *posMsg = new GameAPIPositionMessage("MOVEMENT_INDICATION");
    posMsg->setCommand(MOVEMENT_INDICATION);
    posMsg->setPosition(position);
    sendMessageToOverlay(posMsg);
}

SimpleGameClient::~SimpleGameClient()
{
    // destroy self timer messages
    cancelAndDelete(move_timer);
    cancelAndDelete(packetNotification);
}
