//
// Copyright (C) 2006 Institut fuer Telematik, Universitaet Karlsruhe (TH)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//

/**
 * @file BaseApp.cc
 * @author Bernhard Heep
 */

#include <IPAddressResolver.h>
#include <CommonMessages_m.h>
#include <CommonMessages_m.h>

#include "BaseApp.h"
#include "BaseRpc.h"
#include "BootstrapOracleAccess.h"
#include "GlobalStatisticsAccess.h"
#include "OverlayAccess.h"
#include "UnderlayConfiguratorAccess.h"
#include <UDPControlInfo_m.h>
#include <UDPAppBase.h>
#include <UDPSocket.h>

BaseApp::BaseApp()
{
    notificationBoard = NULL;

    overlay = NULL;
}

BaseApp::~BaseApp()
{
    finishRpcs();
}

int BaseApp::numInitStages() const
{
    return MAX_STAGE_APP + 1;
}

void BaseApp::initialize(int stage)
{
    if(stage == MIN_STAGE_APP) {
        // fetch parameters
        debugOutput = par("debugOutput");

        bootstrapOracle = BootstrapOracleAccess().get();
        underlayConfigurator = UnderlayConfiguratorAccess().get();
        globalStatistics = GlobalStatisticsAccess().get();
        notificationBoard = NotificationBoardAccess().get();

        // subscribe to the notification board
        notificationBoard->subscribe(this, NF_OVERLAY_TRANSPORTADDRESS_CHANGED);
        notificationBoard->subscribe(this, NF_OVERLAY_NODE_LEAVE);
        notificationBoard->subscribe(this, NF_OVERLAY_NODE_GRACEFUL_LEAVE);

        // set overlay pointer
        overlay = OverlayAccess().get(this);

        // determine the terminal's transport address
        thisNode.ip = IPAddressResolver().addressOf(
                                  parentModule()->parentModule()).get4();
        WATCH(thisNode);

        // statistics
        numSent = 0;
        numReceived = 0;
        bytesSent = 0;
        bytesReceived = 0;
        creationTime = simTime();

        WATCH(numSent);
        WATCH(numReceived);
        WATCH(bytesSent);
        WATCH(bytesReceived);

        // register component
        thisCompType = getThisCompType();
        overlay->registerComp(thisCompType, this);
        
        // init rpcs
        initRpcs();
    }

    if(stage >= MIN_STAGE_APP && stage <= MAX_STAGE_APP)
        initializeApp(stage);
}

void BaseApp::initializeApp(int stage)
{
    // ...
}

// Process messages passed up from the overlay.
void BaseApp::handleMessage(cMessage* msg)
{
    // Process self-messages.
    if(msg->isSelfMessage()) {
        // process rpc self-messages
        BaseRpcMessage* rpcMessage = dynamic_cast<BaseRpcMessage*>(msg);
        if (rpcMessage!=NULL) {
            internalHandleRpcMessage(rpcMessage);
            return;
        }
        // process all other self-messages
        handleTimerEvent(msg);
        return;
    }

    // TODO: cleanup code
    if (msg->arrivedOn("from_lowerTier") || msg->arrivedOn("direct_in")) {
        BaseRpcMessage* rpcMessage = dynamic_cast<BaseRpcMessage*>(msg);
        if (rpcMessage!=NULL) {
            internalHandleRpcMessage(rpcMessage);
            return;
        }
        TierReadyMessage* readyMsg = dynamic_cast<TierReadyMessage*>(msg);
        if (readyMsg != NULL) {
            handleReadyMessage(readyMsg);
            return;
        }
        // common API
        CommonAPIMessage* commonAPIMsg = dynamic_cast<CommonAPIMessage*>(msg);
        if(commonAPIMsg != NULL)
            handleCommonAPIMessage(commonAPIMsg);

        else if (msg->arrivedOn("from_lowerTier")) {
            RECORD_STATS(numReceived++; bytesReceived += msg->byteLength());
            handleLowerMessage(msg);
        }
        else
            delete msg;
    } else if(msg->arrivedOn("from_upperTier")) {
        BaseRpcMessage* rpcMessage = dynamic_cast<BaseRpcMessage*>(msg);
        if (rpcMessage!=NULL) {
            internalHandleRpcMessage(rpcMessage);
            return;
        }
        handleUpperMessage(msg);
    } else if (msg->arrivedOn("from_udp")) {
        handleUDPMessage(msg);
    } else if(msg->arrivedOn("trace_in")) {
        handleTraceMessage(msg);
    } else {
        delete msg;
    }

}

CompType BaseApp::getThisCompType()
{
    std::string name(this->name());

    if (name == std::string("tier1")) {
        return TIER1_COMP;
    } else if (name == std::string("tier2")) {
        return TIER2_COMP;
    } else if (name == std::string("tier3")) {
        return TIER3_COMP;
    }

    std::string parentName(this->parentModule()->name());

    if (parentName == std::string("tier1")) {
        return TIER1_COMP;
    } else if (parentName == std::string("tier2")) {
        return TIER2_COMP;
    } else if (parentName == std::string("tier3")) {
        return TIER3_COMP;
    } else {
        throw new cException("BaseApp::getThisCompType(): "
                             "Unknown module type!");
    }

    return INVALID_COMP;

}


void BaseApp::handleTimerEvent(cMessage* msg)
{
    // process self-messages
}

void BaseApp::receiveChangeNotification(int category, cPolymorphic* details)
{
    Enter_Method_Silent();
    if (category == NF_OVERLAY_TRANSPORTADDRESS_CHANGED) {
        handleTransportAddressChangedNotification();
    } else if (category == NF_OVERLAY_NODE_LEAVE) {
        handleNodeLeaveNotification();
    } else if (category == NF_OVERLAY_NODE_GRACEFUL_LEAVE) {
        handleNodeGracefulLeaveNotification();
    }
}

void BaseApp::handleTransportAddressChangedNotification()
{
    // ...
}

void BaseApp::handleNodeLeaveNotification()
{
    // ...
}

void BaseApp::handleNodeGracefulLeaveNotification()
{
    // ...
}

void BaseApp::callRoute(const OverlayKey& key, cMessage* msg,
                        const TransportAddress& hint)
{
    // create route-message (common API)
    KBRroute* routeMsg = new KBRroute();
    routeMsg->setDestKey(key);
    routeMsg->setHint(hint);
    routeMsg->encapsulate(msg);
    routeMsg->setSrcComp(thisCompType);
    routeMsg->setDestComp(thisCompType);

    routeMsg->setType(KBR_ROUTE);

    sendDirect(routeMsg, 0, overlay->getCompRpcGate(OVERLAY_COMP));

    // debug output
    if (debugOutput)
        EV << "[BaseApp::callRoute() @ " << thisNode.ip
           << " (" << thisNode.key.toString(16) << ")]\n"
           << "    Sent message " << id() << "-" << numSent << "\n"
           << "    to destination key " << key.toString(16)
           << endl;

    // count
    RECORD_STATS(numSent++; bytesSent += msg->byteLength());
}

void BaseApp::deliver(OverlayKey& key, cMessage* msg)
{
    // deliver...

    delete msg;
}

void BaseApp::forward(OverlayKey* key, cMessage** msg, NodeHandle* nextHopNode)
{
    // usually do nothing
}

void BaseApp::forwardResponse(const OverlayKey& key, cMessage* msg,
                              const NodeHandle& nextHopNode)
{
    OverlayCtrlInfo* ctrlInfo =
        check_and_cast<OverlayCtrlInfo*>(msg->removeControlInfo());

        //create forwardResponse message (common API)
        KBRforward* forwardMsg = new KBRforward();
        forwardMsg->setDestKey(key);
        forwardMsg->setNextHopNode(nextHopNode);
        forwardMsg->setControlInfo(ctrlInfo);
        forwardMsg->encapsulate(msg);

        forwardMsg->setType(KBR_FORWARD_RESPONSE);

        send(forwardMsg, "to_lowerTier");
}

void BaseApp::update(const NodeHandle& node, bool joined)
{
    if(joined) {
        EV << "[BaseApp::update() @ " << thisNode.ip
           << " (" << thisNode.key.toString(16) << ")]\n"
           << "    (" << node << ", " << joined << ") joined"
           << endl;
    } else {
        EV << "[BaseApp::update() @ " << thisNode.ip
           << " (" << thisNode.key.toString(16) << ")]\n"
           << "    (" << node << ", " << joined << ") left"
           << endl;
    }

}

void BaseApp::handleCommonAPIMessage(CommonAPIMessage* commonAPIMsg)
{
    if (thisCompType != TIER1_COMP)
        opp_error("KBRroute/KBRdeliver/KBRforward/KBRupdate are "
                  "only handled on tier1!");

    cMessage* tempMsg = commonAPIMsg->decapsulate();

    // process interface control information
    OverlayCtrlInfo* overlayCtrlInfo =
        dynamic_cast<OverlayCtrlInfo*>(commonAPIMsg->removeControlInfo());
        if(overlayCtrlInfo != NULL) {
            tempMsg->setControlInfo(overlayCtrlInfo);

            if(debugOutput)
                EV << "[BaseApp::handleCommonAPIMessage() @ " << thisNode.ip
                   << " (" << thisNode.key.toString(16) << ")]\n"
                   << "    Received message from node " << overlayCtrlInfo->getSrcNode().ip
                   << endl;
        }

        switch (commonAPIMsg->getType()) {

        case KBR_DELIVER:
        {
            KBRdeliver* apiMsg = dynamic_cast<KBRdeliver*>(commonAPIMsg);
            OverlayKey key = apiMsg->getDestKey();
            NodeHandle nextHopNode = thisNode;
            //first call forward, then deliver
            forward(&key, &tempMsg, &nextHopNode);

            if(tempMsg != NULL) {
                //if key or nextHopNode is changed send msg back to overlay
                if ((!key.isUnspecified() && key != apiMsg->getDestKey()) ||
                        (!nextHopNode.isUnspecified() &&
                                nextHopNode != thisNode)) {
                    forwardResponse(key, tempMsg, nextHopNode);
                }
                else {
                    RECORD_STATS(numReceived++; bytesReceived += tempMsg->byteLength());
                    //handle RPC first
                    BaseRpcMessage* rpcMessage = dynamic_cast<BaseRpcMessage*>(tempMsg);
                    if (rpcMessage!=NULL) {
                        internalHandleRpcMessage(rpcMessage);
                    }
                    else {
                        deliver(apiMsg->getDestKey(), tempMsg);
                    }
                }
            }
            break;
        }

        case KBR_FORWARD:
        {
            KBRforward* apiMsg = dynamic_cast<KBRforward*>(commonAPIMsg);
            OverlayKey key = apiMsg->getDestKey();
            NodeHandle nextHopNode = apiMsg->getNextHopNode();

            forward(&key, &tempMsg, &nextHopNode);

            //if message ist not deleted send it back
            if(tempMsg != NULL) {
                if(nextHopNode == apiMsg->getNextHopNode())
                    //do we need this?
                    nextHopNode = NodeHandle::UNSPECIFIED_NODE;
                forwardResponse(key, tempMsg, nextHopNode);
            }
            break;
        }

        case KBR_UPDATE:
        {
            KBRupdate* apiMsg = dynamic_cast<KBRupdate*>(commonAPIMsg);
            if (apiMsg->getJoined() &&
                ((apiMsg->getNode().ip == thisNode.ip &&
                 (thisNode.port == -1 ||
                  apiMsg->getNode().port == thisNode.port)) ||
                 (!thisNode.key.isUnspecified() &&
                  apiMsg->getNode().key == thisNode.key)))
                thisNode = apiMsg->getNode();
            else
                update(apiMsg->getNode(), apiMsg->getJoined());

            break;
        }

        default:
        {
            delete tempMsg;
        }
        }
        delete commonAPIMsg;
}

void BaseApp::handleUpperMessage(cMessage* msg)
{
    delete msg;
}

void BaseApp::handleLowerMessage(cMessage* msg)
{
    delete msg;
}

void BaseApp::handleUDPMessage(cMessage *msg)
{
    delete msg;
}

void BaseApp::handleReadyMessage(TierReadyMessage* msg)
{
    delete msg;
}

void BaseApp::handleTraceMessage(cMessage* msg)
{
    throw new cException("This application cannot handle trace data. "
                         "You have to overwrite handleTraceMessage() in your application "
                         "to make trace files work");
}

void BaseApp::sendMessageToOverlay(cMessage* msg)
{
    RECORD_STATS(numSent++; bytesSent += msg->byteLength());

    send(msg, "to_lowerTier");
}

void BaseApp::finish()
{
    // record scalar data
    simtime_t time = globalStatistics->calcMeasuredLifetime(creationTime);

    if(time != 0) {
        globalStatistics->addStdDev("BaseApp: Sent Messages/s", numSent / time);
        globalStatistics->addStdDev("BaseApp: Received Messages/s", numReceived / time);
        globalStatistics->addStdDev("BaseApp: Sent Bytes/s", bytesSent / time);
        globalStatistics->addStdDev("BaseApp: Received Bytes/s", bytesReceived / time);
    }

    finishApp();
}

void BaseApp::finishApp()
{
    // ...
}


void BaseApp::sendRpcMessageWithTransport(TransportType transportType,
                                          CompType destComp,
                                          const TransportAddress& dest,
                                          const OverlayKey& destKey,
                                          BaseRpcMessage* message)
{
    switch (transportType) {
    case UDP_TRANSPORT:
        sendMessageToUDP( dest, message );
        break;
    case ROUTE_TRANSPORT:
        callRoute(destKey, message, dest);
        break;
    case INTERNAL_TRANSPORT: {
        cGate *destCompGate = overlay->getCompRpcGate(destComp);
         if (destCompGate == NULL) {
             throw new cException("BaseApp::sendRpcMessageWithTransport():"
                                  " INTERNAL_RPC to unknown RpcCompType!");
         }
         OverlayCtrlInfo *overlayCtrlInfo = new OverlayCtrlInfo();
         overlayCtrlInfo->setSrcComp(thisCompType);
         overlayCtrlInfo->setDestComp(destComp);
         overlayCtrlInfo->setTransportType(INTERNAL_TRANSPORT);
         message->setControlInfo(overlayCtrlInfo);
         sendDirect( message, 0, destCompGate);
         break;
    }
    default:
        opp_error("BaseApp::sendRpcMessageWithTransport: "
                  "invalid transportType!");
        break;
    }
}

/* copied from BaseOverlay */

void BaseApp::bindToPort(int port)
{
    EV << "[BaseApp::bindToPort() @ " << thisNode.ip << ":  Binding to UDP port " << port
       << endl;

    cMessage *msg = new cMessage("UDP_C_BIND", UDP_C_BIND);
    UDPControlInfo *ctrl = new UDPControlInfo();
    ctrl->setSrcPort(port);
    ctrl->setSockId(UDPSocket::generateSocketId());
    msg->setControlInfo(ctrl);
    send(msg, "to_udp");
}

void BaseApp::sendMessageToUDP(const TransportAddress& destAddr, cMessage *msg)
{
    // send message to UDP, with the appropriate control info attached
    msg->removeControlInfo();
    msg->setKind(UDP_C_DATA);

    UDPControlInfo *ctrl = new UDPControlInfo();
    ctrl->setSrcPort(thisNode.port);
    ctrl->setSrcAddr(thisNode.ip);
    ctrl->setDestAddr(destAddr.ip);
    ctrl->setDestPort(destAddr.port);
    msg->setControlInfo(ctrl);

    EV << "[BaseApp::sendMessageToUDP() @ " << thisNode.ip
       << ":    Sending packet: " << msg
       << endl;

    send(msg, "to_udp");
}

//private
bool BaseApp::internalHandleRpc( BaseCallMessage* msg )
{
    // if RPC was handled return true, else tell the parent class to handle it
    return BaseRpc::internalHandleRpc(msg);
}

bool BaseApp::internalHandleRpcResponse(BaseResponseMessage* msg, int rpcId, simtime_t rtt)
{
    // if RPC was handled return true, else tell the parent class to handle it
    return BaseRpc::internalHandleRpcResponse(msg, rpcId, rtt);
}

