//
// Copyright (C) 2006 Institut fuer Telematik, Universitaet Karlsruhe (TH)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//

/**
 * @file BaseLookup.h
 * @author Sebastian Mies
 */

#ifndef __BASE_LOOKUP_H
#define __BASE_LOOKUP_H

// predefinitions
class LookupListener;
class AbstractLookup;
class BaseLookup;
class BasePathLookup;

/**
 * This class holds the lookup configuration.
 *
 * @author Sebastian Mies
 */ 
class BaseLookupConfiguration
{
public:
    int redundantNodes;   /**< number of next hops in each step */
    int parallelPaths; /**< number of parallel paths */
    int parallelRpcs;  /**< number of nodes to ask in parallel */
    bool secure;       /**< true, if all nodes should be identified with a ping */
    bool merge;        /**< true, if parallel Rpc results should be merged */
    bool failedNodeRpcs; //!< communicate failed nodes
};

// includes
#include <BaseOverlay.h>
#include <RpcListener.h>

#include <OverlayKey.h>
#include <NodeHandle.h>
#include <NodeVector.h>
#include <Comparator.h>

#include <ext/hash_map>
#include <ext/hash_set>
#include <vector>

// used namespaces
using namespace std;
using namespace __gnu_cxx;


/**
 * This class declares an abstract lookup listener.
 * 
 * @author Sebastian Mies
 */
class LookupListener
{
public:
    /**
     * virtual destructor
     */
    virtual ~LookupListener();
    
    /**
     * 
     */
    virtual void lookupFinished( AbstractLookup* lookup ) = 0;
};

/**
 * This class declares an abstract iterative lookup.
 *
 * @author Sebastian Mies
 */
class AbstractLookup
{
public:
    /**
     * Virtual destructor
     */
    virtual ~AbstractLookup();

    /**
     * Lookup siblings for a key
     * 
     * @param key The key to lookup
     * @param numSiblings Number of siblings to lookup
     * @param hopCountMax Maximum hop count
     * @param retries Number of retries if lookup fails
     * @param listener Listener to inform, when the lookup is done
     */
    virtual void lookup( const OverlayKey& key, int numSiblings = 1,
			 int hopCountMax = 0, int retries = 0, 
			 LookupListener* listener = NULL ) = 0;

    /**
     * Returns the result of the lookup
     * 
     * @return The result node vector.
     */
    virtual const NodeVector& getResult() const = 0;

    /**
     * Returns true, if the lookup was successful.
     * 
     * @return true, if the lookup was successful.
     */
    virtual bool isValid() const = 0;

    /**
     * Aborts a running lookup.
     *
     * This method aborts a running lookup without calling the
     * listener and delete the lookup object.
     */
    virtual void abortLookup() = 0;

    /**
     * Returns the total number of hops for all lookup paths.
     * 
     * @return The accumulated number of hops.
     */
    virtual uint getAccumulatedHops() const = 0;

};




/**
 * This class implements a basic greedy lookup strategy.
 * 
 * It uses the standard metric for greedy behaviour. If another
 * metric is needed, the distance function can be replaced by
 * overriding the distance method.
 *
 * @author Sebastian Mies
 */
class BaseLookup : public RpcListener,
		   public AbstractLookup,
		   public Comparator<OverlayKey>
{
    friend class BasePathLookup;
    friend class BaseOverlay;

    //-------------------------------------------------------------------------
    //- Enhanceable methods ---------------------------------------------------
    //-------------------------------------------------------------------------
protected:
    /**
     * This method creates a new path lookup. It may be overloaded
     * to enhance BasePathLookup with some new information/features.
     * 
     * @return The new path lookup
     */
    virtual BasePathLookup* createPathLookup();

    /**
     * This method creates the initial FindNode RPC call message used to
     * query the local findNode() method. It may be overridden to allow
     * additional information being attached before the lookup procedure
     * actually starts
     *
     * @return newly allocated FindNodeCall message
     */
    virtual FindNodeCall* createInitialRpcCall();

    //-------------------------------------------------------------------------
    //- Base configuration and state ------------------------------------------
    //-------------------------------------------------------------------------
protected:
    OverlayKey key;                 /**< key to lookup */
    BaseOverlay* overlay;           /**< ptr to overlay */
    BaseLookupConfiguration config; /**< lookup configuration */
    cObject* firstCallExt;	        /**< additional info for first findNode() */
    LookupListener* listener;       /**< lookup listener */
    vector<BasePathLookup*> paths;  /**< parallel paths */
    bool finished;                  /**< true, if lookup is finished */
    bool success;                   /**< true, if lookup was successful */
    bool running;                   /**< true, if lookup is running */
    int retries;                    /**< number of retries, if lookup fails */
    uint finishedPaths;             /**< number of finished paths */
    uint successfulPaths;           /**< number of successful paths */
    uint accumulatedHops;           /**< total number of hops (for all paths) */

protected://virtual methods: comparator induced by distance in BaseOverlay
    /**
     * compares two OverlayKeys and indicates which one is 
     * closer to the key to look up
     *
     * @param lhs the first OverlayKey
     * @param rhs the second OverlayKey
     * @return -1 if rhs is closer, 0 if lhs and rhs are
     *         equal and 1 if rhs is farther away to the key to lookup
     */
    int compare( const OverlayKey& lhs, const OverlayKey& rhs ) const;

    //-------------------------------------------------------------------------
    //- Siblings and visited nodes management---------------------------------
    //-------------------------------------------------------------------------
protected://fields
    NodeVector siblings;           /**< closest nodes */
    TransportAddress::Set visited;  /**< nodes already visited */
    int numSiblings;               /**< number of siblings */
    int hopCountMax;                /**< maximum hop count */ 

protected://methods
    /**
     * adds a node to the siblings NodeVector
     *
     * @param handle NodeHandle of the node to add
     * @return true if operation was succesfull, false otherwise
     */
    bool addSibling( const NodeHandle& handle );
    
    /**
     * adds/deletes visited nodes to/from the visited TransportAddress::Set
     *
     * @param addr TransportAddress of the node to add
     * @param visited if true node is addes, else node is erased
     */
    void setVisited( const TransportAddress& addr, bool visited = true );
    
    /**
     * indicates if the specified node has been visited before
     *
     * @param addr TransportAddress of the node
     * @return false if addr is not in visited, true otherwise
     */
    bool getVisited( const TransportAddress& addr );

    //-------------------------------------------------------------------------
    //- Parallel RPC distribution ---------------------------------------------
    //-------------------------------------------------------------------------
protected://fields and classes: rpc distribution
    
    class RpcInfo
    {
    public:
        int vrpcId;
        BasePathLookup* path;
    };

    class RpcInfoVector : public vector<RpcInfo>
    {
    public:
        uint32_t nonce;
/*
	vector<RpcInfo> v;

	RpcInfoVector() {};

	RpcInfoVector(const RpcInfoVector& r) {
	    nonce = r.nonce;
	    v = r.v;
	    std::cout << "!!!!!!!!" << nonce << std::endl;
	}
*/
    };
    
    typedef hash_map<TransportAddress, RpcInfoVector, TransportAddress::hashFcn>
	RpcInfoMap;
    RpcInfoMap rpcs;

protected://methods: rpcListener
    void handleRpcResponse(BaseResponseMessage* msg,int rpcId,simtime_t rtt );
    void handleRpcTimeout(BaseCallMessage* msg,
	    const TransportAddress& dest, int rpcId,
	    const OverlayKey& destKey = OverlayKey::UNSPECIFIED_KEY);

protected://methods: rpc distribution
    void sendRpc( const NodeHandle& handle, FindNodeCall* call,
                  BasePathLookup* listener, int rpcId );

    //-------------------------------------------------------------------------
    //- Construction & Destruction --------------------------------------------
    //-------------------------------------------------------------------------
public://construction & destruction
    BaseLookup( BaseOverlay* overlay, const BaseLookupConfiguration& config,
	    const cObject* findNodeExt = NULL );
    virtual ~BaseLookup();

protected:
    void start();
    void stop();
    void checkStop();

    //-------------------------------------------------------------------------
    //- AbstractLookup implementation -----------------------------------------
    //-------------------------------------------------------------------------
public://methods
    void lookup( const OverlayKey& key, int numSiblings = 1,
		 int hopCountMax = 0, int retries = 0,
                 LookupListener* listener = NULL);

    const NodeVector& getResult() const;

    bool isValid() const;
    void abortLookup();

    uint getAccumulatedHops() const;
};




/**
 * This class implements a path lookup.
 *
 * @author Sebastian Mies
 */
class BasePathLookup
{
    friend class BaseLookup;

protected://fields:
    BaseLookup* lookup;

protected://fields: state
    int  hops;
    int  step;
    int  pendingRpcs;
    bool finished;
    bool success;
    NodePairVector nextHops;
    std::map<TransportAddress, NodeHandle> oldNextHops;

protected://methods: rpc handling
    bool accepts( int rpcId );
    void handleResponse( FindNodeResponse* msg );
    void handleTimeout( BaseCallMessage* msg, const TransportAddress& dest,
	    int rpcId );
    void handleFailedNodeResponse( const NodeHandle& src,
	    cMessage* findNodeExt, bool retry);

private:
    void sendRpc( int num, cMessage* FindNodeExt = NULL );

protected:
    BasePathLookup( BaseLookup* lookup );
    virtual ~BasePathLookup();

    /**
     * Creates a find node call message. This method can be
     * overridden to add some additional state information to the
     * FindNodeCall message.
     *
     * @param findNodeExt Pointer to a optional cMessage, that may
     *                    contain overlay specific data to be attached
     *                    to FindNode RPCs and BaseRouteMessages
     * 
     * @returns Pointer to a new FindNodeCall message.
     */
    virtual FindNodeCall* createRpcMessage(cMessage *findNodeExt = NULL);

    /**
     * Adds a NodeHandle to next hops
     */
    void add( const NodeHandle& handle,
	    const NodeHandle& source = NodeHandle::UNSPECIFIED_NODE);
};

#endif

