//
// Copyright (C) 2006 Institut fuer Telematik, Universitaet Karlsruhe (TH)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//

/**
 * @file BaseOverlay.cc
 * @author Bernhard Heep (initial)
 * @author Sebastian Mies (rpcs)
 */

#include <UDPAppBase.h>
#include <UDPSocket.h>
#include <UDPControlInfo_m.h>
#include <IPAddressResolver.h>

#include <BootstrapOracleAccess.h>
#include <UnderlayConfiguratorAccess.h>
#include <GlobalStatisticsAccess.h>

#include <RpcMacros.h>
#include <RecursiveLookup.h>
#include <NodeHandle.h>
#include <OverlayKey.h>

#include "BaseOverlay.h"

//------------------------------------------------------------------------
//--- Initialization & finishing -----------------------------------------
//------------------------------------------------------------------------

BaseOverlay::BaseOverlay()
{
    bootstrapOracle = NULL;
    underlayConfigurator = NULL;
    notificationBoard = NULL;
}

BaseOverlay::~BaseOverlay()
{
    finishLookups();
    finishRpcs();
}

int BaseOverlay::numInitStages() const
{
    return MAX_STAGE_OVERLAY + 1;
}

void BaseOverlay::initialize(int stage)
{
    if (stage == MIN_STAGE_OVERLAY) {
        OverlayKey::setKeyLength(par("keyLength"));

        // fetch some parameters
        debugOutput = par("debugOutput");
        measureNetwInitPhase = par("measureNetwInitPhase");
        collectPerHopDelay = par("collectPerHopDelay");
        localPort = par("localPort");
        hopCountMax = par("hopCountMax");
        drawOverlayTopology = par("drawOverlayTopology");
        iterativeLookup = par("iterativeLookup");
        collectPerHopDelay = par("collectPerHopDelay");
        useCommonAPIforward = par("useCommonAPIforward");
        useNextHopRpc = par("useNextHopRpc");

        // set base lookup parameters
        baseLookupConfig.redundantNodes = par("lookupRedundantNodes");
        baseLookupConfig.parallelPaths = par("lookupParallelPaths");
        baseLookupConfig.parallelRpcs = par("lookupParallelRpcs");
        baseLookupConfig.secure = par("lookupSecure");
        baseLookupConfig.merge = par("lookupMerge");
        baseLookupConfig.failedNodeRpcs	= par("lookupFailedNodeRpcs");

        // statistics
        numSent = 0;
        bytesSent = 0;
        numSignalingSent = 0;
        bytesSignalingSent = 0;
        numReceived = 0;
        bytesReceived = 0;
        numOutbound = 0;
        bytesOutbound = 0;
        numInbound = 0;
        bytesInbound = 0;
        numSignalingReceived = 0;
        bytesSignalingReceived = 0;
        numForwarded = 0;
        bytesForwarded = 0;
        numSignalingForwarded = 0;
        bytesSignalingForwarded = 0;
        numDropped = 0;
        bytesDropped = 0;
        numFindNodeSent = 0;
        bytesFindNodeSent = 0;
        numFindNodeResponseSent = 0;
        bytesFindNodeResponseSent = 0;
        numFailedNodeSent = 0;
        bytesFailedNodeSent = 0;
        numFailedNodeResponseSent = 0;
        bytesFailedNodeResponseSent = 0;

        WATCH(numSent);
        WATCH(bytesSent);
        WATCH(numOutbound);
        WATCH(bytesOutbound);
        WATCH(numSignalingSent);
        WATCH(bytesSignalingSent);
        WATCH(numReceived);
        WATCH(bytesReceived);
        WATCH(numInbound);
        WATCH(bytesInbound);
        WATCH(numSignalingReceived);
        WATCH(bytesSignalingReceived);
        WATCH(numForwarded);
        WATCH(bytesForwarded);
        WATCH(numSignalingForwarded);
        WATCH(bytesSignalingForwarded);
        WATCH(numDropped);
        WATCH(bytesDropped);
        WATCH(numFindNodeSent);
        WATCH(bytesFindNodeSent);
        WATCH(numFindNodeResponseSent);
        WATCH(bytesFindNodeResponseSent);
        WATCH(numFailedNodeSent);
        WATCH(bytesFailedNodeSent);
        WATCH(numFailedNodeResponseSent);
        WATCH(bytesFailedNodeResponseSent);

        // set up UDP
        bindToPort(localPort);

        // find friend modules
        bootstrapOracle = BootstrapOracleAccess().get();
        underlayConfigurator = UnderlayConfiguratorAccess().get();
        notificationBoard = NotificationBoardAccess().get();

        // set up local nodehandle
        thisNode.ip = IPAddressResolver().
                      addressOf(parentModule()->parentModule()).get4();
        thisNode.port = localPort;
        thisNode.key = OverlayKey::UNSPECIFIED_KEY;

        // subscribe to the notification board
        notificationBoard->subscribe(this, NF_OVERLAY_TRANSPORTADDRESS_CHANGED);
        notificationBoard->subscribe(this, NF_OVERLAY_NODE_LEAVE);
        notificationBoard->subscribe(this, NF_OVERLAY_NODE_GRACEFUL_LEAVE);

        // set overlay pointer
        //overlay = this;

        // init visualization with terminal ptr
        if (drawOverlayTopology)
            initVis(parentModule()->parentModule());

        // register component
        registerComp(getThisCompType(), this);

        // init rpcs
        initRpcs();
        initLookups();

        // statistics
        globalStatistics->nodesInitialized++;
        creationTime = simTime();
    }

    if (stage >= MIN_STAGE_OVERLAY && stage <= MAX_STAGE_OVERLAY)
        initializeOverlay(stage);

    if (stage == MAX_STAGE_OVERLAY) {
        if ((bool)par("joinOnApplicationRequest") == false) {
            join();
        }
    }
}

void BaseOverlay::initializeOverlay(int stage)
{
}

void BaseOverlay::finish()
{
    finishOverlay();

    globalStatistics->nodesFinished++;
    recordScalar("BaseOverlay: Sent FindNode Responses",
                 numFindNodeResponseSent);
    recordScalar("BaseOverlay: Sent FindNodeResponse Bytes",
                 bytesFindNodeResponseSent);
    recordScalar("BaseOverlay: Sent FailedNode Messages", numFailedNodeSent);
    recordScalar("BaseOverlay: Sent FailedNode Bytes", bytesFailedNodeSent);
    recordScalar("BaseOverlay: Sent FailedNode Responses",
                 numFailedNodeResponseSent);
    recordScalar("BaseOverlay: Sent FailedNodeResponse Bytes",
                 bytesFailedNodeResponseSent);

    recordScalar("BaseOverlay: Sent Packets", numSent + numSignalingSent);
    recordScalar("BaseOverlay: Outbound Packets", numOutbound);
    recordScalar("BaseOverlay: Sent Bytes", bytesSent + bytesSignalingSent);
    recordScalar("BaseOverlay: Outbound Bytes", bytesOutbound);
    recordScalar("BaseOverlay: Received Packets", numReceived + numSignalingReceived);
    recordScalar("BaseOverlay: Inbound Packets", numInbound);
    recordScalar("BaseOverlay: Received Bytes", bytesReceived + bytesSignalingReceived);
    recordScalar("BaseOverlay: Inbound Bytes", bytesInbound);

    simtime_t time = globalStatistics->calcMeasuredLifetime(creationTime);
    if (time == 0) return;

    if (collectPerHopDelay) {
        std::ostringstream singleHopName;
        HopDelayRecord* hdrl = NULL;
        HopDelayRecord* hdr = NULL;
        for (size_t i = 0; i < singleHopDelays.size();) {
            hdrl = singleHopDelays[i++];
            hdr = hdrl;
            for (size_t j = 1; j <= i; ++j) {
                if (hdr->count == 0) continue;
                singleHopName.str("");
                singleHopName << "BaseOverlay: Average Delay in Hop "
                              << j << " of " << i;
                globalStatistics->addStdDev(singleHopName.str(),
                                            hdr->val / hdr->count);
                ++hdr;
            }
            delete[] hdrl;
        }
        singleHopDelays.clear();
    }

    globalStatistics->addStdDev("BaseOverlay: Sent User Messages/s", numSent / time);
    globalStatistics->addStdDev("BaseOverlay: Sent User Bytes/s", bytesSent / time);
    globalStatistics->addStdDev("BaseOverlay: Sent Outbound Messages/s", numOutbound / time);
    globalStatistics->addStdDev("BaseOverlay: Sent Outbound Bytes/s", bytesOutbound / time);
    globalStatistics->addStdDev("BaseOverlay: Sent Signaling Messages/s", numSignalingSent / time);
    globalStatistics->addStdDev("BaseOverlay: Sent Signaling Bytes/s", bytesSignalingSent / time);
    globalStatistics->addStdDev("BaseOverlay: Sent Total Messages/s", (numSent + numSignalingSent)/time);
    globalStatistics->addStdDev("BaseOverlay: Sent Total Bytes/s", (bytesSent + bytesSignalingSent)/time);
    globalStatistics->addStdDev("BaseOverlay: Sent FindNode Messages/s", numFindNodeSent / time);
    globalStatistics->addStdDev("BaseOverlay: Sent FindNode Bytes/s", bytesFindNodeSent / time);

    globalStatistics->addStdDev("BaseOverlay: Sent FindNodeResponse Messages/s", numFindNodeResponseSent / time);
    globalStatistics->addStdDev("BaseOverlay: Sent FindNodeResponse Bytes/s", bytesFindNodeResponseSent / time);
    globalStatistics->addStdDev("BaseOverlay: Sent FailedNode Messages/s", numFailedNodeSent / time);
    globalStatistics->addStdDev("BaseOverlay: Sent FailedNode Bytes/s", bytesFailedNodeSent / time);
    globalStatistics->addStdDev("BaseOverlay: Sent FailedNodeResponse Messages/s", numFailedNodeResponseSent / time);
    globalStatistics->addStdDev("BaseOverlay: Sent FailedNodeResponse Bytes/s", bytesFailedNodeResponseSent / time);

    globalStatistics->addStdDev("BaseOverlay: Received User Messages/s", numReceived / time);
    globalStatistics->addStdDev("BaseOverlay: Received User Bytes/s", bytesReceived / time);
    globalStatistics->addStdDev("BaseOverlay: Received Inbound Messages/s", numInbound / time);
    globalStatistics->addStdDev("BaseOverlay: Received Inbound Bytes/s", bytesInbound / time);
    globalStatistics->addStdDev("BaseOverlay: Received Signaling Messages/s", numSignalingReceived / time);
    globalStatistics->addStdDev("BaseOverlay: Received Signaling Bytes/s", bytesSignalingReceived / time);
    globalStatistics->addStdDev("BaseOverlay: Received Total Messages/s",
                                (numReceived + numSignalingReceived)/time);
    globalStatistics->addStdDev("BaseOverlay: Received Total Bytes/s",
                                (bytesReceived + bytesSignalingReceived)/time);
    globalStatistics->addStdDev("BaseOverlay: Forwarded User Messages/s", numForwarded / time);
    globalStatistics->addStdDev("BaseOverlay: Forwarded User Bytes/s", bytesForwarded / time);
    globalStatistics->addStdDev("BaseOverlay: Forwarded Signaling Messages/s", numSignalingForwarded / time);
    globalStatistics->addStdDev("BaseOverlay: Forwarded Signaling Bytes/s", bytesSignalingForwarded / time);
    globalStatistics->addStdDev("BaseOverlay: Forwarded Total Messages/s",
                                (numForwarded + numSignalingForwarded)/time);
    globalStatistics->addStdDev("BaseOverlay: Forwarded Total Bytes/s",
                                (bytesForwarded + bytesSignalingForwarded)/time);

    globalStatistics->addStdDev("BaseOverlay: Dropped Messages/s", numDropped / time);
    globalStatistics->addStdDev("BaseOverlay: Dropped Bytes/s", bytesDropped / time);

    globalStatistics->addStdDev("BaseOverlay: Active Time", simTime() - creationTime);

    globalStatistics->doFinish();

}

void BaseOverlay::finishOverlay()
{
}

//------------------------------------------------------------------------
//--- General Overlay Parameters (getter and setters) --------------------
//------------------------------------------------------------------------
bool BaseOverlay::isMalicious()
{
    return bootstrapOracle->isMalicious(getThisNode());
}

NodeHandle& BaseOverlay::getThisNode()
{
    return thisNode;
}

CompType BaseOverlay::getThisCompType()
{
    return OVERLAY_COMP;
}


//------------------------------------------------------------------------
//--- UDP functions copied from the INET framework .----------------------
//------------------------------------------------------------------------
void BaseOverlay::bindToPort(int port)
{
    cModule *node = parentModule()->parentModule();
    IPvXAddress ip = IPAddressResolver().addressOf(node);

    EV << "[BaseOverlay::bindToPort() @ " << ip
       << " (" << thisNode.key.toString(16) << ")]\n"
       << "    Binding to UDP port " << port
       << endl;

    // TODO UDPAppBase should be ported to use UDPSocket sometime, but for now
    // we just manage the UDP socket by hand...
    cMessage *msg = new cMessage("UDP_C_BIND", UDP_C_BIND);
    UDPControlInfo *ctrl = new UDPControlInfo();
    ctrl->setSrcPort(port);
    ctrl->setSockId(UDPSocket::generateSocketId());
    msg->setControlInfo(ctrl);
    send(msg, "to_udp");
}


//------------------------------------------------------------------------
//--- Overlay Common API: Key-based Routing ------------------------------
//------------------------------------------------------------------------

void BaseOverlay::callDeliver(BaseOverlayMessage* msg,
                              const OverlayKey& destKey)
{
    KBRdeliver* deliverMsg = new KBRdeliver();

    OverlayCtrlInfo* overlayCtrlInfo =
        check_and_cast<OverlayCtrlInfo*>(msg->removeControlInfo());

    BaseAppDataMessage* appDataMsg = dynamic_cast<BaseAppDataMessage*>(msg);

    // TODO GIA
    if (appDataMsg != NULL) {
        overlayCtrlInfo->setSrcComp(appDataMsg->getSrcComp());
        overlayCtrlInfo->setDestComp(appDataMsg->getDestComp());
    }

    deliverMsg->setControlInfo(overlayCtrlInfo);
    deliverMsg->setDestKey(destKey);

    deliverMsg->encapsulate(msg->decapsulate());

    deliverMsg->setType(KBR_DELIVER);

    cGate* destGate = getCompRpcGate(static_cast<CompType>(
            overlayCtrlInfo->getDestComp()));

    if (destGate == NULL) {
        opp_error("BaseOverlay::callDeliver(): Unknown destComp!");
    }

    sendDirect(deliverMsg, 0, destGate);

    delete msg;
}

void BaseOverlay::callForward(const OverlayKey& key, BaseRouteMessage* msg,
                              const NodeHandle& nextHopNode)
{
    KBRforward* forwardMsg = new KBRforward();

    forwardMsg->setDestKey(msg->getDestKey());
    forwardMsg->setNextHopNode(nextHopNode);
    forwardMsg->encapsulate(msg->encapsulatedMsg()->decapsulate());

    OverlayCtrlInfo* overlayCtrlInfo =
        new OverlayCtrlInfo();
    overlayCtrlInfo->setTransportType(ROUTE_TRANSPORT);
    overlayCtrlInfo->setHopCount(msg->getHopCount());
    overlayCtrlInfo->setSrcNode(msg->getSrcNode());
    overlayCtrlInfo->setSrcComp(check_and_cast<BaseAppDataMessage*>
        (msg->encapsulatedMsg())->getSrcComp());
    overlayCtrlInfo->setDestComp(check_and_cast<BaseAppDataMessage*>
        (msg->encapsulatedMsg())->getDestComp());

    if (msg->controlInfo() != NULL) {
        OverlayCtrlInfo* ctrlInfo =
            check_and_cast<OverlayCtrlInfo*>(msg->removeControlInfo());

        overlayCtrlInfo->setLastHopAddr(ctrlInfo->getLastHopAddr());
        overlayCtrlInfo->setLastHopPort(ctrlInfo->getLastHopPort());

        delete ctrlInfo;
    }

    forwardMsg->setControlInfo(overlayCtrlInfo);

    forwardMsg->setType(KBR_FORWARD);

    send(forwardMsg, "to_app");

    delete msg;
}

NodeVector* BaseOverlay::local_lookup(const OverlayKey& key,
                                      int num, bool safe)
{
    Enter_Method("local_lookup()");

    if (safe == true) {
        error("BaseOverlay::local_lookup(): safe flag is not implemented!");
    }

    NodeVector* nodeVector = findNode(key, num, num);

    if (((int)nodeVector->size()) > num)
        nodeVector->resize(num);

    return nodeVector;
}

void BaseOverlay::join(const OverlayKey& nodeID)
{
    Enter_Method("join()");

    // set nodeID and IP
    thisNode.ip =
        IPAddressResolver().addressOf(parentModule()->parentModule()).get4();
    if (nodeID.isUnspecified())  {
        setOwnNodeID();//thisNode.key = OverlayKey::random();
    } else {
        thisNode.key = nodeID;
    }

    callUpdate(thisNode, true);

    joinOverlay();
}

void BaseOverlay::setOwnNodeID()
{
    thisNode.key = OverlayKey::random();
}


NodeVector* BaseOverlay::neighborSet(int num)
{
    Enter_Method("neighborSet()");

    return local_lookup(thisNode.key, num, false);
}


void BaseOverlay::callUpdate(const NodeHandle& node, bool joined)
{
    KBRupdate* updateMsg = new KBRupdate("UPDATE");

    updateMsg->setNode(node);
    updateMsg->setJoined(joined);

    updateMsg->setType(KBR_UPDATE);

    send(updateMsg, "to_app");//"to_upperTier");
}

bool BaseOverlay::isSiblingFor(const NodeHandle& node, const OverlayKey& key,
                               int numSiblings, bool* err)
{
    Enter_Method("isSiblingFor()");

    opp_error( "isSiblingFor: Not implemented!" );

    return false;
}

int BaseOverlay::getMaxNumSiblings()
{
    Enter_Method("getMaxNumSiblings()");

    opp_error( "getMaxNumSiblings: Not implemented!" );

    return false;
}

int BaseOverlay::getMaxNumRedundantNodes()
{
    Enter_Method("getMaxNumRedundantNodes()");

    opp_error( "getMaxNumRedundantNodes: Not implemented!" );

    return false;
}

//------------------------------------------------------------------------
//--- Bootstrapping ------------------------------------------------------
//------------------------------------------------------------------------

void BaseOverlay::removeBootstrapNode(const NodeHandle& node)
{
    bootstrapOracle->removePeer(node);
}

const NodeHandle& BaseOverlay::getBootstrapNode()
{
    return bootstrapOracle->getBootstrapNode();
}

void BaseOverlay::registerBootstrapNode(const NodeHandle& node)
{
    bootstrapOracle->registerPeer(node);
}

//------------------------------------------------------------------------
//--- Message Handlers ---------------------------------------------------
//------------------------------------------------------------------------

//private
void BaseOverlay::handleMessage(cMessage* msg)
{
    // process self-messages
    if (msg->isSelfMessage()) {
        // process rpc self-messages
        BaseRpcMessage* rpcMessage = dynamic_cast<BaseRpcMessage*>(msg);
        if (rpcMessage!=NULL) {
            internalHandleRpcMessage(rpcMessage);
            return;
        }
        // process all other self-messages
        handleTimerEvent(msg);
    }

    // process messages from UDP
    else if (msg->arrivedOn("from_udp")) {
        UDPControlInfo* udpControlInfo =
            check_and_cast<UDPControlInfo*>(msg->removeControlInfo());
        OverlayCtrlInfo* overlayCtrlInfo = new OverlayCtrlInfo;
        //TODO TransportAddress
        overlayCtrlInfo->setLastHopAddr(udpControlInfo->srcAddr());
        overlayCtrlInfo->setLastHopPort(udpControlInfo->srcPort());
        overlayCtrlInfo->setTransportType(UDP_TRANSPORT);

        msg->setControlInfo(overlayCtrlInfo);
        delete udpControlInfo;

        BaseOverlayMessage* baseOverlayMsg =
            dynamic_cast<BaseOverlayMessage*>(msg);

        if (baseOverlayMsg == NULL) {
            RECORD_STATS(numDropped++; bytesDropped += msg->byteLength());
            delete msg;
            return;
        }

        // records stats if message is not a UDP "self message"
        if (overlayCtrlInfo->getLastHopAddr() != thisNode.ip ||
                overlayCtrlInfo->getLastHopPort() != thisNode.port) {
            // is this from anywhere else?
            if (baseOverlayMsg->getSignaling() == false)
                RECORD_STATS(numReceived++; bytesReceived +=
                                 baseOverlayMsg->byteLength());
            else
                RECORD_STATS(numSignalingReceived++;bytesSignalingReceived
                             += baseOverlayMsg->byteLength());
        }
        if (overlayCtrlInfo->getLastHopAddr() != thisNode.ip) {
            // is this from a different node?
            RECORD_STATS(numInbound++; bytesInbound +=
                             baseOverlayMsg->byteLength());
        }

        handleBaseOverlayMessage(baseOverlayMsg);
    }

    // process CommonAPIMessages from App
    else if (dynamic_cast<CommonAPIMessage*>(msg) != NULL) {
        if (dynamic_cast<KBRroute*>(msg) != NULL) {
            KBRroute* apiMsg = static_cast<KBRroute*>(msg);

            route(apiMsg->getDestKey(), static_cast<CompType>(apiMsg->getDestComp()),
                  static_cast<CompType>(apiMsg->getSrcComp()), apiMsg->decapsulate(),
                          apiMsg->getHint());
        } else if (dynamic_cast<KBRforward*>(msg) != NULL) {
            KBRforward* apiMsg = static_cast<KBRforward*>(msg);
            OverlayCtrlInfo* overlayCtrlInfo =
                check_and_cast<OverlayCtrlInfo*>
                (msg->removeControlInfo());

            BaseAppDataMessage* dataMsg =
                new BaseAppDataMessage();
            dataMsg->setType(APPDATA);
            dataMsg->setLength(BASEAPPDATA_L(dataMsg));
            dataMsg->setName(msg->encapsulatedMsg()->name());
            dataMsg->encapsulate(msg->decapsulate());
            dataMsg->setSrcComp(overlayCtrlInfo->getSrcComp());
            dataMsg->setDestComp(overlayCtrlInfo->getDestComp());
            dataMsg->setSignaling(false);

            BaseRouteMessage* routeMsg = new BaseRouteMessage(dataMsg->name());
            routeMsg->setType(OVERLAYROUTE);
            routeMsg->setLength(BASEROUTE_L(routeMsg));
            routeMsg->encapsulate(dataMsg);

            routeMsg->setSignaling(false);
            routeMsg->setDestKey(apiMsg->getDestKey());
            routeMsg->setSrcNode(overlayCtrlInfo->getSrcNode());
            routeMsg->setHopCount(overlayCtrlInfo->getHopCount());
            routeMsg->setControlInfo(overlayCtrlInfo);

            // message marked with this-pointer as already forwarded to tier1
            routeMsg->setContextPointer(this);

            sendToKey(apiMsg->getDestKey(), routeMsg, 1, apiMsg->getNextHopNode());
        }

        delete msg;
    }

    else if (dynamic_cast<BaseRpcMessage*>(msg) != NULL) {
        BaseRpcMessage* rpcMessage =
            static_cast<BaseRpcMessage*>(msg);
        internalHandleRpcMessage(rpcMessage);
    }

    // process other messages from App
    else if (msg->arrivedOn("from_app")) {
        handleAppMessage(msg);

    } else {
        opp_error("BaseOverlay::handleMessage(): Received msg with "
                  "unknown type!");
        delete msg;
    }
}

void BaseOverlay::handleBaseOverlayMessage(BaseOverlayMessage* msg,
        const OverlayKey& destKey)
{
    switch (msg->getType()) {
    case OVERLAYSIGNALING:
        handleUDPMessage(msg);
        return;

    case RPC: {
        // process rpc-messages
        BaseRpcMessage* rpcMessage =
            check_and_cast<BaseRpcMessage*>(msg);
        internalHandleRpcMessage(rpcMessage);
        return;
    }

    case APPDATA: {
        BaseAppDataMessage* baseAppDataMsg =
            check_and_cast<BaseAppDataMessage*>(msg);
        callDeliver(baseAppDataMsg, destKey);
        return;
    }

    case OVERLAYROUTE: {
        BaseRouteMessage* baseRouteMsg =
            check_and_cast<BaseRouteMessage*>(msg);

        if (baseRouteMsg->getDestKey().isUnspecified())
            opp_error("BaseOverlay: No destination key in ROUTE msg");


        // if this node is malicious drop the message
        if (isMalicious()) {
            EV << "[BaseOverlay::handleBaseOverlayMessage() @ " << thisNode.ip
            << " (" << thisNode.key.toString(16) << ")]\n"
            << "    BaseRouteMessage gets dropped because this node is malicious"
            << endl;
            RECORD_STATS(numDropped++;
                         bytesDropped += baseRouteMsg->byteLength());
            delete msg;
            return;
        }

        // collect delay-value of completed hop
        if (collectPerHopDelay) {
            baseRouteMsg->setHopDelayArraySize(baseRouteMsg->
                                               getHopDelayArraySize() + 1);
            baseRouteMsg->setHopDelay(baseRouteMsg->getHopDelayArraySize() - 1,
                                      simTime() - baseRouteMsg->getHopStamp());
        }

        bool err;
        if (isSiblingFor(thisNode, baseRouteMsg->getDestKey(),
                         1, &err) /*&& !err*/) {
            OverlayCtrlInfo* overlayCtrlInfo =
                check_and_cast<OverlayCtrlInfo*>(msg->removeControlInfo());

            overlayCtrlInfo->setHopCount(baseRouteMsg->getHopCount());
            overlayCtrlInfo->setSrcNode(baseRouteMsg->getSrcNode());

            BaseOverlayMessage* tmpMsg
            = check_and_cast<BaseOverlayMessage*>
              (baseRouteMsg->decapsulate());
            tmpMsg->setControlInfo(overlayCtrlInfo);

            if (collectPerHopDelay) {
                RECORD_STATS(
                    size_t i;
                    for (i = singleHopDelays.size();
                             i < baseRouteMsg->getHopDelayArraySize();) {
                        singleHopDelays.push_back(new HopDelayRecord[++i]);
                    }
                    i = baseRouteMsg->getHopDelayArraySize() - 1;
                    HopDelayRecord* hdr = singleHopDelays[i];
                    for (size_t j = 0; j <= i; ++j) {
                        hdr[j].count++;
                        hdr[j].val += baseRouteMsg->getHopDelay(j);
                    }
                );
            }
            if (iterativeLookup ||
                forwardMessageRecursive(thisNode, baseRouteMsg)) {
                handleBaseOverlayMessage(tmpMsg, baseRouteMsg->getDestKey());
                delete msg;
            }
            return;
        } else {
            // forward msg if this node is not responsible for the key
            sendToKey(baseRouteMsg->getDestKey(), baseRouteMsg, 1);
            return;
        }
        break;
    }

    default:
        EV << "[BaseOverlay::handleBaseOverlayMessage() @ " << thisNode.ip
        << " (" << thisNode.key.toString(16) << ")]\n"
        << "    Received unknown message from UDP of type " << msg->name()
        << endl;
        break;
    }
}

void BaseOverlay::receiveChangeNotification(int category, cPolymorphic* details)
{
    Enter_Method_Silent();
    if (category == NF_OVERLAY_TRANSPORTADDRESS_CHANGED) {
        handleTransportAddressChangedNotification();
    } else if (category == NF_OVERLAY_NODE_LEAVE) {
        handleNodeLeaveNotification();
    } else if (category == NF_OVERLAY_NODE_GRACEFUL_LEAVE) {
        handleNodeGracefulLeaveNotification();
    }
}

void BaseOverlay::handleTransportAddressChangedNotification()
{
    // get new ip address
    thisNode.ip = IPAddressResolver().addressOf(
                      parentModule()->parentModule()).get4();

    joinOverlay();
}

void BaseOverlay::handleNodeLeaveNotification()
{
    // ...
}

void BaseOverlay::handleNodeGracefulLeaveNotification()
{
    // ...
}

//virtual protected
void BaseOverlay::handleTimerEvent(cMessage* msg)
{
    // timer...
}

//virtual protected
void BaseOverlay::handleAppMessage(cMessage* msg)
{
    delete msg;
}

void BaseOverlay::handleUDPMessage(BaseOverlayMessage* msg)
{
    delete msg;
}

//virtual protected
void BaseOverlay::recordOverlaySentStats(BaseOverlayMessage* msg)
{
    // collect statistics ...
}

void BaseOverlay::setOverlayReady(bool ready)
{
    displayString().setTagArg("i", 1, ready ? "" : "red");
    bootstrapOracle->setOverlayReadyIcon(getThisNode(), ready);

    if (ready) {
        registerBootstrapNode(thisNode);
    } else {
        removeBootstrapNode(thisNode);
    }

    TierReadyMessage* msg = new TierReadyMessage;
    msg->setReady(ready);
    send( msg, "to_app");
}

//------------------------------------------------------------------------
//--- Messages -----------------------------------------------------------
//------------------------------------------------------------------------

void BaseOverlay::sendRouteMessage(const TransportAddress& dest,
                                   BaseRouteMessage* msg,
                                   bool ack)
{
    if (msg && !iterativeLookup && (dest != thisNode)) {
        msg->setHopCount(msg->getHopCount() + 1);
    }
    if (!ack)
        sendMessageToUDP(dest, msg);
    else {
        NextHopCall* nextHopCall = new NextHopCall(msg->name());
        nextHopCall->setLength(NEXTHOPCALL_L(nextHopCall));
        nextHopCall->encapsulate(msg);
        sendUdpRpcCall(dest, nextHopCall);
    }
}
void BaseOverlay::sendMessageToUDP(const TransportAddress& dest,
                                   BaseOverlayMessage* msg)
{
    // if there's still a control info attached to the message, remove it
    cPolymorphic* ctrlInfo = msg->removeControlInfo();
    if (ctrlInfo != NULL)
        delete ctrlInfo;

    // debug message
    if (debugOutput) {
        EV << "[BaseOverlay::sendMessageToUDP() @ " << thisNode.ip
        << " (" << thisNode.key.toString(16) << ")]\n"
        << "    Sending" << msg->name() << " to " << dest.ip
        << endl;
    }

    msg->setKind(UDP_C_DATA);
    UDPControlInfo* udpControlInfo = new UDPControlInfo();
    udpControlInfo->setSrcAddr(thisNode.ip);
    udpControlInfo->setSrcPort(thisNode.port);
    udpControlInfo->setDestAddr(dest.ip);
    udpControlInfo->setDestPort(dest.port);
    msg->setControlInfo(udpControlInfo);

    send(msg, "to_udp");

    if (dest != thisNode) {
        // record statistics, if message is not local
        if (dest.ip != thisNode.ip) {
            RECORD_STATS(numOutbound++; bytesOutbound += msg->byteLength());
        }
        if (msg->getSignaling() == false) {
            RECORD_STATS(numSent++; bytesSent += msg->byteLength());
        } else {
            RECORD_STATS(numSignalingSent++; bytesSignalingSent +=
                             msg->byteLength());
        }
        recordOverlaySentStats(msg);
    }
}


//------------------------------------------------------------------------
//--- RPC Handling -------------------------------------------------------
//------------------------------------------------------------------------

//public
void BaseOverlay::pingKey( const OverlayKey& destKey, simtime_t timeout,
                           int retries, const char* caption,
                           RpcListener* rpcListener, int rpcId )
{
    PingCall* call = createPingCall(caption);
    RECORD_STATS(numPingSent++; bytesPingSent += call->byteLength());
    sendRouteRpcCall(OVERLAY_COMP, destKey, call,
                     timeout, retries, rpcId, rpcListener);
}


//------------------------------------------------------------------------
//--- Basic Routing ------------------------------------------------------
//------------------------------------------------------------------------

static int pendingLookups = 0;

void BaseOverlay::initLookups()
{
    lookups = LookupSet();
}

void BaseOverlay::finishLookups()
{
    while (lookups.size() > 0) {
        (*lookups.begin())->abortLookup();
    }
    lookups.clear();
}

class SendToKeyListener : public LookupListener
{
private:
    BaseOverlay* overlay;
    BaseOverlayMessage* msg;
    GlobalStatistics* globalStatistics;
public:
    SendToKeyListener( BaseOverlay* overlay, BaseOverlayMessage* msg ) {
        this->overlay = overlay;
        this->msg = msg;
        globalStatistics = overlay->globalStatistics;
        pendingLookups++;
    }

    ~SendToKeyListener() {
        pendingLookups--;
        overlay = NULL;
        if (msg != NULL) {
            delete msg;
            msg = NULL;
        }
    }

    virtual void lookupFinished(AbstractLookup *lookup) {
        if (dynamic_cast<BaseRouteMessage*>(msg)) {
            BaseRouteMessage* routeMsg = static_cast<BaseRouteMessage*>(msg);
            if (lookup->isValid()) {
                if (lookup->getResult().size()==0) {
                    cout << "[ERROR] SendToKeyListener: Valid result, "
                    "but empty array." << endl;
                } else {
                    routeMsg->setHopCount(routeMsg->getHopCount()
                                          + lookup->getAccumulatedHops());

                    for (uint i=0; i<lookup->getResult().size(); i++) {
                        overlay->sendRouteMessage(lookup->getResult()[i],
                                                  static_cast<BaseRouteMessage*>
                                                   (routeMsg->dup()),
                                                  overlay->useNextHopRpc);
                    }
                }
            } else {
                EV << "[SendToKeyListener::lookupFinished()]\n"
                << "    Lookup failed - dropping message"
                << endl;
                RECORD_STATS(overlay->numDropped++;
                             overlay->bytesDropped += routeMsg->byteLength());
            }
        } else if (dynamic_cast<LookupCall*>(msg)) {
            LookupCall* call = static_cast<LookupCall*>(msg);
            LookupResponse* response = new LookupResponse();

            if (lookup->isValid()) {

                response->setIsValid(true);
                response->setKey(call->getKey());
                response->setSiblingsArraySize(lookup->getResult().size());
                for (uint i=0; i<lookup->getResult().size(); i++) {
                    response->setSiblings(i, lookup->getResult()[i]);
                }
                if (lookup->getResult().size() == 0) {
                    EV << "[SendToKeyListener::lookupFinished() @ "
                       << overlay->thisNode.ip
                       << " (" << overlay->thisNode.key.toString(16) << ")]\n"
                       << "    LookupCall "
                       << call->getNonce()
                       << " failed! (size=0)" << endl;
                }
            } else {
                response->setIsValid(false);
                EV << "[SendToKeyListener::lookupFinished() @ "
                   << overlay->thisNode.ip
                   << " (" << overlay->thisNode.key.toString(16) << ")]\n"
                   << "    LookupCall "
                   << call->getNonce()
                   << " failed!" << endl;
            }
            overlay->sendRpcResponse(call, response);
            msg = NULL;
        } else {
            throw new cRuntimeError("SendToKeyListener::lookupFinished(): "
                                    "Unkown message type!");
        }
        delete this;
    }
};

void BaseOverlay::route(const OverlayKey& key, CompType destComp,
                        CompType srcComp, cMessage* msg,
                        const TransportAddress& hint)
{
    if (key.isUnspecified() && hint.isUnspecified())
        error("route(): Key and hint unspecified!");

    // create base route message
    BaseAppDataMessage* baseAppDataMsg =
        new BaseAppDataMessage("BaseAppDataMessage");
    baseAppDataMsg->setType(APPDATA);
    baseAppDataMsg->setDestComp(destComp);
    baseAppDataMsg->setSrcComp(srcComp);
    baseAppDataMsg->setLength(BASEAPPDATA_L(baseAppDataMsg));

    baseAppDataMsg->setSignaling(false);
    baseAppDataMsg->encapsulate(msg);

    // debug output
    if (debugOutput) {
        EV << "[BaseOverlay::route() @ " << thisNode.ip
        << " (" << thisNode.key.toString(16) << ")]\n"
        << "    Received message from application"
        << endl;
    }

    if (key.isUnspecified()) {
        sendMessageToUDP(hint, baseAppDataMsg);
    } else {
        sendToKey( key, baseAppDataMsg, 1, hint );
    }
}

bool BaseOverlay::forwardMessageRecursive( const TransportAddress& dest,
                                           BaseRouteMessage* msg)
{
    return true;
}

void BaseOverlay::sendToKey(const OverlayKey& key, BaseOverlayMessage* msg,
                            int numSiblings, const TransportAddress& nextHop)
{
    BaseRouteMessage* routeMsg = NULL;

    if (debugOutput) {
        EV << "[BaseOverlay::sendToKey() @ " << thisNode.ip
        << " (" << thisNode.key.toString(16) << ")]\n"
        << "    Sending " << msg <<  " to " << key
        << endl;
    }

    if (key.isUnspecified())
        error("BaseOverlay::sendToKey(): unspecified destination key!");

    if (msg->getType() != OVERLAYROUTE) {
        routeMsg = new BaseRouteMessage("BaseRouteMessage");
        routeMsg->setType(OVERLAYROUTE);
        routeMsg->setDestKey(key);
        routeMsg->setSrcNode(thisNode);
        routeMsg->setSignaling(msg->getSignaling());
        // copy the name of the inner message
        routeMsg->setName(msg->name());
        routeMsg->setLength(BASEROUTE_L(routeMsg));
        routeMsg->encapsulate(msg);

        OverlayCtrlInfo* routeCtrlInfo = new OverlayCtrlInfo;
        routeCtrlInfo->setLastHopAddr(thisNode.ip);
        routeCtrlInfo->setLastHopPort(thisNode.port);
        routeCtrlInfo->setTransportType(ROUTE_TRANSPORT);
        routeMsg->setControlInfo(routeCtrlInfo);

        //message marked as not already forwarded to tier1
        routeMsg->setContextPointer(NULL);
    } else {
        routeMsg = check_and_cast<BaseRouteMessage*>(msg);
    }

    /* // only used as workaround - uncomment if assert fails!
    if (!routeMsg->controlInfo()) {
        // workaround
        OverlayCtrlInfo* routeCtrlInfo = new OverlayCtrlInfo;
        routeCtrlInfo->setLastHopAddr(thisNode.ip);
        routeCtrlInfo->setLastHopPort(thisNode.port);
        routeCtrlInfo->setTransportType(ROUTE_TRANSPORT);
        routeMsg->setControlInfo(routeCtrlInfo);
    }
    */

    // set timestamp for next hop
    if (collectPerHopDelay) {
        routeMsg->setHopStamp(simTime());
    }

    if (!nextHop.isUnspecified()) {
        // send msg to nextHop if specified (used for join rpcs)
        if (iterativeLookup && (nextHop != thisNode)) {
            routeMsg->setHopCount(routeMsg->getHopCount() + 1);
        }
        assert(routeMsg->controlInfo());
        sendRouteMessage(nextHop, routeMsg, useNextHopRpc);
        return;
    }

    if (iterativeLookup) {
        // create lookup and sent to key
        AbstractLookup* lookup = createLookup(routeMsg);
        lookup->lookup( routeMsg->getDestKey(), numSiblings, hopCountMax,
                        0, new SendToKeyListener( this, routeMsg ));
    } else  {
        NodeVector* nextHops = findNode(routeMsg->getDestKey(),
                getMaxNumRedundantNodes(), numSiblings, routeMsg);

        if (nextHops->size() == 0) {
            EV << "[BaseOverlay::sendToKey() @ " << thisNode.ip
               << " (" << thisNode.key.toString(16) << ")]\n"
               << "    FindNode() returned NULL - dropping message"
               << endl;
            // statistics
            RECORD_STATS(numDropped++; bytesDropped += routeMsg->byteLength());
            delete routeMsg;
        } else {
            // delete message if the hop count maximum is exceeded
            if (routeMsg->getHopCount() >= hopCountMax) {

                EV << "[BaseOverlay::sendToKey() @ " << thisNode.ip
                   << " (" << thisNode.key.toString(16) << ")]\n"
                   << "    Discards " << routeMsg->name() << " from "
                   << routeMsg->getSrcNode().ip << "\n"
                   << "    The hop count maximum has been exceeded ("
                   << routeMsg->getHopCount() << ">="
                   << hopCountMax << ")"
                   << endl;
                // statistics
                RECORD_STATS(numDropped++;
                             bytesDropped += routeMsg->byteLength());
                delete routeMsg;
                delete nextHops;
                return;
            }

            // callForward to app
            if (useCommonAPIforward &&
                dynamic_cast<BaseAppDataMessage*>(
                        routeMsg->encapsulatedMsg()) &&
                routeMsg->contextPointer() == NULL) {
                callForward(routeMsg->getDestKey(), routeMsg, (*nextHops)[0]);
                delete nextHops;
                return;
            }

            //message marked as not already forwarded
            routeMsg->setContextPointer(NULL);

            NodeHandle* nextHop = &((*nextHops)[0]);
            if (*nextHop == thisNode) {
                bool isSibling, err;
                isSibling = isSiblingFor(thisNode,routeMsg->getDestKey(),
                        numSiblings, &err);
                if (isSibling && !err) {
                    //EV << "[BaseOverlay::sendToKey() @ " << thisNode.ip
                    //   << " (" << thisNode.key.toString(16) << ")]\n"
                    //   << "    Forwards msg for key " << routeMsg->getDestKey() "\n"
                    //   << "    to node " << (*nextHops)[0]
                    //   << endl;
                    delete nextHops;
                    assert(routeMsg->controlInfo());
                    handleBaseOverlayMessage(routeMsg, key);
                    return;
                } else if (!isSibling && err) {
                    if(nextHops->size() > 1) {
                        nextHop = &((*nextHops)[1]);
                    } else {
                        EV << "[BaseOverlay::sendToKey() @ " << thisNode.ip
                           << " (" << thisNode.key.toString(16) << ")]\n"
                           << "    ERROR forwarding message " << *routeMsg
                           << "\n    from " << routeMsg->getSrcNode()
                           << " to " << routeMsg->getDestKey()
                           << "\n    (nextHops->size() = " << nextHops->size()
                           << ")" << endl;
                        RECORD_STATS(numDropped++; bytesDropped +=
                            routeMsg->byteLength());
                        delete routeMsg;
                        delete nextHops;
                        return;
                    }
                } else if (!isSibling && !err) {
                    error("nextHop = thisNode but isSiblingsFor() is false!");
                } else {
                    error("isSiblingsFor() is true with an error: "
                          "Erroneous method isSiblingFor()!");
                }
            }
            // forward msg if this node is not responsible for the key
            OverlayCtrlInfo* overlayCtrlInfo =
                    dynamic_cast<OverlayCtrlInfo*>(msg->controlInfo());

            // records statistics, if we forward this message
            if ((overlayCtrlInfo != NULL) && (overlayCtrlInfo->getLastHopAddr()
                    != thisNode.ip)) {
                if (routeMsg->getSignaling() == false) {
                    RECORD_STATS(numForwarded++; bytesForwarded +=
                            routeMsg->byteLength());
                } else {
                    RECORD_STATS(numSignalingForwarded++;
                            bytesSignalingForwarded +=
                            routeMsg->byteLength());
                }
            }

            if (forwardMessageRecursive(*nextHop, routeMsg)) {
                delete routeMsg->removeControlInfo();
                sendRouteMessage(*nextHop, routeMsg, useNextHopRpc);
            }
        }
        delete nextHops;
    }
}

//protected: create a iterative lookup class
AbstractLookup* BaseOverlay::createLookup(const BaseOverlayMessage* msg,
                                          const cObject* findNodeExt)
{
    //Enter_Method_Silent();

    AbstractLookup* newLookup;
    iterativeLookup ?
        (newLookup = new BaseLookup(this, baseLookupConfig, findNodeExt)) :
        (newLookup = new RecursiveLookup(this));

    lookups.insert( newLookup );
    return newLookup;
}

void BaseOverlay::removeLookup( AbstractLookup* lookup )
{
    lookups.erase(lookup);
}

//virtual public
OverlayKey BaseOverlay::distance( const OverlayKey& x,
                                  const OverlayKey& y ) const
{
    opp_error("BaseOverlay::distance(): Not implemented!");
    return OverlayKey::UNSPECIFIED_KEY;
}

//protected: find closest nodes
NodeVector* BaseOverlay::findNode(const OverlayKey& key,
                                  int numRedundantNodes,
                                  int numSiblings,
                                  BaseOverlayMessage* msg)
{
    opp_error( "findNode: Not implemented!" );
    return NULL;
}

//protected: join the overlay with a given nodeID
void BaseOverlay::joinOverlay()
{
//  std::cout << "BaseOverlay::joinOverlay(): Not implemented!" << endl;
    return;
}

bool BaseOverlay::handleFailedNode(const TransportAddress& failed)
{
    return false;
}

//------------------------------------------------------------------------
//--- RPCs ---------------------------------------------------------------
//------------------------------------------------------------------------

void BaseOverlay::sendRpcMessageWithTransport(TransportType transportType,
                                              CompType destComp,
                                              const TransportAddress& dest,
                                              const OverlayKey& destKey,
                                              BaseRpcMessage* message)
{
    switch (transportType) {
    case UDP_TRANSPORT:
        sendMessageToUDP( dest, message );
        break;
    case ROUTE_TRANSPORT:
        sendToKey( destKey, message, 1, dest );
        break;
    case INTERNAL_TRANSPORT: {
        cGate *destCompGate = getCompRpcGate(destComp);
        if (destCompGate == NULL) {
            opp_error("BaseOverlay::sendRpcMessageWithTransport():"
                      " INTERNAL_RPC to unknown RpcCompType!");
        }
        OverlayCtrlInfo *overlayCtrlInfo = new OverlayCtrlInfo();
        overlayCtrlInfo->setSrcComp(getThisCompType());
        overlayCtrlInfo->setDestComp(destComp);
        overlayCtrlInfo->setTransportType(INTERNAL_TRANSPORT);
        message->setControlInfo(overlayCtrlInfo);
        sendDirect( message, 0, destCompGate);
        break;
    }
    default:
        opp_error("BaseOverlay::sendRpcMessageWithTransport: "
                  "invalid transportType!");
        break;
    }
}

//private
bool BaseOverlay::internalHandleRpc( BaseCallMessage* msg )
{
    // call rpc stubs
    RPC_SWITCH_START( msg );
    RPC_DELEGATE( FindNode, findNodeRpc );
    RPC_DELEGATE( FailedNode, failedNodeRpc );
    RPC_DELEGATE( Lookup, lookupRpc );
    RPC_DELEGATE( NextHop, nextHopRpc );
    RPC_SWITCH_END( );

    // if RPC was handled return true, else tell the parent class to handle it
    return RPC_HANDLED || BaseRpc::internalHandleRpc(msg);
}

bool BaseOverlay::internalHandleRpcResponse(BaseResponseMessage* msg,
                                            int rpcId, simtime_t rtt)
{
    // if RPC was handled return true, else tell the parent class to handle it
    return BaseRpc::internalHandleRpcResponse(msg, rpcId, rtt);
}

void BaseOverlay::internalHandleRpcTimeout(BaseCallMessage* msg,
                                           const TransportAddress& dest,
                                           int rpcId, const OverlayKey& destKey)
{
    BaseRpc::internalHandleRpcTimeout(msg, dest, rpcId, destKey);

    RPC_SWITCH_START( msg )
        RPC_ON_CALL( NextHop )
        {
            //TODO neighborFailed(dest);
            BaseRouteMessage* tempMsg
                = check_and_cast<BaseRouteMessage*>(msg->decapsulate());

            if (!tempMsg->controlInfo()) {
                OverlayCtrlInfo* overlayCtrlInfo = new OverlayCtrlInfo;
                overlayCtrlInfo->setLastHopAddr(thisNode.ip);
                overlayCtrlInfo->setLastHopPort(thisNode.port);
                overlayCtrlInfo->setHopCount(tempMsg->getHopCount());
                overlayCtrlInfo->setSrcNode(tempMsg->getSrcNode());
                overlayCtrlInfo->setTransportType(UDP_TRANSPORT);
                tempMsg->setControlInfo(overlayCtrlInfo);
            }

            handleBaseOverlayMessage(tempMsg, destKey);
            break;
        }
    RPC_SWITCH_END( )
}

//protected: statistic helpers for BaseLookup
void BaseOverlay::countFindNodeCall( const FindNodeCall* call )
{
    RECORD_STATS(numFindNodeSent++;
                 bytesFindNodeSent += call->byteLength());
}

void BaseOverlay::countFailedNodeCall( const FailedNodeCall* call )
{
    RECORD_STATS(numFailedNodeSent++;
                 bytesFailedNodeSent += call->byteLength());
}

//private: rpc stub
void BaseOverlay::findNodeRpc( FindNodeCall* call )
{
    // if this node is malicious don't answer a findNodeCall
    if (isMalicious()) {
        EV << "[BaseOverlay::findNodeRpc() @ " << thisNode.ip
        << " (" << thisNode.key.toString(16) << ")]\n"
        << "    Node ignores findNodeCall because this node is malicious"
        << endl;
        delete call;
        return;
    }

    FindNodeResponse* findNodeResponse =
        new FindNodeResponse("FindNodeResponse");

    NodeVector* nextHops = findNode(call->getLookupKey(),
                                    call->getNumRedundantNodes(),
                                    call->getNumSiblings(), call);

    findNodeResponse->setClosestNodesArraySize(nextHops->size());

    for (uint i=0; i < nextHops->size(); i++) {
        findNodeResponse->setClosestNodes(i, (*nextHops)[i]);
    }

    bool err;
    if (isSiblingFor(thisNode, call->getLookupKey(), call->getNumSiblings(),
                     &err)) {
        findNodeResponse->setSiblings(true);
    }

    findNodeResponse->setLength(FINDNODERESPONSE_L(findNodeResponse));

    if (call->hasObject("findNodeExt")) {
        cMessage* findNodeExt = (cMessage*)call->removeObject("findNodeExt");
        findNodeResponse->addObject(findNodeExt);
        findNodeResponse->addLength(findNodeExt->length());
    }

    RECORD_STATS(numFindNodeResponseSent++; bytesFindNodeResponseSent +=
                     findNodeResponse->byteLength());

    delete nextHops;

    sendRpcResponse(call, findNodeResponse);
}

void BaseOverlay::failedNodeRpc( FailedNodeCall* call )
{
    FailedNodeResponse* failedNodeResponse =
        new FailedNodeResponse("FailedNodeResponse");
    failedNodeResponse->setTryAgain(handleFailedNode(call->getFailedNode()));
    failedNodeResponse->setLength(FAILEDNODERESPONSE_L(failedNodeResponse));

    if (call->hasObject("findNodeExt")) {
        cMessage* findNodeExt = check_and_cast<cMessage*>(
                                    call->removeObject("findNodeExt"));
        failedNodeResponse->addObject(findNodeExt);
        failedNodeResponse->addLength(findNodeExt->length());
    }

    RECORD_STATS(numFailedNodeResponseSent++; bytesFailedNodeResponseSent +=
                     failedNodeResponse->byteLength());

    sendRpcResponse(call, failedNodeResponse);
}

void BaseOverlay::lookupRpc(LookupCall* call)
{
    // create lookup and sent to key
    AbstractLookup* lookup = createLookup(call);
    lookup->lookup(call->getKey(), call->getNumSiblings(), hopCountMax,
                   1, new SendToKeyListener( this, call ));

}

void BaseOverlay::nextHopRpc(NextHopCall* call)
{
    BaseRouteMessage* routeMsg
        = check_and_cast<BaseRouteMessage*>(call->decapsulate());

    OverlayCtrlInfo* overlayCtrlInfo =
        check_and_cast<OverlayCtrlInfo*>(call->controlInfo()->dup());
    overlayCtrlInfo->setHopCount(routeMsg->getHopCount());
    overlayCtrlInfo->setSrcNode(routeMsg->getSrcNode());

    /*
    std::cout << overlayCtrlInfo->getHopCount() << "\n"
              << overlayCtrlInfo->getLastHopAddr() << "\n"
              << overlayCtrlInfo->getLastHopPort() << "\n"
              << overlayCtrlInfo->getSrcNode() << "\n"
              << overlayCtrlInfo->getSrcComp() << "\n"
              << overlayCtrlInfo->getDestComp() << "\n" << std::endl;
    */

    routeMsg->setControlInfo(overlayCtrlInfo);
    assert(routeMsg->controlInfo());

    std::string temp("ACK: [");
    (temp += routeMsg->name()) += "]";

    NextHopResponse* response
        = new NextHopResponse(temp.c_str());
    response->setLength(NEXTHOPRESPONSE_L(response));
    sendRpcResponse(call, response);

    handleBaseOverlayMessage(routeMsg, routeMsg->getDestKey());
}
void BaseOverlay::insertBootstrapNode(const NodeHandle& node)
{
    Enter_Method_Silent();

    EV << "New bootstrap node " << node << endl;
}


void BaseOverlay::registerComp(CompType compType, cModule *module)
{
    cGate *gate = NULL;

    if (module != NULL)
        gate = module->gate("direct_in");

    compModuleList[compType] = make_pair<cModule*, cGate*>(module, gate);
}

cModule* BaseOverlay::getCompModule(CompType compType)
{
    CompModuleList::iterator it = compModuleList.find(compType);

    if (it != compModuleList.end())
        return it->second.first;
    else
        return NULL;
}

cGate* BaseOverlay::getCompRpcGate(CompType compType)
{
    CompModuleList::iterator it = compModuleList.find(compType);

    if (it != compModuleList.end())
        return it->second.second;
    else
        return NULL;
}

