//
// Copyright (C) 2006 Institut fuer Telematik, Universitaet Karlsruhe (TH)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//

/**
 * @file BaseRpc.cc
 * @author Bernhard Heep (initial)
 * @author Sebastian Mies (rpcs)
 * @author Gregoire Menuel (separation from BaseRpc)
 */

#include <CommonMessages_m.h>
#include <BootstrapOracleAccess.h>
#include <UnderlayConfiguratorAccess.h>
#include <GlobalStatisticsAccess.h>
#include <GlobalParametersAccess.h>


#include "BaseRpc.h"
#include "RpcMacros.h"



//------------------------------------------------------------------------
//--- Initialization & finishing -----------------------------------------
//------------------------------------------------------------------------


//------------------------------------------------------------------------
//--- RPC Handling -------------------------------------------------------
//------------------------------------------------------------------------

BaseRpc::BaseRpc()
{
    defaultRpcListener = NULL;
}

//private
void BaseRpc::initRpcs()
{
    // set friend modules
    globalStatistics = GlobalStatisticsAccess().get();
    globalParameters = GlobalParametersAccess().get();
        
    rpcsPending = 0;
    rpcStates.clear();

    defaultRpcListener = new RpcListener();

    //set ping cache
    numPingSent = 0;
    bytesPingSent = 0;
    numPingResponseSent = 0;
    bytesPingResponseSent = 0;
        
    WATCH(numPingSent);
    WATCH(bytesPingSent);
    WATCH(numPingResponseSent);
    WATCH(bytesPingResponseSent);

    cModule *mod = parentModule();
    pingCache = NULL;
    while (pingCache == NULL) {
        pingCache = (PingCache*)mod->submodule("pingCache");
        mod = mod->parentModule();
        if (!mod)
            opp_error ("BaseRpc::initRpc: Module type contains no ping cache!");
    }
}

//private
void BaseRpc::finishRpcs()
{
    // stop all rpcs
    for (RpcStates::iterator i = rpcStates.begin();
        i != rpcStates.end(); i++) {
        cancelAndDelete(i->second.callMsg);
        cancelAndDelete(i->second.timeoutMsg);
    }
    rpcStates.clear();

    // delete default rpc listener
    if (defaultRpcListener!=NULL) {
        delete defaultRpcListener;
        defaultRpcListener = NULL;
    }
}

//public
uint32_t BaseRpc::sendRpcMessage(const TransportAddress& dest,
                                 BaseCallMessage* msg,
                                 RpcListener* rpcListener,
                                 const OverlayKey& destKey,
                                 int rpcId, simtime_t timeout,
                                 int retries)
{
    if (destKey.isUnspecified()) {
        if (getThisCompType() == OVERLAY_COMP) {            
            return sendUdpRpcCall(dest, msg, timeout, retries,
                                  rpcId, rpcListener);
        } else {
            return sendRouteRpcCall(getThisCompType(), dest, msg,
                                    timeout, retries, rpcId, rpcListener);
        }
    } 
    // send message to key
    return sendRouteRpcCall(getThisCompType(), dest, destKey, msg,
                       timeout, retries, rpcId, rpcListener);
}  

uint32_t BaseRpc::sendRpcMessage(RpcAddrType destType,
                                 const TransportAddress& dest,
                                 BaseCallMessage* msg,
                                 RpcListener* rpcListener,
                                 const OverlayKey& destKey,
                                 int rpcId, simtime_t timeout,
                                 int retries)
{
    TransportType transportType = INVALID_TRANSPORT;
    CompType destComp = INVALID_COMP;
    
    switch (destType) {
    case RPC_TO_OVERLAY:
        transportType = INTERNAL_TRANSPORT;
        destComp = OVERLAY_COMP;
        break;
    case RPC_TO_TIER1:
        transportType = INTERNAL_TRANSPORT;
        destComp = TIER1_COMP;
        break;
    case RPC_TO_TIER2:
        transportType = INTERNAL_TRANSPORT;
        destComp = TIER2_COMP;
        break;
    case RPC_TO_TIER3:
        transportType = INTERNAL_TRANSPORT;
        destComp = TIER3_COMP;
        break;
    case RPC_TO_KEY:
        transportType = ROUTE_TRANSPORT;
        destComp = getThisCompType();
        break;
    case RPC_TO_UDP:
        if (getThisCompType() == OVERLAY_COMP) {
            transportType = UDP_TRANSPORT;
        } else {
            transportType = ROUTE_TRANSPORT;
            destComp = getThisCompType();
        }
        break;
    default:
        opp_error("BaseRpc::sendRpcMessage(): Deprecated parameters!");
    }
    
    return sendRpcCall(transportType, destComp, dest, destKey, msg,
                       timeout, retries, rpcId, rpcListener);
}

uint32_t BaseRpc::sendRpcCall(TransportType transportType,
                              CompType destComp,
                              const TransportAddress& dest,
                              const OverlayKey& destKey,
                              BaseCallMessage* msg,
                              simtime_t timeout,
                              int retries,
                              int rpcId,
                              RpcListener* rpcListener)
{
    // create nonce, timeout and set default parameters
    uint nonce;
    do {
        nonce = intuniform( 0, 2147483647);
    } while (rpcStates.count(nonce) > 0);

    if (timeout == -1) {
        switch (transportType) {
        case INTERNAL_TRANSPORT:
            timeout = 0;
            break;
        case UDP_TRANSPORT:
            timeout = globalParameters->getRpcUdpTimeout();
            break;
        case ROUTE_TRANSPORT:
            timeout = (destKey.isUnspecified() ? 
                       globalParameters->getRpcUdpTimeout() :
                       globalParameters->getRpcKeyTimeout());                        
            break;
        default:
            throw new cException("BaseRpc::sendRpcMessage(): "
                                 "Unknown RpcTransportType!");
        }
    }
    
    if (rpcListener == NULL)
        rpcListener = defaultRpcListener;

    // create state
    RpcState state;
    state.id = rpcId;
    state.timeSent = simulation.simTime();
    state.dest = dest;
    state.destKey = destKey;
    state.srcComp = getThisCompType();
    state.destComp = destComp;
    state.listener = rpcListener;
    state.timeoutMsg = new RpcTimeoutMessage();
    state.timeoutMsg->setNonce(nonce);
    state.retries = retries;
    state.transportType = transportType;

    if (rpcStates.count(nonce) > 0)
        error("RPC nonce collision");

    // set message parameters
    msg->setNonce(nonce);
    msg->setSrcNode(thisNode);
    msg->setType(RPC);

    // save copy of call message in RpcState
    state.callMsg = dynamic_cast<BaseCallMessage*>(msg->dup());
    assert(!msg->encapsulatedMsg() || !msg->encapsulatedMsg()->controlInfo());
            
    // register state
    rpcStates[nonce] = state;

    // schedule timeout message
    if (timeout!=0)
        scheduleAt( simulation.simTime() + timeout, state.timeoutMsg);

    sendRpcMessageWithTransport(transportType, destComp, dest, destKey, msg);

    return nonce;
}


//public
void BaseRpc::cancelRpcMessage( uint32_t nonce )
{
    if (rpcStates.count(nonce)==0)
        return;
    RpcState state = rpcStates[nonce];
    rpcStates.erase(nonce);
    cancelAndDelete(state.callMsg);
    cancelAndDelete(state.timeoutMsg);
}

//protected
void BaseRpc::internalHandleRpcMessage( BaseRpcMessage* msg)
{
    // check if this is a rpc call message
    BaseCallMessage* rpCall = dynamic_cast<BaseCallMessage*>(msg);
    if (rpCall != NULL) {
        bool rpcHandled = true;
        if (!handleRpc(rpCall)) rpcHandled = internalHandleRpc(rpCall);
        if (!rpcHandled) {
            EV << "[BaseRpc::internalHandleRpcMessage() @ " << thisNode.ip
               << " (" << thisNode.key.toString(16) << ")]\n"
               << "    Error: RPC '" << msg->fullName()<< "' was not handled"
               << endl;
            delete msg;
        }
        return;
    }

    // get nonce
    int nonce = msg->getNonce();

    // nonce known? no -> delete message and return
    if (rpcStates.count(nonce)==0) {
        EV << "[BaseRpc::internalHandleRpcMessage() @ " << thisNode.ip
           << " " << thisNode.key.toString(16) << ")]\n"
           << "    RPC: Nonce Unknown"
           << endl;    	
        delete msg;
        return;
    }

    // get state and remove from map
    RpcState state = rpcStates[nonce];
    rpcStates.erase(nonce);

    // is timeout message?
    if (msg->isSelfMessage() &&
        (dynamic_cast<RpcTimeoutMessage*>(msg) != NULL)) {
        // yes-> inform listener

        // retry?
        state.retries--;
        simtime_t timeout = simulation.simTime() - state.timeSent;
        if (state.retries>=0) {
            sendRpcMessageWithTransport(state.transportType, state.destComp,
                                        state.dest,
                                        state.destKey,
                                        dynamic_cast<BaseCallMessage*>
                                        (state.callMsg->dup()));
            if (timeout!=0)
                scheduleAt( simulation.simTime() + timeout, msg );

            state.timeSent = simulation.simTime();
            rpcStates[nonce] = state;
            return;
        }

        // inform listener
        if ( state.listener != NULL )
            state.listener->handleRpcTimeout( state.callMsg, state.dest,
                                              state.id, state.destKey );

        // inform overlay
        internalHandleRpcTimeout(state.callMsg, state.dest,
                                 state.id, state.destKey);
        handleRpcTimeout( state.callMsg, state.dest, state.id, state.destKey );

    } else { // no-> handle rpc response

        // get parameters
        simtime_t rtt = simulation.simTime() - state.timeSent;
        BaseResponseMessage* response
        = dynamic_cast<BaseResponseMessage*>(msg);

        // inform listener
        if ( state.listener != NULL )
            state.listener->handleRpcResponse( response, state.id, rtt );

        // inform overlay
        internalHandleRpcResponse(response, state.id, rtt);
        handleRpcResponse(response, state.id, rtt);

        // delete response
        delete response;
    }

    // delete messages
    cancelAndDelete(state.callMsg);
    cancelAndDelete(state.timeoutMsg);

    // clean up pointers
    state.callMsg = NULL;
    state.timeoutMsg = NULL;
}

//private
bool BaseRpc::internalHandleRpc( BaseCallMessage* msg )
{
    RPC_SWITCH_START( msg );
    RPC_DELEGATE( Ping, pingRpc );
    RPC_SWITCH_END( );

    return RPC_HANDLED;
}

bool BaseRpc::internalHandleRpcResponse(BaseResponseMessage* msg, int rpcId, simtime_t rtt)
{
    // call rpc stubs
    RPC_SWITCH_START( msg );
    RPC_ON_RESPONSE( Ping ) {
        pingRpcResponse(_PingResponse, rpcId, rtt);
        break;
    }
    RPC_SWITCH_END( );
    
    return RPC_HANDLED;
}

void BaseRpc::internalHandleRpcTimeout(BaseCallMessage* msg,
                                       const TransportAddress& dest,
                                       int rpcId, const OverlayKey& destKey)
{
    RPC_SWITCH_START( msg ) {
        RPC_ON_CALL( Ping ) {
            pingCache->addressTimeout(dest);
        }
    }
    RPC_SWITCH_END( )
}
//virtual protected
bool BaseRpc::handleRpc( BaseCallMessage* msg )
{
    return false;
}

void BaseRpc::sendRpcResponse(TransportType transportType,
                              const TransportAddress &dest,
                              const OverlayKey &destKey,
                              BaseCallMessage* call,
                              BaseResponseMessage* response)
{
    if (call==NULL || response==NULL) {
        if ( call!=NULL )
            delete call;
        return;
    }
    
    OverlayCtrlInfo* overlayCtrlInfo =
            check_and_cast<OverlayCtrlInfo*>(call->removeControlInfo());
            
    response->setSrcNode(thisNode);
    response->setType(RPC);
    response->setNonce(call->getNonce());
    sendRpcMessageWithTransport(transportType,
                                static_cast<CompType>
                                (overlayCtrlInfo->getSrcComp()),
                                dest, destKey, response);
    delete overlayCtrlInfo;
    delete call;

}

//protected
void BaseRpc::sendRpcResponse( BaseCallMessage* call,
                               BaseResponseMessage* response )
{
    TransportType transportType;
    
    OverlayCtrlInfo* overlayCtrlInfo =
        check_and_cast<OverlayCtrlInfo*>(call->controlInfo());
     
    if (overlayCtrlInfo->getTransportType() == INTERNAL_TRANSPORT) {
        transportType = INTERNAL_TRANSPORT;
    } else if (getThisCompType() == OVERLAY_COMP) {
        transportType = UDP_TRANSPORT;
    } else {
        transportType = ROUTE_TRANSPORT;
    }
    
    sendRpcResponse(transportType, call->getSrcNode(), call->getSrcNode().key,
                    call, response);
}

//protected
void BaseRpc::sendInternalRpcResponse(BaseCallMessage* call,
                                      BaseResponseMessage* response)
{
    sendRpcResponse(call, response);
}

void BaseRpc::sendRpcResponse(int destType, BaseCallMessage* call,
                              BaseResponseMessage* response)
{

    TransportType transportType = INVALID_TRANSPORT;

    switch (destType) {
    case RPC_TO_OVERLAY:
    case RPC_TO_TIER1:
    case RPC_TO_TIER2:
    case RPC_TO_TIER3:
        transportType = INTERNAL_TRANSPORT;
        break;
    case RPC_TO_KEY:
        transportType = ROUTE_TRANSPORT;
        break;
    case RPC_TO_UDP:
        if (getThisCompType() == OVERLAY_COMP) {
            transportType = UDP_TRANSPORT;
        } else {
            transportType = ROUTE_TRANSPORT;
        }
        break;
    default:
        opp_error("BaseRpc::sendRpcResponse(): Deprecated parameters!");
    }
    
    sendRpcResponse(transportType, call->getSrcNode(),
                    call->getSrcNode().key, call, response);
}

// ping RPC stuff
void BaseRpc::pingResponse(PingResponse* response, int rpcId, simtime_t rtt) 
{
}

void BaseRpc::pingTimeout(PingResponse* response, int rpcId) 
{
}

void BaseRpc::pingRpc( PingCall* call )
{
    PingResponse* response = ping(call);
    RECORD_STATS(numPingResponseSent++; bytesPingResponseSent +=
        response->byteLength());
    sendRpcResponse(call, response );
}

PingResponse* BaseRpc::ping( PingCall* call )
{
    std::string pongName(call->name());
    pongName[1] = 'O';
    PingResponse* response = new PingResponse(pongName.c_str());

    response->setLength( PINGRESPONSE_L(response) );
    return response;
}

void BaseRpc::pingRpcResponse(PingResponse* response, int rpcId, simtime_t rtt)
{
    pingCache->insertAddressRtt(response->getSrcNode(), rtt);
    pingResponse(response, rpcId, rtt);
}

PingCall* BaseRpc::createPingCall( const char* caption )
{
    PingCall* call;

    if (caption) {
        call = new PingCall(caption);
    } else {
        call = new PingCall("PING");
    }

    call->setLength(PINGCALL_L(call));
    return call;
}

void BaseRpc::pingNode(const TransportAddress& dest, simtime_t timeout,
                       int retries, const char* caption,
                       RpcListener* rpcListener, int rpcId, bool overrideCache)
{
    if (overrideCache || pingCache->getAddressRtt(dest) == PINGSTATE_UNKNOWN) {
        /* if unknown or overriding, send ping call as normal */
        PingCall* call = createPingCall(caption);
        RECORD_STATS(numPingSent++; bytesPingSent += call->byteLength());
        sendUdpRpcCall(dest, call, timeout, retries, rpcId, rpcListener);
        pingCache->setAddressPinged(dest);
    } else if (pingCache->getAddressRtt(dest) != PINGSTATE_WAITING) {        
        /* if it's time-out or a known value, reply with fake ping response */
        PingResponse *pr = new PingResponse();
        pr->setNonce(0);
        pr->setSrcNode(pingCache->getNodeHandle(dest));
        handleRpcResponse(pr, rpcId, pingCache->getAddressRtt(dest));
        delete pr;
    }
}
