// Copyright (C) 2006 Institut fuer Telematik, Universitaet Karlsruhe (TH)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//

/**
 * @file BaseRpc.h
 * @author Bernhard Heep
 * @author Sebastian Mies
 * @author Gregoire Menuel
 */

#ifndef __BASERPC_H_
#define __BASERPC_H_


class BootstrapOracle;
class UnderlayConfigurator;

#include <omnetpp.h>
#include <GlobalParameters.h>
#include "PingCache.h"

// rpc includes
#include <RpcListener.h>
#include <RpcMacros.h>
#include <CommonMessages_m.h>

using namespace __gnu_cxx;

// TODO: deprecated - use TransportType instead!
enum RpcAddrType {
    RPC_TO_OVERLAY,
    RPC_TO_TIER1,
    RPC_TO_TIER2,
    RPC_TO_TIER3,
    RPC_TO_UPPERTIER,
    RPC_TO_LOWERTIER,
    RPC_TO_UDP,
    RPC_TO_KEY
};


/**
 * Base class for RPCs
 *
 * Base class for RPCs.
 * 
 * @author Bernhard Heep (initial)
 * @author Sebastian Mies (rpc, lookup)
 * @author Gregoire Menuel (separation from BaseRpc)
 */
class BaseRpc : public RpcListener,
                public cSimpleModule
{

    //------------------------------------------------------------------------
    //--- Construction / Destruction -----------------------------------------
    //------------------------------------------------------------------------

public:

    BaseRpc();

    /** Virtual destructor */
    //------------------------------------------------------------------------
    //--- Statistics ---------------------------------------------------------
    //------------------------------------------------------------------------

private://fields: statistics


protected://fields: overlay attributes

    // overlay identity
    NodeHandle thisNode;   //*< NodeHandle to this node */

    // overlay common parameters
    bool debugOutput;           //*< debug output ? */
    
    // references to global modules
    GlobalParameters* globalParameters;  //*< pointer to the GlobalParameters module  */
    GlobalStatistics* globalStatistics;         /**< pointer to GlobalStatistics module in this node */
            
    //------------------------------------------------------------------------
    //--- RPC Handling -------------------------------------------------------
    //------------------------------------------------------------------------

private://classes: rpc handling

    class RpcState
    {
    public:
        int id;
        int retries;
        TransportType transportType;
        CompType destComp;
        CompType srcComp;
        RpcListener* listener;
        TransportAddress dest;
        OverlayKey destKey;
        BaseCallMessage* callMsg;
        RpcTimeoutMessage* timeoutMsg;
        simtime_t timeSent;
    };

private://fields: rpc handling

    typedef hash_map<int,RpcState> RpcStates;

    int rpcsPending;
    RpcListener* defaultRpcListener;
    RpcStates rpcStates;
    double rpcUdpTimeout, rpcKeyTimeout;
    
protected://methods: rpc handling

    /**
     * Handles internal rpc requests.<br>
     * 
     * This method is used to implement basic functionionality in
     * the BaseRpc.
     * 
     * @param msg The call message
     * @return bool true, if call has been handled.
     */
    virtual bool internalHandleRpc( BaseCallMessage* msg );

    /**
    * Handles rpc responses internal in base classes<br>
    * 
    * This method is used to implement basic functionionality in
    * the BaseRpc.
    * 
    * @param msg The call message
    * @param rpcId The ID of the call
    * @param rtt the time between sending the call and receiving the response
    * @return bool true, if call has been handled.
    */
    virtual bool internalHandleRpcResponse(BaseResponseMessage* msg, int rpcId,
                                           simtime_t rtt);
    
    /**
    * Handles rpc timeouts internal in base classes<br>
    * 
    * This method is used to implement basic functionionality in
    * the BaseRpc.
    * 
    * @param msg The call message
    * @param dest The node that did not response
    * @param rpcId The ID of the call
    * @param destKey The key of the call if used
    * @return bool true, if call has been handled.
    * @todo return bool?
    */
    virtual void internalHandleRpcTimeout(BaseCallMessage* msg,
                                          const TransportAddress& dest,
                                          int rpcId, const OverlayKey& destKey);
    /**
     * Initializes Remote-Procedure state.
     */
    void initRpcs( );

    /**
     * Deinitializes Remote-Procedure state.
     */
    void finishRpcs( );

protected://methods: rpc handling

    /**
     * Handles incoming rpc messages and delegates them to the
     * corresponsing listeners or handlers.
     * 
     * @param msg The message to handle.
     */
    virtual void internalHandleRpcMessage( BaseRpcMessage* msg);

protected:

    /**
     * Sends a Remote-Procedure-Call message to the underlay.<br>
     * @deprecated Use sendRouteRpcCall(), sendInternalRpcCall(),
     * sendUdpRpcCall() instead!<br>
     *
     * Compatibility function.
     * See the other sendRpcMessage function for more information     
     * @param dest Destination node handle
     * @param msg RPC Call Message
     * @param rpcListener RPC Listener
     * @param destKey first route the RPC to the node that is
     *        responsible for kestkey
     * @param rpcId RPC id
     * @param timeout RPC timeout
     * @param retries How often we try to resent rpc call, if it gets lost
     * @return The nonce of the RPC
     */
    uint32_t sendRpcMessage(const TransportAddress& dest,
                            BaseCallMessage* msg,
                            RpcListener* rpcListener = NULL,
                            const OverlayKey& destKey
                            = OverlayKey::UNSPECIFIED_KEY,
                            int rpcId = -1, simtime_t timeout = -1,
                            int retries = 0) __attribute__ ((deprecated));
    
    /**
     * Sends a Remote-Procedure-Call message to the underlay.<br>
     * @deprecated Use sendRouteRpcCall(), sendInternalRpcCall(),
     * sendUdpRpcCall() instead!<br>
     * 
     * If no timeout is provided, a default value of
     * globalParameters.rpcUdpTimeout for
     * underlay and globalParameters.rpcKeyTimeout for a overlay rpc
     * is used. After a timeout the message gets retransmitted for
     * at maximum retries times. The destKey
     * attribute is kept untouched. 
     * 
     * @param destType The type of destination (RPC_TO_UDP, RPC_TO_KEY, RPC_TO_LOWERTIER, RPC_TO_UPPERTIER)
     * @param dest Destination node handle
     * @param msg RPC Call Message
     * @param rpcListener RPC Listener
     * @param destKey first route the RPC to the node that is
     *        responsible for kestkey
     * @param rpcId RPC id
     * @param timeout RPC timeout
     * @param retries How often we try to resent rpc call, if it gets lost
     * @return The nonce of the RPC
     */
    uint32_t sendRpcMessage(RpcAddrType destType, const TransportAddress& dest,
                            BaseCallMessage* msg,
                            RpcListener* rpcListener = NULL,
                            const OverlayKey& destKey
                                = OverlayKey::UNSPECIFIED_KEY,
                            int rpcId = -1, simtime_t timeout = -1,
                            int retries = 0 ) __attribute__ ((deprecated));

private:
    
    /**
     * Sends a Remote-Procedure-Call message to the underlay.<br>
     * USE ONE OF THE WRAPPER FUNCTIONS INSTEAD: <br>
     * sendRouteRpcCall(), sendInternalRpcCall(), or sendUdpRpcCall()
     * 
     * If no timeout is provided, a default value of
     * globalParameters.rpcUdpTimeout for
     * underlay and globalParameters.rpcKeyTimeout for a overlay rpc
     * is used. After a timeout the message gets retransmitted for
     * at maximum retries times. The destKey
     * attribute is kept untouched. 
     * 
     * @param transportType The type of transport
     * @param destComp The destination component
     * @param dest Destination node handle
     * @param destKey route the RPC to the node that is
     *        responsible for destkey
     * @param msg RPC Call Message
     * @param rpcListener RPC Listener     
     * @param timeout RPC timeout
     * @param retries How often we try to resent rpc call, if it gets lost
     * @param rpcId RPC id
     * @param rpcListener RPC Listener
     * @return The nonce of the RPC
     */
    uint32_t sendRpcCall(TransportType transportType,
                         CompType destComp,
                         const TransportAddress& dest,
                         const OverlayKey& destKey,
                         BaseCallMessage* msg,
                         simtime_t timeout, int retries,
                         int rpcId, RpcListener* rpcListener);

protected:
    
    /**
     * Routes a Remote-Procedure-Call message to an OverlayKey.<br> 
     * 
     * If no timeout is provided, a default value of
     * globalParameters.rpcUdpTimeout for
     * underlay and globalParameters.rpcKeyTimeout for a overlay rpc
     * is used. After a timeout the message gets retransmitted for
     * at maximum retries times. The destKey
     * attribute is kept untouched. 
     * 
     * @param destComp The destination component
     * @param dest Destination node handle (if specified, used as first hop)
     * @param destKey Destination OverlayKey (if unspecified, the message will
     *                be sent to dest using the overlay's UDP port)
     * @param msg RPC Call Message
     * @param timeout RPC timeout
     * @param retries How often we try to resent rpc call, if it gets lost
     * @param rpcId RPC id
     * @param rpcListener RPC Listener
     * @return The nonce of the RPC
     */
    inline uint32_t sendRouteRpcCall(CompType destComp,
                                     const TransportAddress& dest,
                                     const OverlayKey& destKey,
                                     BaseCallMessage* msg,
                                     simtime_t timeout = -1,
                                     int retries = 0,
                                     int rpcId = -1,
                                     RpcListener* rpcListener = NULL)
    {
        if (dest.isUnspecified() && destKey.isUnspecified())
            opp_error("BaseRpc::sendRouteRpcCall() with both key and "
                      "transportAddress unspecified!"); 
        return sendRpcCall(ROUTE_TRANSPORT, destComp, dest, destKey, msg,
                           timeout, retries, rpcId, rpcListener);
    }

    /**
     * Routes a Remote-Procedure-Call message to an OverlayKey.<br> 
     * 
     * If no timeout is provided, a default value of
     * globalParameters.rpcUdpTimeout for
     * underlay and globalParameters.rpcKeyTimeout for a overlay rpc
     * is used. After a timeout the message gets retransmitted for
     * at maximum retries times. The destKey
     * attribute is kept untouched. 
     * 
     * @param destComp The destination component
     * @param destKey Destination OverlayKey
     * @param msg RPC Call Message
     * @param timeout RPC timeout
     * @param retries How often we try to resent rpc call, if it gets lost
     * @param rpcId RPC id
     * @param rpcListener RPC Listener
     * @return The nonce of the RPC
     */
    inline uint32_t sendRouteRpcCall(CompType destComp,
                                     const OverlayKey& destKey,
                                     BaseCallMessage* msg,
                                     simtime_t timeout = -1,
                                     int retries = 0,
                                     int rpcId = -1,
                                     RpcListener* rpcListener = NULL)
    {
        return sendRpcCall(ROUTE_TRANSPORT, destComp,
                           TransportAddress::UNSPECIFIED_NODE,
                           destKey, msg, timeout, retries, rpcId,
                           rpcListener);
    }

    /**
     * Sends a Remote-Procedure-Call message using the overlay's UDP port<br> 
     * This replaces ROUTE_DIRECT calls!
     * 
     * If no timeout is provided, a default value of
     * globalParameters.rpcUdpTimeout for
     * underlay and globalParameters.rpcKeyTimeout for a overlay rpc
     * is used. After a timeout the message gets retransmitted for
     * at maximum retries times. The destKey
     * attribute is kept untouched. 
     * 
     * @param destComp The destination component
     * @param dest Destination node handle
     * @param msg RPC Call Message
     * @param timeout RPC timeout
     * @param retries How often we try to resent rpc call, if it gets lost
     * @param rpcId RPC id
     * @param rpcListener RPC Listener
     * @return The nonce of the RPC
     */
    inline uint32_t sendRouteRpcCall(CompType destComp,
                                     const TransportAddress& dest,
                                     BaseCallMessage* msg,
                                     simtime_t timeout = -1,
                                     int retries = 0,
                                     int rpcId = -1,
                                     RpcListener* rpcListener = NULL)
    {
        return sendRpcCall(ROUTE_TRANSPORT, destComp, dest,
                           OverlayKey::UNSPECIFIED_KEY, msg, timeout,
                           retries, rpcId, rpcListener);
    }

    /**
     * Sends a Remote-Procedure-Call message to the underlay<br> 
     * 
     * If no timeout is provided, a default value of
     * globalParameters.rpcUdpTimeout for
     * underlay and globalParameters.rpcKeyTimeout for a overlay rpc
     * is used. After a timeout the message gets retransmitted for
     * at maximum retries times. The destKey
     * attribute is kept untouched. 
     * 
     * @param dest Destination node handle
     * @param msg RPC Call Message
     * @param timeout RPC timeout
     * @param retries How often we try to resent rpc call, if it gets lost
     * @param rpcId RPC id
     * @param rpcListener RPC Listener
     * @return The nonce of the RPC
     */
    inline uint32_t sendUdpRpcCall(const TransportAddress& dest,
                                   BaseCallMessage* msg,                             
                                   simtime_t timeout = -1,
                                   int retries = 0, int rpcId = -1,
                                   RpcListener* rpcListener = NULL)
    {
        return sendRpcCall(UDP_TRANSPORT, INVALID_COMP, dest,
                           OverlayKey::UNSPECIFIED_KEY, msg,
                           timeout, retries, rpcId, rpcListener);
    }

    /**
     * Sends an internal Remote-Procedure-Call between two tiers<br> 
     * 
     * If no timeout is provided, a default value of
     * globalParameters.rpcUdpTimeout for
     * underlay and globalParameters.rpcKeyTimeout for a overlay rpc
     * is used. After a timeout the message gets retransmitted for
     * at maximum retries times. The destKey
     * attribute is kept untouched. 
     * 
     * @param destComp Destination component
     * @param msg RPC Call Message
     * @param timeout RPC timeout
     * @param retries How often we try to resent rpc call, if it gets lost
     * @param rpcId RPC id
     * @param rpcListener RPC Listener
     * @return The nonce of the RPC
     */
    inline uint32_t sendInternalRpcCall(CompType destComp,
                                        BaseCallMessage* msg,
                                        simtime_t timeout = -1,
                                        int retries = 0,
                                        int rpcId = -1,
                                        RpcListener* rpcListener = NULL)
    {
        return sendRpcCall(INTERNAL_TRANSPORT, destComp,
                           TransportAddress::UNSPECIFIED_NODE,
                           OverlayKey::UNSPECIFIED_KEY, msg,
                           timeout, retries, rpcId, rpcListener);
    }

    /**
     * Cancels a Remote-Procedure-Call.
     * 
     * @param nonce The nonce of the RPC
     */
    void cancelRpcMessage(uint32_t nonce);

    /**
     * Send Remote-Procedure response message and deletes call
     * message. 
     * 
     * @param transportType The transport used for this RPC
     * @param dest The TransportAddress of the destination
     *             (hint for ROUTE_TRANSPORT)
     * @param destKey The destination key for a ROUTE_TRANSPORT
     * @param call The corresponding call message to the reponse
     * @param response The response message
     */
    void sendRpcResponse(TransportType transportType,
                         const TransportAddress &dest,
                         const OverlayKey &destKey,
                         BaseCallMessage* call,
                         BaseResponseMessage* response);
    /**
     * Send Remote-Procedure response message and deletes call
     * message.
     * 
     * @param destType The destination type used for this RPC
     * @param call The corresponding call message to the reponse
     * @param response The response message
     */    
    void sendRpcResponse(int destType, BaseCallMessage* call,
                         BaseResponseMessage* response)
                         __attribute__ ((deprecated));

    /**
     * Send Remote-Procedure response message to UDP and deletes call
     * message.
     * 
     * @param call The corresponding call message to the reponse
     * @param response The response message
     */
    void sendRpcResponse(BaseCallMessage* call,
                         BaseResponseMessage* response);
    
    /**
     * Send internal RPC response message and deletes call
     * message.<br>
     * 
     * An internal RPC response message is sent to callingTier of the
     * call message. If call or reponse is NULL no reponse will be sent.
     * This is deprecated - use sendRpcResponse(call, response) instead.
     * 
     * @param call The corresponding call message to the reponse
     * @param response The response message
     */
    void sendInternalRpcResponse(BaseCallMessage* call,
                                 BaseResponseMessage* response )
                                 __attribute__ ((deprecated)); 

    /**
     * ping a node by its TransportAddress
     * 
     * Statistics are collected by this method.
     *
     * @param dest the node to ping
     * @param timeout RPC timeout
     * @param retries how often to retry after timeout
     * @param caption special name for the ping call (instead of "PING")
     * @param rpcListener RPC Listener
     * @param rpcId RPC id
     * @param overrideCache ignore ping cache
     */
    void pingNode(const TransportAddress& dest,
                  simtime_t timeout = -1,
                  int retries = 0,
                  const char* caption = NULL,
                  RpcListener* rpcListener = NULL,
                  int rpcId = -1,
                  bool overrideCache = false);

protected:
    /**
     * Processes Remote-Procedure-Call invokation messages.<br>
     * 
     * This method should be overloaded when the overlay provides
     * RPC functionality.
     * 
     * @return true, if rpc has been handled
     */
    virtual bool handleRpc( BaseCallMessage* msg );

    /**
     * Return the component type of this module
     * 
     * This method is overloaded by BaseOverlay/BaseApp and returns
     * the appropriate component type of this module.
     * 
     * @return the component type of this module
     */
    virtual CompType getThisCompType() = 0;

protected://methods: rpc handling

    int numPingSent; 
    int bytesPingSent;
    int numPingResponseSent;
    int bytesPingResponseSent;

    virtual void sendRpcMessageWithTransport(TransportType transportType,
                                             CompType destComp,
                                             const TransportAddress& dest,
                                             const OverlayKey& destKey,
                                             BaseRpcMessage* message) = 0;
    
    void pingRpc( PingCall* call );
    PingResponse* ping( PingCall* call );
    void pingRpcResponse(PingResponse* response, int rpcId, simtime_t rtt);
    PingCall* createPingCall( const char* caption );
    
    virtual void pingResponse(PingResponse* pingResponse, int rpcId, simtime_t rtt);
    virtual void pingTimeout(PingResponse* pingResponse, int rpcId);
    
    PingCache *pingCache;
};

#endif
