// Copyright (C) 2008 Institut fuer Telematik, Universitaet Karlsruhe (TH)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//

/**
 * @file TopologyVis.cc
 * @author Bernhard Heep
 */

#include <omnetpp.h>

#include <NodeHandle.h>
#include <BootstrapOracle.h>
#include <BootstrapOracleAccess.h>

#include "TopologyVis.h"


TopologyVis::TopologyVis()
{
    thisTerminal = NULL;
    bootstrapOracle = NULL;
}

void TopologyVis::initVis(cModule* terminal)
{
    thisTerminal = terminal;
    bootstrapOracle = BootstrapOracleAccess().get();

    // set up arrow-gates
    thisOutGateArray = thisTerminal->gate("overlayNeighborArrowOut");
    thisInGateArray = thisTerminal->gate("overlayNeighborArrowIn");
    thisTerminal->setGateSize("overlayNeighborArrowOut", 1);
    thisTerminal->setGateSize("overlayNeighborArrowIn", 1);

}

void TopologyVis::showOverlayNeighborArrow(const NodeHandle& neighbor,
                                           bool flush, char* displayString)
{
    if (!ev.isGUI() || !thisTerminal)
        return;

    char red[] = "o=red,1";
    if (displayString == NULL)
        displayString = red;

    cModule* neighborTerminal;
    cGate* neighborInGateArray;

    // flush
    if (flush) {
        for (int l = 0; l < thisOutGateArray->size(); l++) {
            cGate* tempGate =
                thisTerminal->gate("overlayNeighborArrowOut", l)->toGate();

            thisTerminal->gate("overlayNeighborArrowOut", l)->disconnect();
            if (tempGate != NULL)
                compactGateArray(tempGate->ownerModule(), IN);
        }
        thisTerminal->setGateSize("overlayNeighborArrowOut" ,0);
    }

    if (bootstrapOracle->getPeerInfo(neighbor) == NULL)
        return;

    neighborTerminal
        = simulation.module(bootstrapOracle
                            ->getPeerInfo(neighbor)->getModuleID());
    if (neighborTerminal != NULL) {
        neighborInGateArray = neighborTerminal->gate("overlayNeighborArrowIn");
    } else
        return;

    if (thisTerminal == neighborTerminal)
        return;

    //do not draw double
    for (int i = 0; i < thisOutGateArray->size(); i++)
        if (thisTerminal->gate("overlayNeighborArrowOut", i)
                ->toGate() != NULL &&
                neighborTerminal == thisTerminal
                    ->gate("overlayNeighborArrowOut", i)
                    ->toGate()->ownerModule())
            return;

    // IN
    int i = 0;
    if (neighborInGateArray->size() == 0) {
        neighborTerminal->setGateSize("overlayNeighborArrowIn", 1);
    } else {
        for (i = 0; i < neighborInGateArray->size() - 1; i++) {
            if (!(neighborTerminal->gate("overlayNeighborArrowIn", i)
                    ->isConnectedOutside()))
                break;
        }
        if (neighborTerminal->gate("overlayNeighborArrowIn", i)
                ->isConnectedOutside()) {
            neighborTerminal->setGateSize("overlayNeighborArrowIn", i + 2);
            i++;
        }
    }

    // OUT
    int j = 0;
    if (thisOutGateArray->size() == 0)
        thisTerminal->setGateSize("overlayNeighborArrowOut", 1);
    else {
        for (j = 0; j < (thisOutGateArray->size() - 1); j++) {
            if (!(thisTerminal->gate("overlayNeighborArrowOut", j)
                    ->isConnectedOutside()))
                break;
        }
        if (thisTerminal->gate("overlayNeighborArrowOut", j)
                ->isConnectedOutside()) {
            thisTerminal->setGateSize("overlayNeighborArrowOut", j + 2);
            j++;
        }
    }

    thisTerminal->gate("overlayNeighborArrowOut", j)->
    connectTo(neighborTerminal->gate("overlayNeighborArrowIn", i));

    thisTerminal->gate("overlayNeighborArrowOut", j)->
    setDisplayString(displayString);
}

void TopologyVis::deleteOverlayNeighborArrow(const NodeHandle& neighbor)
{
    if (!ev.isGUI() || !thisTerminal)
        return;

    cModule* neighborTerminal
        = simulation.module(bootstrapOracle
                            ->getPeerInfo(neighbor)->getModuleID());

    if (neighborTerminal == NULL)
        return;

    //find gate
    bool compactOut = false;
    bool compactIn = false;
    for (int i = 0; i < thisOutGateArray->size(); i++) {
        // NULL-Gate?
        if (thisTerminal->gate("overlayNeighborArrowOut", i)
                ->toGate() == NULL) {
            compactOut = true;
            continue;
        }

        if (thisTerminal->gate("overlayNeighborArrowOut", i)
                ->toGate()->ownerModule()->id() == neighborTerminal->id()) {
            thisTerminal->gate("overlayNeighborArrowOut", i)->disconnect();
            compactOut = true;
            compactIn = true;
        }
    }

    //compact OUT-array
    if (compactOut)
        compactGateArray(thisTerminal, OUT);
    //compact IN-array
    if (compactIn)
        compactGateArray(neighborTerminal, IN);
}

void TopologyVis::compactGateArray(cModule* terminal,
                                   enum VisDrawDirection dir)
{
    cGate* gateArray = (dir == OUT ? terminal->gate("overlayNeighborArrowOut")
                                   : terminal->gate("overlayNeighborArrowIn"));
    const char* gateName = (dir == OUT ? "overlayNeighborArrowOut"
                                         : "overlayNeighborArrowIn");

    for (int j = 0; j < gateArray->size() - 1; j++) {
        if (terminal->gate(gateName, j)->isConnectedOutside())
            continue;

        cGate* tempGate = NULL;
        int k = 1;
        while ((tempGate == NULL) && ((j + k) != gateArray->size())) {
            tempGate = (dir == OUT ?
                        terminal->gate(gateName, j + k)->toGate() :
                        terminal->gate(gateName, j + k)->fromGate());
            k++;
        }

        if (tempGate == NULL)
            break;

        if (dir == OUT) {
            terminal->gate(gateName, j + k - 1)->disconnect();
            terminal->gate(gateName, j)->connectTo(tempGate);
        } else {
            tempGate->disconnect();
            tempGate->connectTo(terminal->gate(gateName, j));
        }
    }

    int nullGates = 0;
    for (int j = 0; j < gateArray->size(); j++)
        if (!terminal->gate(gateName, j)->isConnectedOutside())
            nullGates++;

    terminal->setGateSize(gateName, gateArray->size() - nullGates);
}
