//
// Copyright (C) 2006 Institut fuer Telematik, Universitaet Karlsruhe (TH)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//

/**
 * @file TransportAddress.h
 * @author Markus Mauch
 * @author Sebastian Mies
 */

#ifndef __TRANSPORTADDRESS_H_
#define __TRANSPORTADDRESS_H_

//class TransportAddress;

#include <stdint.h>
#include <omnetpp.h>
#include <IPvXAddress.h>

#include <ext/hash_set>
#include <ext/hash_map>

using namespace __gnu_cxx;
using namespace std;

/**
 * This class implements a common transport address.<br>
 *
 * It covers the complete node information, like IP-Address,
 * and port. The information can be sparse, so parts can be
 * omited by setting the property to an unspecified value.
 *
 * @author Markus Mauch
 * @author Sebastian Mies
 */
class TransportAddress
{
public:
    
    /**
     * defines a hash function for TransportAddress
     */ 
    class hashFcn
    {
    public:
        size_t operator()( const TransportAddress& h1 ) const
        {
            return h1.hash();
        }
    };

public://collection typedefs

    typedef hash_set<TransportAddress, hashFcn> Set; /**< a hashed set of TransportAddresses */

public://fields

    IPvXAddress ip; /**< the ip of this TransportAddress object */
    int port; /**< the port of this TransportAddress object */

public://construction

    /**
     * Constructs a unspecified TransportAddress
     */
    TransportAddress( );

    /**
     * Standard destructor
     */
    virtual ~TransportAddress( ) {};

    /**
     * Copy constructor.
     *
     * @param handle The TransportAddress to copy
     */
    TransportAddress( const TransportAddress& handle );

    /**
     * Complete constructor.
     *
     * @param ip The IPvXAddress
     * @param port The UDP-Port
     */
    TransportAddress(const IPvXAddress& ip, int port = -1);

public://static fields

    static const TransportAddress UNSPECIFIED_NODE; /**< TransportAddress without specified ip and port */

public://methods: delegates to OverlayKey and IPvXAddress

    /**
     * compares this to a given TransportAddress
     *
     * @param rhs the TransportAddress this is compared to
     * @return true if both IPvXAddress and port are equal, false otherwise
     */
    bool operator==(const TransportAddress& rhs) const;
    
    /**
     * compares this to a given TransportAddress
     *
     * @param rhs the TransportAddress this is compared to
     * @return true if both IPvXAddress and port are not equal, false otherwise
     */
    bool operator!=(const TransportAddress& rhs) const;
    
    /**
     * compares this to a given TransportAddress
     *
     * @param rhs the TransportAddress this is compared to
     * @return true if this->ip is smaller than rhs->ip, false otherwise
     */
    bool operator< (const TransportAddress& rhs) const;

    /**
     * compares this to a given TransportAddress
     *
     * @param rhs the TransportAddress this is compared to
     * @return true if this->ip is greater than rhs->ip, false otherwise
     */
    bool operator> (const TransportAddress& rhs) const;
    
    /**
     * compares this to a given TransportAddress
     *
     * @param rhs the TransportAddress this is compared to
     * @return true if this->ip is smaller than or equal to rhs->ip, false otherwise
     */
    bool operator<=(const TransportAddress& rhs) const;
    
    /**
     * compares this to a given TransportAddress
     *
     * @param rhs the TransportAddress this is compared to
     * @return true if this->ip is greater than or equal to rhs->ip, false otherwise
     */    
    bool operator>=(const TransportAddress& rhs) const;

public://methods: operators

    /**
     * assigns ip and port of rhs to this->ip and this->port
     * 
     * @param rhs the TransportAddress with the defined ip and port
     * @return this TransportAddress object
     */
    TransportAddress& operator=(const TransportAddress& rhs);

public://methods: setters and getters

    /**
     * sets this->ip to ip and port to -1 if unspecified
     * 
     * @param ip the new IPvXAddress
     * @param port the new port
     * */
    void setAddress( const IPvXAddress& ip, int port = -1 );
    
    /**
     * sets this->port to the given port
     * 
     * @param port the new port
     * */
    void setPort( int port );

    /**
     * returns ip address
     * 
     * @return this->ip
     */
    const IPvXAddress& getAddress() const;
    
    /**
     * returns port
     * 
     * @return this->port
     */
    int getPort() const;

    /**
     * indicates if TransportAddress is specified
     * 
     * @return true, if TransportAddress has no ip or port specified, false otherwise
     */
    bool isUnspecified() const;

public://methods: hashing

    /**
     * creates a hash value of ip and port
     * 
     * @return the hash value
     */
    size_t hash() const;

public://methods: c++ streaming

    /**
     * standard output stream for TransportAddress,
     * gives out ip and port 
     * 
     * @param os the ostream
     * @param n the TransportAddress
     * @return the output stream
     */
    friend std::ostream& operator<<(std::ostream& os, const TransportAddress& n);

private://methods:

    /**
     * throws an opp_error if this or handle is unspecified
     * 
     * @param handle the TransportAddress to check
     */
    void assertUnspecified( const TransportAddress& handle ) const;

public:

    /**
     * serializes the object into a buffer
     * 
     * @param b the buffer
     */
    virtual void netPack(cCommBuffer *b);
    
    /**
     * deserializes the object from a buffer
     * 
     * @param b the buffer
     */
    virtual void netUnpack(cCommBuffer *b);
};

/**
 * netPack for TransportAddress
 * 
 * @param b the buffer
 * @param obj the TransportAddress to serialise
 */ 
inline void doPacking(cCommBuffer *b, TransportAddress& obj) {obj.netPack(b);}

/**
 * netUnpack for TransportAddress
 * 
 * @param b the buffer
 * @param obj the TransportAddress to unserialise
 */
inline void doUnpacking(cCommBuffer *b, TransportAddress& obj) {obj.netUnpack(b);}

#endif
