//
// Copyright (C) 2007 Institut fuer Telematik, Universitaet Karlsruhe (TH)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//

/**
 * @file Broose.cc
 * @author Jochen Schenk
 */

#include "Broose.h"
#include <RpcMacros.h>
#include <IPAddressResolver.h>

using namespace std;

Define_Module(Broose);

void Broose::initializeOverlay(int stage)
{
    // because of IPAddressResolver, we need to wait until interfaces
    // are registered, address auto-assignment takes place etc.
    if(stage != MIN_STAGE_OVERLAY)
        return;

    // fetch some parameters
    bucketSize = par("bucketSize"); // = k
    rBucketSize = par("rBucketSize"); // = k'
    parallelRequests = par("parallelRequests"); // not implemented yet
    joinDelay = par("joinDelay");
    shiftingBits = par("shiftingBits");
    pingDelay = par("pingDelay");
    userDist = par("userDist");
    refreshTime = par("refreshTime");
    numberRetries = par("numberRetries");

    //statistics
    bucketCount = 0;
    bucketBytesSent = 0;
    broosePingCount = 0;
    broosePingBytesSent = 0;

    //init local parameters
    chooseLookup = 0;
    receivedJoinResponse = 0;
    receivedBBucketLookup = 0;
    numberBBucketLookup = 0;
    receivedLBucketLookup = 0;
    numberLBucketLookup = 0;
    powShiftingBits = (int) pow(2.0, shiftingBits);
    keyLength = OverlayKey::getLength();
    numFailedPackets = 0;
    refresh = false;
    numPings = 0;
    maxPings = 0;

    // add some watches
    WATCH(receivedJoinResponse);
    WATCH(receivedBBucketLookup);
    WATCH(numberBBucketLookup);
    WATCH(receivedLBucketLookup);
    WATCH(numberLBucketLookup);
    WATCH(numPings);
    WATCH(maxPings);

    // get module references - these are references to the buckets
    // each node have in order to get a lookup done
    rBucket = new BrooseBucket*[powShiftingBits];

    for (int i = 0; i < powShiftingBits; i++) {
        rBucket[i] = check_and_cast<BrooseBucket*>
                     (parentModule()->submodule("rBucket",i));
    }

    lBucket = check_and_cast<BrooseBucket*>
              (parentModule()->submodule("lBucket"));

    bBucket = check_and_cast<BrooseBucket*>
              (parentModule()->submodule("bBucket"));

    // create join and bucket timer
    join_timer = new cMessage("join_timer");
    bucket_timer = new cMessage("bucket_timer");
}

void Broose::joinOverlay()
{
    changeState(INIT);

    // if the bootstrap node is unspecified we are the only node in the network
    // so we can skip the "normal" join protocol
    if (bootstrapNode.isUnspecified()) {
        changeState(READY);
    }
}


void Broose::changeState(int state)
{
    switch (state) {
    case INIT: {
        protoState = INIT;

        if (!refresh) {
            // Calculate node's id by hashing its IP address
            thisNode.key = OverlayKey::sha1(
                               const_cast<char*>(thisNode.ip.str().c_str()));
            callUpdate(thisNode, true);
            thisBrooseNode = BrooseHandle(thisNode);
            updateTooltip();
        } else {
            removeBootstrapNode(thisNode);
        }

        // find a new bootstrap node and enroll to the bootstrap list
        bootstrapNode = getBootstrapNode();

        cancelEvent(join_timer);
        scheduleAt(simulation.simTime(), join_timer);

        // initialize respectively clear the buckets
        for (int i = 0; i < powShiftingBits; i++) {
            rBucket[i]->initializeBucket(shiftingBits, i,
                                         thisBrooseNode, rBucketSize, this);
        }

        lBucket->initializeBucket(-shiftingBits, 0, thisBrooseNode,
                                  powShiftingBits*rBucketSize, this);
        bBucket->initializeBucket(0, 0, thisBrooseNode, 7*bucketSize,
                                  this, true);

        // if we have restarted the join protocol reset parameters
        refresh = false;
        receivedBBucketLookup = 0;
        receivedLBucketLookup = 0;
        receivedJoinResponse = 0;

        assert(parentModule()->parentModule());
        parentModule()->parentModule()->bubble("Enter INIT state.");
        break;
    }

    case RSET: {
        protoState = RSET;

        BrooseBucket* tmpBucket = new BrooseBucket();
        tmpBucket->initializeBucket(0, 0, thisNode,
                                    powShiftingBits*rBucketSize, this);

        for (int i = 0; i < powShiftingBits; i++) {
            int size = rBucket[i]->getSize();

            for (int j = 0; j < size; j++) {
                tmpBucket->add(rBucket[i]->get(j));
            }
        }

        BucketCall** bCall = new BucketCall*[tmpBucket->getSize()];
        for (uint i = 0; i < tmpBucket->getSize(); i++) {
            bCall[i] = new BucketCall("LBucketCall");
            bCall[i]->setBucketType(LEFT);
            bCall[i]->setProState(PRSET);
            bCall[i]->setLength(BUCKETCALL_L(bcall[i]));

            sendUdpRpcCall(tmpBucket->get(i), bCall[i], numberRetries,
                           joinDelay);
        }

        // half of the calls must return to init a state change
        numberBBucketLookup = tmpBucket->getSize();
        numberBBucketLookup = (numberBBucketLookup == 1) ?
                              numberBBucketLookup : (numberBBucketLookup/2);

        delete tmpBucket;

        assert(parentModule()->parentModule());
        parentModule()->parentModule()->bubble("Enter RSET state.");
        break;
    }

    case BSET: {
        protoState = BSET;

        // half of the calls must return to init a state change
        numberLBucketLookup = bBucket->getSize();
        numberLBucketLookup = (numberLBucketLookup == 1) ?
                              numberLBucketLookup : (numberLBucketLookup/2);

        // send messages to all entries of the B Bucket
        int size2 = bBucket->getSize();
        BucketCall** bCall2 = new BucketCall*[size2];
        for (int i = 0; i < size2; i++) {
            bCall2[i] = new BucketCall("LBucketCall");
            bCall2[i]->setBucketType(LEFT);
            bCall2[i]->setProState(PBSET);
            bCall2[i]->setLength(BUCKETCALL_L(bcall2[i]));

            sendUdpRpcCall(bBucket->get(i), bCall2[i], numberRetries,
                           joinDelay);
        }

        assert(parentModule()->parentModule());
        parentModule()->parentModule()->bubble("Enter BSET state.");
        break;
    }

    case READY: {
        protoState = READY;
//        registerBootstrapNode(thisNode);

        //fill the bucket also with this node
        for (int i = 0; i < powShiftingBits; i++) {
            rBucket[i]->add(thisBrooseNode);
        }
        
        lBucket->add(thisBrooseNode);
        bBucket->add(thisBrooseNode);

        // to disable the ping protocol a pingDelay or
        // refreshTime of zero was given
        if (!(pingDelay == 0 || refreshTime == 0)) {
            cancelEvent(bucket_timer);
            scheduleAt(simulation.simTime() + pingDelay, bucket_timer);
        }

        assert(parentModule()->parentModule());
        parentModule()->parentModule()->bubble("Enter READY state.");

        updateTooltip();
        break;
    }

    }
    setOverlayReady(protoState == READY);
}

void Broose::handleUDPMessage(BaseOverlayMessage* msg)
{
    error("Broose::handleUDPMessage called!");
}

void Broose::handleTimerEvent(cMessage* msg)
{
    if(msg->isName("join_timer"))
        handleJoinTimerExpired(msg);
    else if (msg->isName("bucket_timer"))
        handleBucketTimerExpired(msg);
    else
        error("Broose::handleTimerEvent - no other timer currently in use!");
}

void Broose::handleJoinTimerExpired(cMessage* msg)
{
    if (protoState == READY)
        return;

    if (!bootstrapNode.isUnspecified()) {
        // create new lookup message
        BucketCall* bCallArray[powShiftingBits];

        // do lookups for key >> shiftingBits for each prefix
        OverlayKey newKey = thisNode.key >> shiftingBits;
        for (int i = 0; i < powShiftingBits; i++) {
            OverlayKey add(i);
            add = add << (keyLength - shiftingBits);
            add += newKey;

            bCallArray[i] = new BucketCall("BBucketCall");
            bCallArray[i]->setBucketType(BROTHER);
            bCallArray[i]->setBucketIndex(i);
            bCallArray[i]->setProState(PINIT);
            bCallArray[i]->setLength(BUCKETCALL_L(bCallArray[i]));

            // restart join protocol if one call times out
            // otherwise the node might be isolated
            sendRouteRpcCall(OVERLAY_COMP, bootstrapNode, add,
                             bCallArray[i]);
        }
    } else {
        // if the bootstrap node is unspecified we are the only node in the network
        // so we can skip the "normal" join protocol
        changeState(READY);
    }
}

void Broose::handleBucketTimerExpired(cMessage* msg)
{
    BrooseBucket* tmpBucket = new BrooseBucket();
    tmpBucket->initializeBucket (0, 0, thisBrooseNode,
                                 (2*powShiftingBits*rBucketSize + 7*bucketSize),
                                 this);

    for (int i = 0; i < powShiftingBits; i++) {
        for(uint j = 0; j < rBucket[i]->getSize(); j++) {
            if ((simulation.simTime() - rBucket[i]->getLastSeen(
                        rBucket[i]->get(j))) > refreshTime
                    || rBucket[i]->getRTT(rBucket[i]->get(j)) == -1) {

                tmpBucket->add(BrooseHandle(rBucket[i]->get(j)));
            }
        }
    }

    for (uint i = 0; i < lBucket->getSize(); i++) {
        if ((simulation.simTime() - lBucket->getLastSeen(
                    lBucket->get(i))) > refreshTime
                || lBucket->getRTT(lBucket->get(i)) == -1) {

            tmpBucket->add(BrooseHandle(lBucket->get(i)));
        }
    }

    for (uint i = 0; i < bBucket->getSize(); i++) {
        if ((simulation.simTime() - bBucket->getLastSeen(
                    bBucket->get(i))) > refreshTime
                || bBucket->getRTT(bBucket->get(i)) == -1 ) {

            tmpBucket->add(BrooseHandle(bBucket->get(i)));
        }
    }

    maxPings = tmpBucket->getSize();

    if (maxPings != 0) {
        BroosePingCall** array = new BroosePingCall*[tmpBucket->getSize()];

        for (uint i = 0; i < tmpBucket->getSize(); i++) {
            array[i] = new BroosePingCall("PingCall");
            array[i]->setLength(BROOSEPINGCALL_L(array[i]));

            sendUdpRpcCall(tmpBucket->get(i), array[i]);
        }
    } else {
        numPings = 0;
        scheduleAt(simulation.simTime() + pingDelay, bucket_timer);
    }
    delete tmpBucket;
}


int Broose::getMaxNumSiblings()
{
    return bucketSize;
}

int Broose::getMaxNumRedundantNodes()
{
    return bucketSize;
}


NodeVector* Broose::findNode(const OverlayKey& key,
                             int numRedundantNodes,
                             int numSiblings,
                             BaseOverlayMessage* msg)
{
    NodeVector* nextHop = new NodeVector(1);
    BrooseFindNodeExtMessage *findNodeExt = NULL;
    bool err;

    //return the closest nodes
    if (isSiblingFor(thisNode, key, numSiblings, &err)) {
        BrooseBucket* tmpBBucket = new BrooseBucket();
        BrooseHandle node;
        node.key = key;
        tmpBBucket->initializeBucket(0, 0, node, bucketSize, this);
        int size;

        if (bBucket->getSize() > bucketSize)
            size = bucketSize;
        else
            size = bBucket->getSize();

        for (uint i= 0; i < bBucket->getSize();i++)
            tmpBBucket->add(bBucket->get(i));

        // todo: this has to be returned, if baselookup can deal
        // with the complete vector:
        for (int i = 0; i < size; i++)
            nextHop->push_back(tmpBBucket->get(i));

//        nextHop->push_back(tmpBBucket->get(0));

        delete tmpBBucket;
        return nextHop;
    }

    if (msg != NULL) {
        if (!msg->hasObject("findNodeExt")) {
            findNodeExt = new BrooseFindNodeExtMessage("findNodeExt");

            OverlayKey routeKey = thisNode.key;
            // estimate distance
            int dist = max(rBucket[0]->longestPrefix(),
                           rBucket[1]->longestPrefix()) + 1 + userDist;

            if ((dist % shiftingBits) != 0)
                dist += (shiftingBits - (dist % shiftingBits));

            if (dist > keyLength) {
                if ((keyLength % shiftingBits) == 0) {
                    dist = keyLength;
                } else {
                    dist = (keyLength - keyLength % shiftingBits);
                }
            }

            if ((chooseLookup++) % 2 == 0) {
                // init left shifting lookup
                findNodeExt->setRightShifting(false);

                int prefix = 0;
                for (int i = 0; i < dist; i++) {
/*                    prefix += thisNode.key.bitAtPlace(i+1) 
                            (int)pow(2.0, dist - i - 1); */
                    prefix += thisNode.key.get(thisNode.key.getLength() - i - 1, 1) << (dist - i - 1);
                }

                OverlayKey pre(prefix);
                routeKey = key >> dist;
                routeKey += (pre << key.getLength() - dist);

                dist = -dist;
            } else {
                // init right shifting lookup
                findNodeExt->setRightShifting(true);
            }

            //add contact for next Hop
            findNodeExt->setLastNode(thisNode);
            findNodeExt->setRouteKey(routeKey);
            findNodeExt->setStep(dist);
            findNodeExt->setLength(BROOSEFINDNODEEXTMESSAGE_L);

            msg->addObject( findNodeExt );
        }

        findNodeExt = (BrooseFindNodeExtMessage*) msg->getObject("findNodeExt");
    }


    // check for messages which couldn't be routed
    if (findNodeExt->getStep() == 0) {
        if (!isSiblingFor(thisNode, key, numSiblings, &err)) {
            //	    cout << "Message failed - destKey " << key << "@ simtime " << simulation.simTime()
            // << " @node " << thisNode.ip << " lastNode " << findNodeExt->getLastNode().ip << endl;
            numFailedPackets++;
        } else
            error("Broose::findNode - unexpected Error");
        return nextHop;
    }

    if (findNodeExt->getRightShifting() == false) {
        // Left Shifting Lookup

        // can't handle left shifting lookup in BSET-State
        if (protoState == BSET)
            return nextHop;

        // update buckets with last hop
        addNode(findNodeExt->getLastNode());
        setLastSeen(findNodeExt->getLastNode());

        // replace last hop contact information with
        // this hop contact information
        findNodeExt->setLastNode(thisNode);


        // calculate routing key
        findNodeExt->setRouteKey((findNodeExt->getRouteKey()) << shiftingBits);
        findNodeExt->setStep(findNodeExt->getStep() + shiftingBits);

        // On last hop exchange routeKey for destKey especially
        // useful when using lookupNodeIds
        NodeHandle nextNode;

        if (findNodeExt->getStep() == 0)
            nextNode = lBucket->getClosestNode(key);
        else
            nextNode = lBucket->getClosestNode(findNodeExt->getRouteKey());

        nextHop->push_back(nextNode);
    } else {
        // Right Shifting Lookup

        // update buckets with last hop
        addNode(findNodeExt->getLastNode());
        setLastSeen(findNodeExt->getLastNode());

        // replace last hop contact information with
        // this hop contact information
        findNodeExt->setLastNode(thisNode);

        // calculate routing key
        int prefix = 0;
        int dist = findNodeExt->getStep();
        OverlayKey routeKey = findNodeExt->getRouteKey() >> shiftingBits;
        for (int i = 0; i < shiftingBits; i++)
//             prefix += key.bitAtPlace(dist-i) * (int) pow (2.0, i);
            prefix += key.get(key.getLength() - dist + i, 1) << i;
        OverlayKey pre(prefix);
        routeKey += (pre << (routeKey.getLength()-shiftingBits));

        findNodeExt->setRouteKey(routeKey);
        findNodeExt->setStep(dist - shiftingBits);

        // On last hop exchange routeKey for destKey especially
        // useful when using lookupNodeIds
        NodeHandle nextNode;

        if (findNodeExt->getStep() == 0)
            nextNode = rBucket[prefix]->getClosestNode(key);
        else
            nextNode = rBucket[prefix]->getClosestNode(routeKey);

        nextHop->push_back(nextNode);
    }

    if ((*nextHop)[0] == thisNode) {
        return (findNode(key, numRedundantNodes, numSiblings, msg));
    } else
        return nextHop;
}

void Broose::finishOverlay()
{
    // print out statistics and delete pointers

    // delete timer
    cancelEvent(join_timer);
    delete join_timer;

    cancelEvent(bucket_timer);
    delete bucket_timer;

    // remove this node from the bootstrap list
    removeBootstrapNode(thisNode);

    simtime_t time = globalStatistics->calcMeasuredLifetime(creationTime);
    if(time == 0) return;

    globalStatistics->addStdDev("Number of non-routable packets/s", numFailedPackets / time);

    globalStatistics->addStdDev("Broose: Sent BUCKET Messages/s", bucketCount / time);
    globalStatistics->addStdDev("Broose: Sent BUCKET Byte/s", bucketBytesSent / time);

    globalStatistics->addStdDev("Broose: Sent BROOSEPING Messages/s", broosePingCount / time);
    globalStatistics->addStdDev("Broose: Sent BROOSEPING Bytes/s", broosePingBytesSent / time);
}

void Broose::recordOverlaySentStats(BaseOverlayMessage* msg)
{
    BaseOverlayMessage* innerMsg;

    if (msg->getType() == OVERLAYROUTE)
        innerMsg = dynamic_cast<BaseOverlayMessage*>(msg->encapsulatedMsg());
    else
        innerMsg = msg;

    switch (innerMsg->getType()) {

    case RPC: {
        if ((dynamic_cast<BucketCall*>(innerMsg) != NULL) ||
                (dynamic_cast<BucketResponse*>(innerMsg) != NULL)) {
            RECORD_STATS(bucketCount++; bucketBytesSent +=
                             msg->byteLength());
        } else if ((dynamic_cast<BroosePingCall*>(innerMsg) != NULL) ||
                   (dynamic_cast<BroosePingResponse*>(innerMsg) != NULL)) {
            RECORD_STATS(broosePingCount++; broosePingBytesSent +=
                             msg->byteLength());
        }
        break;
    }
    }
}

void Broose::displayBucketState()
{
    EV << "[Broose::displayBucketState() @ " << thisNode.ip
    << " (" << thisNode.key.toString(16) << ")]" << endl;
    for (int i = 0; i < powShiftingBits; i++) {
        EV << "    Content of rBucket[" << i << "]: ";
        rBucket[i]->output();
    }
    EV << "    Content of lBucket: ";
    lBucket->output();
    EV << "    Content of bBucket: ";
    bBucket->output();
    EV << endl;
}


bool Broose::isSiblingFor(const NodeHandle& node,
                          const OverlayKey& key,
                          int numSiblings,
                          bool* err)
{
// TODO: node != thisNode doesn't work yet
    if (key.isUnspecified())
        error("Broose::isSiblingFor(): key is unspecified!");

    if (node != thisNode)
        error("Broose::isSiblingsFor(): "
              "node != thisNode is not implemented!");

    if (numSiblings > getMaxNumSiblings()) {
        opp_error("Broose::isSiblingFor(): numSiblings too big!");
    }
    // set default number of siblings to consider
    if (numSiblings == -1) numSiblings = getMaxNumSiblings();

    if (protoState != READY) {
        *err = true;
        return false;
    }

    return bBucket->keyInRange(key);
}

void Broose::updateTooltip()
{
    if (ev.isGUI()) {
        std::stringstream ttString;

        // show our ip and key in tooltip
        ttString << thisNode.ip << " " << thisNode.key;

        parentModule()->parentModule()->displayString().
        setTagArg("tt", 0, ttString.str().c_str());
        parentModule()->displayString().
        setTagArg("tt", 0, ttString.str().c_str());
        displayString().setTagArg("tt", 0, ttString.str().c_str());

    }
}

void Broose::binaryOutput(const OverlayKey& key)
{
    EV << "[Broose::binaryOutput() @ " << thisNode.ip;
    if (key.isUnspecified()) {
        EV << " (<unspec>)]"
        << endl;
    } else {
        EV << " (";
        for (unsigned int i = 1; i <= key.getLength(); i++) {
/*            EV << key.bitAtPlace(i);*/
            EV << key.get(key.getLength() - i, 1);
        }
        EV << ")]" << endl;
    }
}

bool Broose::handleRpc(BaseCallMessage* msg)
{
    if (protoState == BSET || protoState == READY) {
        // delegate messages
        RPC_SWITCH_START( msg )
        // RPC_DELEGATE( <messageName>[Call|Response], <methodToCall> )
        RPC_DELEGATE(Bucket, handleBucketRequestRpc);
        RPC_DELEGATE(BroosePing, handleBroosePingRequestRpc);
        RPC_SWITCH_END( )
        return RPC_HANDLED;
    } else {
//        EV << "[Broose::handleRpc() @ " << thisNode.ip
//           << " (" << thisNode.key.toString(16) << ")]\n"
//           << "    Received RPC call and state != READY || BSET!"
//           << endl;
        return false;
    }
}

void Broose::handleRpcResponse(BaseResponseMessage* msg, int rpcId,
                               simtime_t rtt)
{
    RPC_SWITCH_START(msg)
    RPC_ON_RESPONSE( BroosePing ) {
        handleBroosePingResponseRpc(_BroosePingResponse, rtt);
        EV << "[Broose::handleRpcResponse() @ " << thisNode.ip
        << " (" << thisNode.key.toString(16) << ")]\n"
        << "    Ping RPC Response received: id=" << rpcId << "\n"
        << "    msg=" << *_BroosePingResponse << " rtt=" << rtt
        << endl;
        break;
    }
    RPC_ON_RESPONSE( Bucket ) {
        handleBucketResponseRpc(_BucketResponse);
        EV << "[Broose::handleRpcResponse() @ " << thisNode.ip
        << " (" << thisNode.key.toString(16) << ")]\n"
        << "    Bucket RPC Response received: id=" << rpcId << "\n"
        << "    msg=" << *_BucketResponse << " rtt=" << rtt
        << endl;
        break;
    }
    RPC_SWITCH_END( )
}

void Broose::handleRpcTimeout(BaseCallMessage* msg, const TransportAddress& dest,
                              int rpcId, const OverlayKey& destKey)
{
    RPC_SWITCH_START(msg)
    RPC_ON_CALL( BroosePing ) {
        handleBroosePingTimeout(_BroosePingCall, dest, destKey);
        EV << "[Broose::handleRpcTimeout() @ " << thisNode.ip
        << " (" << thisNode.key.toString(16) << ")]\n"
        << "    Ping RPC Call timed out: id=" << rpcId << "\n"
        << "    msg=" << *_BroosePingCall
        << endl;
        break;
    }
    RPC_ON_CALL( FindNode ) {
        handleFindNodeTimeout(_FindNodeCall, dest, destKey);
        EV << "[Broose::handleRpcTimeout() @ " << thisNode.ip
        << " (" << thisNode.key.toString(16) << ")]\n"
        << "    Find Node RPC Call timed out: id=" << rpcId << "\n"
        << "    msg=" << *_FindNodeCall
        << endl;
        break;
    }
    RPC_ON_CALL( Bucket ) {
        handleBucketTimeout(_BucketCall);
        EV << "[Broose::handleRpcTimeout() @ " << thisNode.ip
        << " (" << thisNode.key.toString(16) << ")]\n"
        << "    Bucket RPC Call timed out: id=" << rpcId << "\n"
        << "    msg=" << *_BucketCall
        << endl;
        break;
    }
    RPC_SWITCH_END( )
}


void Broose::handleFindNodeTimeout(FindNodeCall* findNode,
                                   const TransportAddress& dest,
                                   const OverlayKey& destKey)
{
    for (int i = 0; i < powShiftingBits; i++) {
        if (rBucket[i]->getFailedResponses(BrooseHandle(dest, destKey)) == numberRetries)
            rBucket[i]->remove
            (BrooseHandle(dest, destKey));
        else
            rBucket[i]->increaseFailedResponses(BrooseHandle(dest, destKey));
    }

    if (lBucket->getFailedResponses(BrooseHandle(dest, destKey)) == numberRetries)
        lBucket->remove
        (BrooseHandle(dest, destKey));
    else
        lBucket->increaseFailedResponses(BrooseHandle(dest, destKey));

    if (bBucket->getFailedResponses(BrooseHandle(dest, destKey)) == numberRetries)
        bBucket->remove
        (BrooseHandle(dest, destKey));
    else
        bBucket->increaseFailedResponses(BrooseHandle(dest, destKey));
}
void Broose::handleBucketRequestRpc(BucketCall* msg)
{
    if (msg->getBucketType() == LEFT) {
        // can't handle LBucketRequest in BSET-State
        if (protoState == BSET) {
            delete msg;
            return;
        }

        // return L-Bucket
        int size = lBucket->getSize();
        BucketResponse* bResponse = new BucketResponse("LBucketResponse");
        bResponse->setBucketType(msg->getBucketType());
        bResponse->setProState(msg->getProState());
        bResponse->setNodeNum(size);
        bResponse->setNodesArraySize(size);

        for (int i = 0; i < size; i++) {
            bResponse->setNodes(i, lBucket->get(i));
        }

        bResponse->setLength(BUCKETRESPONSE_L(bResponse));

        addNode(msg->getSrcNode());
        setLastSeen(msg->getSrcNode());

        sendRpcResponse(msg, bResponse );
    } else if(msg->getBucketType() == BROTHER) {
        // return B-Bucket
        int size = bBucket->getSize();
        BucketResponse* bResponse = new BucketResponse("BBucketResponse");
        bResponse->setBucketType(msg->getBucketType());
        bResponse->setBucketIndex(msg->getBucketIndex());
        bResponse->setProState(msg->getProState());
        bResponse->setNodeNum(size);
        bResponse->setNodesArraySize(size);

        for (int i = 0; i < size; i++) {
            bResponse->setNodes(i, bBucket->get(i));
        }
        bResponse->setLength(BUCKETRESPONSE_L(bResponse));

        sendRpcResponse(msg, bResponse );
    } else
        error("Broose::handleBucketRequestRpc() - Wrong Bucket Type!");
}

void Broose::handleBucketResponseRpc(BucketResponse* msg)
{
    if (msg->getBucketType() == LEFT) {
        switch (protoState) {
        case RSET:
            if (msg->getProState() == PRSET) {
                for (int i = 0; i < msg->getNodeNum(); i++) {
                    bBucket->add(BrooseHandle(msg->getNodes(i)));
                }
                receivedBBucketLookup++;

                if (receivedBBucketLookup == numberBBucketLookup)
                    changeState(BSET);
            }
            break;
        case BSET:
            if (msg->getProState() == PBSET) {
                for (int i = 0; i < msg->getNodeNum(); i++) {
                    lBucket->add(BrooseHandle(msg->getNodes(i)));
                }
                receivedLBucketLookup++;

                if(receivedLBucketLookup == numberLBucketLookup)
                    changeState(READY);
            }
            break;
        }
    } else if(msg->getBucketType() == BROTHER) {
        switch(protoState) {
        case INIT:
            if (msg->getProState() == PINIT) {
                int k = msg->getBucketIndex();

                for (int i = 0; i < msg->getNodeNum(); i++) {
                    rBucket[k]->add(msg->getNodes(i));
                }

                receivedJoinResponse++;
                if (receivedJoinResponse == powShiftingBits)
                    changeState(RSET);
            }
        }
    } else
        error("Broose::handleBucketRequestRpc() - unknown error.");
}


void Broose::handleBucketTimeout(BucketCall* msg)
{
    if (protoState == READY)
        return;
    else {
        refresh = true;
        changeState(INIT);
    }
}


void Broose::handleBroosePingRequestRpc(BroosePingCall* msg)
{
    BroosePingResponse* pingResponse = new BroosePingResponse("PingResponse");
    pingResponse->setLength(BROOSEPINGRESPONSE_L(pingResponse));

    // add pinging node to all buckets and update lastSeen of node
    addNode(msg->getSrcNode());
    setLastSeen(msg->getSrcNode());

    sendRpcResponse(msg, pingResponse);
}

void Broose::handleBroosePingResponseRpc(BroosePingResponse* msg,
        simtime_t rtttime)
{
    // if node respond reset failedResponses and add lastSeen to node
    setLastSeen(msg->getSrcNode());
    resetFailedResponses(msg->getSrcNode());
    setRTT(msg->getSrcNode(), rtttime);

    numPings++;

    if (numPings == maxPings) {
        numPings = 0;
        scheduleAt(simulation.simTime() + pingDelay, bucket_timer);
    }
}

void Broose::handleBroosePingTimeout(BroosePingCall* msg,
                                     const TransportAddress& dest,
                                     const OverlayKey& destKey)
{
    for (int i = 0; i < powShiftingBits; i++) {
        if (rBucket[i]->getFailedResponses(BrooseHandle(dest, destKey)) == numberRetries)
            rBucket[i]->remove(BrooseHandle(dest, destKey));
        else
            rBucket[i]->increaseFailedResponses(BrooseHandle(dest, destKey));
    }

    if (lBucket->getFailedResponses(BrooseHandle(dest, destKey)) == numberRetries) {
        lBucket->remove(BrooseHandle(dest, destKey));
    } else {
        lBucket->increaseFailedResponses(BrooseHandle(dest, destKey));
    }

    if (bBucket->getFailedResponses(BrooseHandle(dest, destKey)) == numberRetries) {
        bBucket->remove(BrooseHandle(dest, destKey));
    } else {
        bBucket->increaseFailedResponses(BrooseHandle(dest, destKey));
    }

    numPings++;

    if (numPings == maxPings) {
        numPings = 0;
        scheduleAt(simulation.simTime() + pingDelay, bucket_timer);
    }
}

void Broose::setLastSeen(NodeHandle node)
{
    for (int i = 0; i < powShiftingBits; i++) {
        rBucket[i]->setLastSeen(BrooseHandle(node), simulation.simTime());
    }

    lBucket->setLastSeen(BrooseHandle(node), simulation.simTime());
    bBucket->setLastSeen(BrooseHandle(node), simulation.simTime());
}

void Broose::addNode(NodeHandle node, int bucket)
{
    if (bucket == 0) {
        // add node to all buckets
        for (int i = 0; i < powShiftingBits; i++) {
            rBucket[i]->add
            (BrooseHandle(node));
        }

        lBucket->add(BrooseHandle(node));
        bBucket->add(BrooseHandle(node));
    } else
        error("Broose::addNode() - not implemented");
}

void Broose::resetFailedResponses(NodeHandle node)
{
    for (int i = 0; i < powShiftingBits; i++) {
        rBucket[i]->resetFailedResponses(BrooseHandle(node));
    }

    lBucket->resetFailedResponses(BrooseHandle(node));
    bBucket->resetFailedResponses(BrooseHandle(node));
}

void Broose::setRTT(NodeHandle node, simtime_t rtttime)
{
    for (int i = 0; i < powShiftingBits; i++) {
        rBucket[i]->setRTT(BrooseHandle(node), rtttime);
    }

    lBucket->setRTT(BrooseHandle(node), rtttime);
    bBucket->setRTT(BrooseHandle(node), rtttime);
}
