//
// Copyright (C) 2006 Institut fuer Telematik, Universitaet Karlsruhe (TH)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

#include "Kademlia.h"
#include "KademliaMessage_m.h"

#include <IPAddressResolver.h>
#include <IPvXAddress.h>
#include <InterfaceTable.h>
#include <IPv4InterfaceData.h>


Define_Module(Kademlia);

std::ostream& operator<<(std::ostream& os, const KademliaBucket* n)
{
    if (n == NULL)
        os << "NULL";
    else
        for (KademliaBucket::const_iterator i=n->begin(); i !=n->end(); i++) {
            os << *i;
            if (i+1 != n->end())
                os << endl;
        }
    return os;
}
;

class KademliaLookupListener : public LookupListener
{
private:
    Kademlia* overlay;
public:
    KademliaLookupListener(Kademlia* overlay)
    {
        this->overlay = overlay;
    }

    virtual void lookupFinished(AbstractLookup *lookup)
    {
        overlay->lookupFinished();
        delete this;
    }
};

//-----------------------------------------------------------------------------

void Kademlia::initializeOverlay(int stage)
{
    if (stage != MIN_STAGE_OVERLAY)
        return;

    // setup kademlia parameters
    this->stabilizeDelay = par("stabilizeDelay");
    this->removalDelay = par("removalDelay");
    this->k = par("k");
    this->b = par("b");
    this->s = par("s");

    // self-messages
    stabilizeTimer = new cMessage("stabilizeTimer");

    comparator = NULL;
    siblingTable = NULL;
}

Kademlia::~Kademlia()
{
}

void Kademlia::finishOverlay()
{
    // deinitialize routing
    routingDeinit();

    // remove comparator and timer
    cancelEvent(stabilizeTimer);
    delete comparator;
    delete stabilizeTimer;
}

void Kademlia::joinOverlay()
{
    // initialize routing
    routingDeinit();
    routingInit();

    // add bootstrap nodes
    //for (int i=0; i < k; i++) {

    NodeHandle handle = getBootstrapNode();
    if (!handle.isUnspecified()) {
        routingAdd(handle, true);
    } else {
        // as the first packet leaves the node, this node is accessible
        registerBootstrapNode(thisNode);
        state = READY;
    }
    //    }

    // schedule refresh timer
    cancelEvent(stabilizeTimer);
    scheduleAt(simulation.simTime(), stabilizeTimer);
}

//-----------------------------------------------------------------------------

void Kademlia::routingInit()
{
    // set join state
    state = JOIN;

    // setup comparator
    comparator = new KeyDistanceComparator<KeyXorMetric>( thisNode.getKey() );

    // set key length
    this->keyLength = OverlayKey::getLength();

    // calculate number of buckets: ( (2^b)-1 ) * ( keylength / b )
    this->numBuckets = ( (1L << b) - 1L ) * (keyLength / b );

    // init routing and sibling table
    this->siblingTable = new KademliaBucket ( s * 5, comparator );

    // initialize pointers
    this->routingTable.assign(this->numBuckets, (KademliaBucket*)NULL);

    WATCH_VECTOR(*siblingTable);
    WATCH_VECTOR(routingTable);

    updateTooltip();
}

void Kademlia::routingDeinit()
{
    // set refresh state (to prevent registering this node
    // at the BootstrapOracle in lookupFinished(), if routingDeinit()
    // was called, because this node gets deleted
    state = REFRESH;

    // remove current node handle from the bootstrap list
    if (!thisNode.key.isUnspecified()) {
        removeBootstrapNode(thisNode);
    }

    // delete buckets
    for (uint i=0; i<this->routingTable.size(); i++) {
        if (this->routingTable[i] != NULL) {
            delete this->routingTable[i];
            this->routingTable[i] = NULL;
        }
    }
    this->routingTable.clear();

    // clear sibling table
    if (this->siblingTable != NULL) {
        delete this->siblingTable;
        this->siblingTable = NULL;
    }

    if (comparator != NULL) {
        delete comparator;
        comparator = NULL;
    }
}

int Kademlia::getMaxNumSiblings()
{
    return s;
}

int Kademlia::getMaxNumRedundantNodes()
{
    return k;
}

int Kademlia::routingBucketIndex(const OverlayKey& key)
{
    // calculate XOR distance
    OverlayKey delta = key ^ getThisNode().getKey();

    // find first subinteger that is not zero...
    int i;
    for (i = key.getLength()-b; i>=0 && delta.get(i, b) == 0; i-=b)
        ;
    if (i<0)
        return -1;
    return (i / b) * ( (1 << b) - 1 ) + (delta.get(i, b) - 1 );
}

KademliaBucket* Kademlia::routingBucket(const OverlayKey& key, bool ensure)
{
    // get bucket index
    int num = routingBucketIndex(key);
    if (num < 0)
        return NULL;

    // get bucket and allocate if neccessary
    KademliaBucket* bucket = routingTable[ num ];
    if (bucket == NULL && ensure)
        bucket = routingTable[ num ] = new KademliaBucket( k, comparator );

    // return bucket
    return bucket;
}

bool Kademlia::routingAdd(const NodeHandle& handle, bool isAlive)
{
    // never add unspecified node handles
    if (handle.isUnspecified() || handle.key == getThisNode().getKey() )
        return false;

    // bucket index
    KademliaBucket::iterator i;
    bool result = false;

    // convert node handle
    KademliaNodeHandle kadHandle = handle;
    kadHandle.setLastSeen(simulation.simTime() );

    /* check if node is already a sibling -----------------------------------*/
    if ( (i = siblingTable->findIterator(handle.getKey() ) )
            != siblingTable->end() ) {

        // not alive? -> do not change routing information
        if (!isAlive)
            return true;

        // refresh sibling
        (*i) = kadHandle;
        return true;
    }

    /* check if node is already in a bucket ---------------------------------*/
    KademliaBucket* bucket = routingBucket(handle.getKey(), false);
    if (bucket != NULL && (i = bucket->findIterator(handle.getKey() ) )
            != bucket->end() ) {

        // not alive? -> do not change routing information
        if (!isAlive)
            return true;

        // remove old handle
        bucket->erase(i);

        // re-add to tail 
        bucket->push_back(kadHandle);

        return true;
    }

    /* check if node can be added to the sibling list -----------------------*/
    if (siblingTable->isAddable(handle) ) {

        // check if sibling list is full so a handle is preemted from the list
        if (siblingTable->isFull()) {

            // get handle thats about to be preempted
            KademliaNodeHandle& oldHandle = siblingTable->back();

            // add handle to the sibling list
            siblingTable->add(kadHandle);

            // TODO: only call update() for real siblings
            callUpdate(oldHandle, false);
            callUpdate(kadHandle, true);

            // change, so that the preemted handle is added to a bucket
            kadHandle = oldHandle;

            // return always true, since the handle has been added
            result |= true;
        } else {

            // simply add the handle and stop
            siblingTable->add(kadHandle);
            callUpdate(kadHandle, true);
            updateTooltip();
            return true;
        }
        updateTooltip();
    }

    /* add node to the appropritate bucket, if not full ---------------------*/
    bucket = routingBucket(kadHandle.getKey(), true);
    if (!bucket->isFull()) {
        EV << "Kademlia::routingAdd(): Adding new node " << kadHandle
        << " to bucket " << routingBucketIndex(kadHandle.getKey()) << endl;
        bucket->push_back(kadHandle);
        result |= true;
    }
    return result;
}

bool Kademlia::routingRemove(const OverlayKey& key)
{
    return routingTimeout(key, true);
}

bool Kademlia::routingTimeout(const OverlayKey& key, bool immediately)
{
    // key unspecified? yes -> ignore
    if (key.isUnspecified())
        return false;

    // bucket index
    KademliaBucket::iterator i;

    /* check if the node is one of the siblings -----------------------------*/
    if ( (i = siblingTable->findIterator(key) ) != siblingTable->end() ) {

        i->incStaleCount();
        if (i->getStaleCount() > removalDelay || immediately) {
            // remove from sibling table
            callUpdate(*i, false);
            updateTooltip();
            siblingTable->erase(i);

            // try to refill with new closest contact
            // TODO!
            return true;
        }
    }

    /* check if node is already in a bucket ---------------------------------*/
    KademliaBucket* bucket = routingBucket(key, false);
    if (bucket != NULL && (i = bucket->findIterator(key) ) != bucket->end() ) {

        i->incStaleCount();
        if (i->getStaleCount() > removalDelay || immediately) {
            // remove from routing table
            bucket->erase(i);
            return true;
        }
    }
    return false;
}

bool Kademlia::isSiblingFor(const NodeHandle& node, const OverlayKey& key,
        int numSiblings, bool* err)
{
    if (key.isUnspecified())
        error("Kademlia::isSiblingFor(): key is unspecified!");

    if (state != READY) {
        *err = true;
        return false;
    }

    if (numSiblings > getMaxNumSiblings()) {
        opp_error("Kademlia::isSiblingFor(): numSiblings too big!");
    }
    // set default number of siblings to consider
    if (numSiblings == -1)
        numSiblings = getMaxNumSiblings();

    if (numSiblings == 0) {
        *err = false;
        return (node.getKey() == key);
    }

    // return false, if sibling table is not full
    if (siblingTable->size() <= (uint)numSiblings ) {
        //       *err = true;
        //        return false;
        //cout << "sibling table not full" << endl;
    }

    KeyDistanceComparator<KeyXorMetric>* comp =
            new KeyDistanceComparator<KeyXorMetric>( key );

    // create result vector
    NodeVector* result = new NodeVector( numSiblings, comp );

    //TODO now:searching all buckets
    // add items from buckets
    int mainIndex = routingBucketIndex(key);

    for (int i=1; !result->isFull() && i<numBuckets*3; i++) {
        int index = mainIndex + ( ((i&1)==1) ? -1 : 1 ) * (i / 2 );
        if (index<0 || index>=numBuckets)
            continue;

        // add bucket to result vector
        KademliaBucket* bucket = routingTable[index];
        if (bucket!=NULL) {
            for (KademliaBucket::iterator i=bucket->begin(); i!=bucket->end(); i++) {
                result->add( *i);
                EV << "Kademlia::routingGetClosestNodes(): Adding "
                << *i << " from bucket " << index << endl;
            }
        }
    }

    for (KademliaBucket::iterator i=siblingTable->begin(); i
            != siblingTable->end(); i++)
        result->add( *i);

    // add local node
    result->add(thisNode);

    *err = false;
    delete comp;

    if (result->contains(node.key)) {
        /*
         if (numSiblings == 1) {
         std::cout << siblingTable->size() << " " << result->size() << std::endl;
         for (NodeVector::iterator i = result->begin();
         i != result->end(); i++)
         std::cout << *i << std::endl;
         std::cout << std::endl;
         }
         */
        delete result;
        return true;
    } else {
        delete result;
        return false;
    }

    // the following doesn't work!!!

    // get distance from this node
    //    OverlayKey d1 = getThisNode().getKey()^key;

    // get siblings range
    //    OverlayKey d2 = getThisNode().getKey()^
    //                    (siblingTable->at(numSiblings).getKey());

    //    return d1 < d2;
}

bool Kademlia::forwardMessageRecursive(const TransportAddress& dest,
        BaseRouteMessage* msg)
{
    if (msg->getSrcNode() != thisNode) {
        routingAdd(msg->getSrcNode(), true);

        NodeVector* nextHops = findNode(msg->getDestKey(), k, s, msg);
        KademliaRoutingInfoMessage* kadRoutingInfoMsg =
                new KademliaRoutingInfoMessage();

        kadRoutingInfoMsg->setSrcNode(thisNode);
        kadRoutingInfoMsg->setNextHopsArraySize(nextHops->size());
        for (uint i = 0; i < nextHops->size(); i++) {
            kadRoutingInfoMsg->setNextHops(i, (*nextHops)[i]);
        }

        sendMessageToUDP(msg->getSrcNode(), kadRoutingInfoMsg);
    }
    return true;
}

NodeVector* Kademlia::findNode(const OverlayKey& key, int numRedundantNodes,
        int numSiblings, BaseOverlayMessage* msg)
{
    if ((numRedundantNodes > getMaxNumRedundantNodes()) || (numSiblings
            > getMaxNumSiblings())) {

        opp_error("(Kademlia::findNode()) numRedundantNodes or numSiblings "
            "too big!");
    }

    // create temporary comparator
    KeyDistanceComparator<KeyXorMetric>* comp =
            new KeyDistanceComparator<KeyXorMetric>( key );

    // select result set size
    bool err;

    int resultSize = isSiblingFor(thisNode, key, numSiblings, &err) ?
                (numSiblings ? numSiblings : 1) : numRedundantNodes;

    // create result vector
    NodeVector* result = new NodeVector( resultSize, comp );

    // add items from buckets
    int mainIndex = routingBucketIndex(key);

    for (int i=1; !result->isFull() && i<numBuckets*3; i++) {
        int index = mainIndex + ( ((i&1)==1) ? -1 : 1 ) * (i / 2 );
        if (index<0 || index>=numBuckets)
            continue;

        // add bucket to result vector
        KademliaBucket* bucket = routingTable[index];
        if (bucket!=NULL) {
            for (KademliaBucket::iterator i=bucket->begin(); i!=bucket->end(); i++) {
                result->add( *i);
                EV << "Kademlia::routingGetClosestNodes(): Adding "
                << *i << " from bucket " << index << endl;
            }
        }
    }

    // add nodes from sibling table 
    for (KademliaBucket::iterator i=siblingTable->begin(); i
            != siblingTable->end(); i++)
        result->add( *i);
    //TODO: Only add if needed!!

    // add local node
    result->add(thisNode);

    if (isSiblingFor(thisNode, key, numSiblings, &err)) {
        if ((*result)[0].key != key) {
            EV << "!!!!! Node " << (*result)[0] << " should be close to "
            << key;
            if (siblingTable->contains(key)) {
                EV << " +++" << endl;
            } else {
                EV << " ---" << endl;
            }
        }
    }

    delete comp;
    return result;
}

//-----------------------------------------------------------------------------


void Kademlia::handleTimerEvent(cMessage* msg)
{
    if (msg->isName("stabilizeTimer")) {
        handleStabilizeTimerExpired(msg);
    }
}

void Kademlia::handleUDPMessage(BaseOverlayMessage* msg)
{
    OverlayCtrlInfo* ctrlInfo =
            check_and_cast<OverlayCtrlInfo*>(msg->removeControlInfo());
    KademliaRoutingInfoMessage* kadRoutingInfoMsg = check_and_cast<
            KademliaRoutingInfoMessage*>(msg);

    routingAdd(kadRoutingInfoMsg->getSrcNode(), true);

    for (uint i = 0; i < kadRoutingInfoMsg->getNextHopsArraySize(); i++) {
        routingAdd(kadRoutingInfoMsg->getNextHops(i), true);
    }

    delete ctrlInfo;
    delete msg;
}

//In Kademlia this method is used to maintain the routing table.
bool Kademlia::handleRpc(BaseCallMessage* msg)
{
    RPC_SWITCH_START( msg )
    RPC_ON_CALL( Ping ) {
        // add active node
        routingAdd( _PingCall->getSrcNode(), true );
        break;
    }
    RPC_ON_CALL(FindNode)
    {
        // add active node
        routingAdd(_FindNodeCall->getSrcNode(), true);
        break;
    }
    RPC_SWITCH_END( )
    return false;
}

//In Kademlia this method is used to maintain the routing table.
void Kademlia::handleRpcResponse(BaseResponseMessage* msg, int rpcId,
        simtime_t rtt)
{
    // add node that reponded
    routingAdd(msg->getSrcNode(), true);

    RPC_SWITCH_START( msg )
    RPC_ON_RESPONSE( Ping ) {
        // add active node
        routingAdd( _PingResponse->getSrcNode(), true );
        break;
    }
    RPC_ON_RESPONSE(FindNode)
    {
        // add active node
        routingAdd(_FindNodeResponse->getSrcNode(), true);
        // add inactive nodes
        for (uint i=0; i<_FindNodeResponse->getClosestNodesArraySize(); i++)
            routingAdd(_FindNodeResponse->getClosestNodes(i), false);
        break;
    }
    RPC_SWITCH_END( )
}

// In Kademlia this method is used to maintain the routing table.
void Kademlia::handleRpcTimeout(BaseCallMessage* msg,
        const TransportAddress& dest, int rpcId, const OverlayKey& destKey)
{
    RPC_SWITCH_START( msg )
    RPC_ON_CALL( Ping ) {
        routingTimeout( destKey );
        break;
    }
    RPC_ON_CALL(FindNode)
    {
        routingTimeout(destKey);
        break;
    }
    RPC_SWITCH_END( )
}

void Kademlia::lookupFinished()
{

    if (state == JOIN) {
        // registerBootstrapNode( thisNode );
        setOverlayReady(true);
    }

    state = READY;
}

// handle a expired stabilize timer
void Kademlia::handleStabilizeTimerExpired(cMessage* msg)
{
    if (state == READY)
        state = REFRESH;

    // refresh buckets
    createLookup()->lookup(~getThisNode().getKey(), getMaxNumSiblings(), 0, 0,
                new KademliaLookupListener(this));
    createLookup()->lookup(getThisNode().getKey(), getMaxNumSiblings(), 0, 0,
                new KademliaLookupListener(this));
    createLookup()->lookup(getThisNode().getKey() + OverlayKey::ONE, 0, 0, 0,
                new KademliaLookupListener(this));

    for (int i=0; i<20; i++) {
        createLookup()->lookup(getThisNode().getKey() ^ 
                (OverlayKey::ONE<<(keyLength-i-1)), 0, 0, 0,
                new KademliaLookupListener(this));
    }

    // refresh sibling table
    for (int i = 0; i < s && i < (int)siblingTable->size(); i++) {
        NodeHandle& handle = siblingTable->at(i);
        FindNodeCall* findNodeCall = new FindNodeCall();
        findNodeCall->setLookupKey(handle.getKey() );
        findNodeCall->setSrcNode(getThisNode() );
        findNodeCall->setNumRedundantNodes(getMaxNumRedundantNodes() );
        findNodeCall->setNumSiblings(getMaxNumSiblings() );
        findNodeCall->setLength(FINDNODECALL_L(call) );

        RECORD_STATS(numFindNodeSent++;
                     bytesFindNodeSent += findNodeCall->byteLength());

        sendUdpRpcCall(handle, findNodeCall);
    }

    if (state == REFRESH)
        state = READY;

    // schedule next stabilization process
    cancelEvent(stabilizeTimer);
    scheduleAt(simulation.simTime() + stabilizeDelay, stabilizeTimer);
}

//virtual public: xor metric
OverlayKey Kademlia::distance(const OverlayKey& x, const OverlayKey& y) const
{
    return x^y;
}

void Kademlia::updateTooltip()
{
    if (ev.isGUI()) {
        std::stringstream ttString;

        // show our nodeId in a tooltip
        ttString << "This: " << thisNode << endl << "Siblings: "
                << siblingTable->size();

        parentModule()->parentModule()->displayString().
        setTagArg("tt", 0, ttString.str().c_str());
        parentModule()->displayString().
        setTagArg("tt", 0, ttString.str().c_str());
        displayString().setTagArg("tt", 0, ttString.str().c_str());
    }
}

