//
// Copyright (C) 2006 Institut fuer Telematik, Universitaet Karlsruhe (TH)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//

/**
 * @file Kademlia.h
 * @author Sebastian Mies
 */

#ifndef __KADEMLIA_H_
#define __KADEMLIA_H_

// OverSim includes
#include <oversim.h>

// Kademlia includes
#include "KademliaNodeHandle.h"
#include "KademliaBucket.h"


/**
 * Kademlia overlay module
 *
 * @author Sebastian Mies
 */
class Kademlia : public BaseOverlay {
private:
    
    KeyDistanceComparator<KeyXorMetric>* comparator;

    KademliaBucket*  siblingTable;
    std::vector<KademliaBucket*> routingTable;
    int numBuckets;

private://methods: routing table maintainance

    void routingInit();
    void routingDeinit();

    /**
     * Returns the index of the bucket the key would reside
     * with respect to kademlia parameters
     * 
     * @param key The key of the node
     * @return int The index of the bucket
     */
    int routingBucketIndex( const OverlayKey& key );

    /**
     * Returns a Bucket or <code>NULL</code> if the bucket has not
     * yet allocated. If ensure is true, the bucket allocation is
     * ensured.
     * 
     * @param key The key of the node
     * @param ensure If true, the bucket allocation is ensured
     * 
     * @return Bucket* The Bucket
     */
    KademliaBucket* routingBucket( const OverlayKey& key, bool ensure );

    /**
     * Adds a node to the routing table
     * 
     * @param handle handle to add
     * @param isAlive true, if it is known that the node is alive
     * @return true, if the node was known or has been added
     */
    bool routingAdd( const NodeHandle& handle, bool isAlive );

    /**
     * Removes a node from the routing table
     * 
     * @param key Key of the Node
     * @return true, if the node has been removed
     */
    bool routingRemove( const OverlayKey& key );

    /**
     * Removes a node after a number of timeouts or immediatly
     * if immediately is true (behaves like routingRemove).
     * 
     * @param key Node's key to remove
     * @param immediately If true, the node is removed immediately
     * @return true, if the node has been removed
     */
    bool routingTimeout( const OverlayKey& key, bool immediately = false );
    
    bool forwardMessageRecursive( const TransportAddress& dest,
                                  BaseRouteMessage* msg );

protected://fields: kademlia parameters

    int k; /*< number of redundant graphs */
    int b; /*< number of bits to consider */
    int s; /*< number of siblings         */
    int r; /*< number of replicas         */

    int removalDelay; /*< number of timouts until node is removed from routingtable */
    int keyLength; 

    double    stabilizeDelay; 
    cMessage* stabilizeTimer; 

public://methods: overloaded from BaseOverlay

    ~Kademlia();

    // see BaseOverlay.h
    void initializeOverlay(int stage);

    // see BaseOverlay.h
    void finishOverlay();

    // see BaseOverlay.h
    void joinOverlay();

    // see BaseOverlay.h
    bool isSiblingFor( const NodeHandle& node,const OverlayKey& key,
			int numSiblings, bool* err );

    // see BaseOverlay.h
    int getMaxNumSiblings();

    // see BaseOverlay.h
    int getMaxNumRedundantNodes();

    // see BaseOverlay.h
    void handleTimerEvent(cMessage* msg);

    // see BaseOverlay.h
    void handleUDPMessage(BaseOverlayMessage* msg);

    // see BaseOverlay.h
    bool handleRpc( BaseCallMessage* msg );

protected://methods from BaseOverlay

    // see BaseOverlay.h
    NodeVector* findNode(const OverlayKey& key,
			 int numRedundantNodes,
			 int numSiblings,
			 BaseOverlayMessage* msg);

    // see BaseOverlay.h
    void handleRpcResponse( BaseResponseMessage* msg, int rpcId,
			    simtime_t rtt );

    // see BaseOverlay.h
    void handleRpcTimeout(BaseCallMessage* msg, const TransportAddress& dest,
			  int rpcId, const OverlayKey& destKey);

protected://methods implemented for Kademlia

    /**
     * handle a expired stabilize timer
     *
     * @param msg the timer self-message
     */
    void handleStabilizeTimerExpired(cMessage* msg);

    // see BaseOverlay.h
    OverlayKey distance( const OverlayKey& x, const OverlayKey& y ) const;

    /**
     * updates information shown in GUI
     */
    void updateTooltip();

    // see BaseOverlay.h
    virtual void lookupFinished();

    friend class KademliaLookupListener;
};

#endif
