//
// Copyright (C) 2006 Institut fuer Telematik, Universitaet Karlsruhe (TH)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

#ifndef __KADEMLIA_NODE_HANDLE_H
#define __KADEMLIA_NODE_HANDLE_H

/**
 * Kademlia NodeHandle
 * 
 * @author Sebastian Mies
 */
class KademliaNodeHandle : public NodeHandle {
private:
    uint8_t staleCount;
    simtime_t lastSeen;
public:
    /**
     * Constructs an unspecified NodeHandle
     */
    KademliaNodeHandle( ) : NodeHandle() 
    {
        staleCount = 0;
    }

    /**
     * Copy constructor.
     *
     * @param handle The NodeHandle to copy
     */
    KademliaNodeHandle( const KademliaNodeHandle& handle ) : NodeHandle( handle ) 
    {
        staleCount = handle.staleCount;
    }

    KademliaNodeHandle( const NodeHandle& handle )
    {
        this->ip = handle.ip;
        this->port = handle.port;
        this->key = handle.key;
        staleCount = 0;
    }

    /**
     * Complete constructor.
     *
     * @param key The OverlayKey
     * @param ip The IPvXAddress
     * @param port The UDP-Port
     */
    KademliaNodeHandle( const OverlayKey& key,
                const IPvXAddress& ip,
                int port) : NodeHandle( key, ip, port)
    {
        staleCount = 0;
    }

    inline uint8_t getStaleCount() const {
        return staleCount;
    }

    inline void setStaleCount( uint8_t staleCount ) {
        this->staleCount = staleCount;
    }

    inline void resetStaleCount() {
        this->setStaleCount(0);
    }

    inline void incStaleCount() {
        this->staleCount++;
    }

    inline void setLastSeen( simtime_t lastSeen ) {
        this->lastSeen = lastSeen;
    }

    inline simtime_t getLastSeen() {
        return this->lastSeen;
    }
};

/**
 * Class for extracting the relevant OverlayKey from a type used as template
 * parameter T_key for NodeVector<T, T_key> - Version for plain NodeHandle.
 * 
 * @author Sebastian Mies
 * @author Felix M. Palmen
 */
template <>
struct KeyExtractor<KademliaNodeHandle>
{
    static const OverlayKey& key(const KademliaNodeHandle& node)
    {
	return node.key;
    };
};

#endif
