//
// Copyright (C) 2006 Institut fuer Telematik, Universitaet Karlsruhe (TH)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//

/**
* @file ConnectivityProbe.cc
* @author Helge Backhaus
*/

#include "ConnectivityProbe.h"

Define_Module(ConnectivityProbe);

VTopologyNode::VTopologyNode(int moduleID)
{
    this->moduleID = moduleID;
    componentCounter = 0;
}

Vast* VTopologyNode::module() const
{
    return check_and_cast<Vast*>(simulation.module(moduleID));
}

void ConnectivityProbe::initialize()
{
    probeIntervall = par("connectivityProbeIntervall");
    probeTimer = new cMessage("probeTimer");
    if(probeIntervall != 0.0f) {
        scheduleAt(simulation.simTime() + probeIntervall, probeTimer);

        cOV_NodeCount.setName("Total Node Count");
        cOV_MaximumComponent.setName("Maximum Component Size");
        cOV_ZeroMissingNeighbors.setName("Nodes with 0 Missing Neighbors");
        cOV_OneMissingNeighbor.setName("Nodes with 1 Missing Neighbor");
        cOV_TwoMissingNeighbor.setName("Nodes with 2 Missing Neighbor");
        cOV_ThreeMissingNeighbor.setName("Nodes with 3 Missing Neighbor");
        cOV_FourMissingNeighbor.setName("Nodes with 4 Missing Neighbor");
        cOV_FiveMissingNeighbors.setName("Nodes with 5 Missing Neighbors");
        cOV_SeveralMissingNeighbors.setName("Nodes with more than 5 Missing Neighbors");
        cOV_AverageMissingNeighbors.setName("Average Missing Neighbors per Node");
        cOV_MaxMissingNeighbors.setName("Max Missing Neighbors per Node");
    }
}

void ConnectivityProbe::handleMessage(cMessage* msg)
{
    // catch self timer messages
    if(msg->isName("probeTimer")) {
        //reset timer
        cancelEvent(probeTimer);
        scheduleAt(simulation.simTime() + probeIntervall, msg);

        // fill topology with all VAST modules
        extractTopology();

        int Component = 1;
        unsigned int maxComponent = 0;
        for(VTopology::iterator itTopology = Topology.begin(); itTopology != Topology.end(); ++itTopology) {
            int count = findComponent(itTopology->second.module()->getThisNode().key, Component);
            if(count > maxComponent) {
                maxComponent = count;
            }
            if(maxComponent == Topology.size()) {
                break;
            }
            ++Component;
        }

        cOV_NodeCount.record((double)Topology.size());
        cOV_MaximumComponent.record((double)maxComponent);

        int mnZero, mnOne, mnTwo, mnThree, mnFour, mnFive, mnSeveral, mnMax;
        double mnAverage = 0.0;
        double AOIWidthSqr = Topology.begin()->second.module()->getAOI();
        AOIWidthSqr *= AOIWidthSqr;
        mnZero = mnOne = mnTwo = mnThree = mnFour = mnFive = mnSeveral = mnMax = 0;

        for(VTopology::iterator itTopology = Topology.begin(); itTopology != Topology.end(); ++itTopology) {
            NodeSet SiteHandles;
            Vector2D vastPosition = itTopology->second.module()->getPosition();
            int missing = 0;
            for(VTopology::iterator itI = Topology.begin(); itI != Topology.end(); ++itI) {
                if(itI != itTopology && vastPosition.distanceSqr(itI->second.module()->getPosition()) <= AOIWidthSqr) {
                    SiteHandles.insert(itI->second.module()->getThisNode());
                }
            }

            for(NodeSet::iterator itSiteHandles = SiteHandles.begin(); itSiteHandles != SiteHandles.end(); ++itSiteHandles) {
                if(itTopology->second.module()->getList()->Sites.find(*itSiteHandles) == itTopology->second.module()->getList()->Sites.end()) {
                    ++missing;
                }
            }

            switch(missing) {
                case 0:
                    ++mnZero;
                break;
                case 1:
                    ++mnOne;
                break;
                case 2:
                    ++mnTwo;
                break;
                case 3:
                    ++mnThree;
                break;
                case 4:
                    ++mnFour;
                break;
                case 5:
                    ++mnFive;
                break;
                default:
                    ++mnSeveral;
            }
            mnAverage += missing;
            if(mnMax < missing) {
                mnMax = missing;
            }
        }
        mnAverage /= (double)Topology.size();

        cOV_ZeroMissingNeighbors.record((double)mnZero);
        cOV_OneMissingNeighbor.record((double)mnOne);
        cOV_TwoMissingNeighbor.record((double)mnTwo);
        cOV_ThreeMissingNeighbor.record((double)mnThree);
        cOV_FourMissingNeighbor.record((double)mnFour);
        cOV_FiveMissingNeighbors.record((double)mnFive);
        cOV_SeveralMissingNeighbors.record((double)mnSeveral);
        cOV_AverageMissingNeighbors.record(mnAverage);
        cOV_MaxMissingNeighbors.record((double)mnMax);

        Topology.clear();
    }
}

void ConnectivityProbe::extractTopology()
{
    for(int i=0; i<=simulation.lastModuleId(); i++) {
        cModule* module = simulation.module(i);
        if(module && dynamic_cast<Vast*>(module)) {
            Vast* vast = check_and_cast<Vast*>(module);
            if(vast->getState() == BaseOverlay::READY) {
                VTopologyNode temp(i);
                Topology.insert(std::make_pair(vast->getThisNode().key, temp));
            }
        }
    }
}

int ConnectivityProbe::findComponent(OverlayKey key, int Component)
{
    VTopology::iterator itEntry = Topology.find(key);
    if(itEntry != Topology.end() && itEntry->second.componentCounter < Component) {
        int visited = 1;
        itEntry->second.componentCounter = Component;
        Vast* vast = itEntry->second.module();
        for(SiteMap::iterator itSites = vast->getList()->Sites.begin(); itSites != vast->getList()->Sites.end(); ++itSites) {
            visited += findComponent(itSites->first.key, Component);
        }
        return visited;
    }
    return 0;
}

ConnectivityProbe::~ConnectivityProbe()
{
    // destroy self timer messages
    cancelAndDelete(probeTimer);
}
