//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//


#include "YMMOGMessage_m.h"
#include "YMMOG.h"
#include "TreeNode.h"
#include <iostream>
Define_Module(YMMOG);

void YMMOG::initialize()
{
	// Get Parameters
	timeSlotSize = par("timeSlotSize");
	maxPlayerLeafs = par("maxPlayerLeafs");
	treeWidth = par("treeWidth");
	subscriptionRange = par("subscriptionRange");
	unsubscriptionRange = par("unsubscriptionRange");
	logLevel = par("logLevel");
	maxFailed = par("maxFailed");
	maxFailedPings = par("maxFailedPings");
		
	
	thisNodeBandwith = (check_and_cast<cBasicChannel *>(parentModule()->parentModule()->gate("out",0)->channel()))->datarate() / (1000);
	thisNodeDelay = (check_and_cast<cBasicChannel *>(parentModule()->parentModule()->gate("out",0)->channel()))->delay();
	magicLog = check_and_cast<SubSpaceTable*>(simulation.moduleByPath("globalObserver.globalFunctions[0]"));

	//Initiate Timeslot
	maxPlayerOnline = 0;
	computedEvents = 0;
		
	// timers:
	timeSlotTimer = new cMessage("timeSlotEnd");
	pingTimer = new cMessage("pingTimer");
	delayTimer = new cMessage("delayTimer");
	playerPingTimer = new cMessage("playerPingTimer");
	scheduleAt(simTime() + 30, playerPingTimer);
	pingDelay = 0.2;
	delayDelay = 1;
	scheduleAt(simTime() + timeSlotSize, timeSlotTimer);
	sending = true;
	
	//Player things
	player_id = 0;
	xpos = 0;
	ypos = 0;
	currentSubSpaceId =0;
	lastCheckedX = 0;
	lastCheckedY = 0;
	// Watches:
	WATCH(player_id);
	WATCH(xpos);
	WATCH(ypos);
	WATCH(currentSubSpaceId);
	WATCH_VECTOR(subscribed_SubSpaces);
	WATCH( thisNodeBandwith );
	WATCH( thisNodeDelay );
	
	WATCH ( computedEvents );
	WATCH ( maxPlayerOnline );
	
	WATCH_PTRVECTOR ( responsible_nodes );
	WATCH_LIST(backup_nodes);
	WATCH_LIST(intermediate_nodes);
	
	//Statistics
	savedRequests=0;
	passedSubSpaces=0;
	WATCH(savedRequests);
	WATCH(passedSubSpaces);
	
	//NodeLog
	WATCH_LIST(nodeLog);	
}

void YMMOG::finish()
{	
	magicLog->record(-1,8,simTime()-sessionBegin);
	finishStats();
	globalStatistics->addStdDev("Own Bandwith: ", (thisNodeBandwith*1000 / 8));
	globalStatistics->addStdDev("Own Delay: ", thisNodeDelay);
	
	nodeLog.clear();
	subscribed_SubSpaces.clear();
	cancelEvent(delayTimer);
	delete delayTimer;
	for(i = responsible_nodes.begin(); i != responsible_nodes.end(); ++i) {
		delete *i;
	}
	responsible_nodes.clear();
	intermediate_nodes.clear();
	backup_nodes.clear();
}


void YMMOG::processSelfTimers(cMessage* msg) {
	if(msg->isName("resendTimer")) {
		processResendTimer();
	} else if( msg->isName("timeSlotEnd") ) {
		processTimeSlotEnd();
	}
	else if( msg->isName("pingTimer") ) {
		processPingTimer();
	}
	else if( msg->isName("delayTimer")) {
		processDelayTimer();
	}
	else if( msg->isName("stabilizeTimer") ) {
		processStabilizeTimer();
	}
	else if( msg->isName("playerPingTimer") ) {
		processPlayerPingTimer();
	}
	else if(msg->isName("bandwithTimer") ) {
		processBandwithTimer();
	}	
}

void YMMOG::processAppMessages(cMessage* msg) {
		YMMOGMessage* ymmog = (YMMOGMessage*) msg;
		if( ymmog->getCommand() == LOGIN ) {
			lobbyIp = bootstrapOracle->getBootstrapNode().ip.get4();
			LoginMessage* loginMsg = new LoginMessage("InitialLogin");
			loginMsg->setCommand(LOGIN);
			loginMsg->setServiceType(REQUEST);
			loginMsg->setBandwith( (int) thisNodeBandwith );
			sendMsg(loginMsg, lobbyIp);
			delete msg;
		}
		else if ( ymmog->getCommand() == PUBLISH ) {
			ClientEvent* clientEvent = (ClientEvent*) msg;
			Event event = clientEvent->getEvent();
			// Compute responsible SubSpace from X/Y coods
			int subSpaceId = (event.ypos/subSpaceSize)*gameSpaceSize + (event.xpos/subSpaceSize);
			if( event.type == PLAYERPOSITION ) {
				xpos = event.xpos;
				ypos = event.ypos;
				// We moved ourself, so lets check if we passed a subspace and need to check our subscriptions
				if( abs( xpos - lastCheckedX) > subscriptionRange /2  || abs( ypos - lastCheckedY) > subscriptionRange/2 ) { 	checkSubscriptions();
				}
				if( subSpaceId != currentSubSpaceId ) {
					passedSubSpaces++;
					currentSubSpaceId = subSpaceId;
					checkSubscriptions();
				}
			}

			if( findSubSpace(subSpaceId) ) {
				sendEvent(subSpaceId, player_id, event);
			}
			else  {
				requestIp(subSpaceId, RESPONSIBLENODE);
				//checkSubscriptions();
				//We are not subscribed to SubSpace yet.. enqueue event and subscribe to subspace first.
				/** missing **/
			}
			delete msg;
		}
}

void YMMOG::processPing(PingMessage* pingMsg) {
	receivedPings++;
	if( pingMsg->getServiceType() == CONFIRMATION) {
		resetPing( pingMsg->getSrcip() );
		delete pingMsg;
	} else {
		pingMsg->setServiceType(CONFIRMATION);
		pingMsg->setSent(simTime());
		sendMsg(pingMsg, pingMsg->getSrcip() );
	}
}

void YMMOG::sendPing(PingMessage* openPing) {
	PingMessage* pingMsg = new PingMessage("BACKUPPING");
	pingMsg->setCommand(PING);
	pingMsg->setServiceType(REQUEST);
	pingMsg->setSent(simTime());
	sendMsg(pingMsg, openPing->getDestip());
}


bool YMMOG::deleteResponsibleNodeData(int id) {
	for(i=responsible_nodes.begin(); i != responsible_nodes.end(); ++i) {	
		if ( (*i)->getId() == id ) {
			delete *i;
			responsible_nodes.erase(i);
			return true; 
		}
	}
	return false;
}
bool YMMOG::deleteIntermediateNodeDataById( int subSpaceId ) {
	for(in = intermediate_nodes.begin(); in != intermediate_nodes.end(); ++in) {
		if( in->getSubSpaceId() == subSpaceId) {
			intermediate_nodes.erase(in);
			return true;
		}
	}
	return false;
}

void YMMOG::deleteBackupNodeDataById( int subSpaceId) {
	for( b = backup_nodes.begin(); b != backup_nodes.end(); ++b) {
		if( b->getSubSpace().getId() == subSpaceId) {
			backup_nodes.erase(b);
			break;
		}
	}
}

ResponsibleNodeData* YMMOG::findResponsibleNodeDataById(int id) {
	for(i=responsible_nodes.begin(); i != responsible_nodes.end(); ++i) {	
		if ( (*i)->getId() == id ) {
			return (*i); 
		}
	}
	return NULL;
}

BackupNodeData* YMMOG::findBackupNodeDataById(int subSpaceId) {
	for(b=backup_nodes.begin(); b != backup_nodes.end(); ++b)  {
		if( b->getId() == subSpaceId ) {
			return &(*b);
		}
	}
	return NULL;
}

IntermediateNodeData* YMMOG::findIntermediateNodeDataById(int subSpaceId) {
	for( in = intermediate_nodes.begin(); in != intermediate_nodes.end() ; in++) {
		if( in->getSubSpaceId() == subSpaceId) {
			return &(*in);
		}
	}
	return NULL;
}


void YMMOG::sendEvent(int subSpaceId, int playerId, Event event) {	
	PublishMessage* publishMsg = new PublishMessage("send_Event");
	publishMsg->setCommand(PUBLISH);
	publishMsg->setServiceType(REQUEST);
	publishMsg->setPlayerId( playerId );
	publishMsg->setSubSpaceId( subSpaceId );
	publishMsg->setPublishedEvent( event );
	// Send the Message
	sendMsg( publishMsg, getSubSpace(subSpaceId)->getRespIp() );
}

void YMMOG::processResendTimer() {
       double min=simTime() + resendDelay;
       double now = simTime();
       std::list<YMMOGMessage*>::iterator i = dupedMessages.begin();
       while( i != dupedMessages.end() ) {
       //&& (*i)->getCommand() != PLAYERLIST
              if( (*i)->getTimeStamp() + resendDelay < now ) {
                     // we have to resend it
                     (*i)->setTimeStamp(now);
                     (*i)->setFailed((*i)->getFailed() + 1);
                     if( (*i)->getFailed() > maxFailed ) {
                     	    if( (*i)->getCommand() == LOGIN ) {
                     	        YMMOGMessage* tmp = *i;
                     	        tmp->setFailed(0);
                            	i = dupedMessages.erase(i);
                           	sendMsg(tmp, tmp->getDestip());
                     	    }
                     	    else if( (*i)->getCommand() == REQUESTIP) {
                     	    	RequestIpMessage* rmsg = dynamic_cast<RequestIpMessage*>(*i);
                     	    	i = dupedMessages.erase(i);
                     	    	if( rmsg != NULL ) {
                     	    		requestIp(rmsg->getSubSpaceId(), rmsg->getNodeType() );
                     	    	}
                     	    	delete rmsg;	
                     	    }
                     	    else {
                     	  	  delete (*i);
                          	  i = dupedMessages.erase(i);
                            	droppedMessages++;
                            }
                     }
                     else {
                            YMMOGMessage* tmp = *i;
                            i = dupedMessages.erase(i);
                            sendMsg(tmp, tmp->getDestip());
                     }
		}
		else {
			if( ((*i)->getTimeStamp() + resendDelay < min) &&  (*i)->getTimeStamp() + resendDelay > simTime() ) min = (*i)->getTimeStamp() + resendDelay;
			++i;
		}
	}
	selfMessages++;
	cancelEvent(resendTimer);
	scheduleAt(min + 0.05, resendTimer);
}



void YMMOG::processLogin(LoginMessage* loginMsg) {
/** 
This Function handles the login Messages recieved from the Lobby Server.
It extracts the initial position and playerid and forwards it via Message to the GameClient Applikation.
It also subscribe to the first subspaces we got in the loginMessage and initiates some Timers.
**/
	receivedLogins++;
	if( loginMsg->getServiceType() == RESPONSE ) { 
		if( player_id != 0 ) {
			// We're already logged in it seems our confirmation was dropped. Resend it
			loginMsg->setServiceType(CONFIRMATION);
			sendMsg(loginMsg, loginMsg->getSrcip());
			return;
		}
		player_id = loginMsg->getId ();
		xpos = loginMsg->getXpos ();
		ypos = loginMsg->getYpos ();
		subSpaceSize = loginMsg->getSubSpaceSize();
		gameSpaceSize = loginMsg->getGameSpaceSize();
		currentSubSpaceId = (ypos/subSpaceSize)*gameSpaceSize + (xpos/subSpaceSize);
		// Lets Subscribe to the first responsible Nodes we got
		for(unsigned int k=0; k < loginMsg->getResponsibleNodesArraySize(); k++) {
			subscribe( loginMsg->getResponsibleNodes(k).getId(), loginMsg->getResponsibleNodes(k).getRespIp() ); 
		}
		deleteFromDupedMessages(loginMsg);
		loginMsg->setServiceType(CONFIRMATION);
		sendMsg(loginMsg, loginMsg->getSrcip());
		
		/** Forward data to gameclient**/
		ClientLogin* clientLogin = new ClientLogin("Clientlogin");
		clientLogin->setCommand(LOGIN);
		clientLogin->setPlayerId(player_id);
		clientLogin->setXpos(xpos);
		clientLogin->setYpos(ypos);
		clientLogin->setGameSpaceSize(gameSpaceSize*subSpaceSize - 1);
		send(clientLogin, "to_app");
	}
}

void YMMOG::processTimeSlotEnd() {
/**
	This Function is called every time a timeslot ends.
	It sends the ordered EventList of every responsible Node to all Players subscribed to this Node.
	and the PlayerList to the BackupNodeData
**/
	if( sending ) {
		std::vector<ResponsibleNodeData*>::iterator i;
		for(i=responsible_nodes.begin(); i != responsible_nodes.end(); ++i) {
			int subSpaceId = (*i)->getId();
			magicLog->record(subSpaceId, 4,(*i)->getDroppedEvents());
			std::vector<Event> eventList = (*i)->endTimeSlot(simTime() );  //creates the ordered Eventlist
			int timeSlot = (*i)->getSubSpace().getTimeSlot();
			computedEvents += eventList.size(); // global statistic
			std::list<IPAddress> playerList = (*i)->getList();
			std::list<IPAddress>::iterator pL;
			for( pL = playerList.begin(); pL != playerList.end() ; ++pL ) {
				sendEventList(REQUEST, subSpaceId, timeSlot, eventList, *pL);
			}
			sendPlayerList( COMPLETE, subSpaceId);
			eventList.clear();
		}
		sending = false;
	}
	else { sending = true; }
	// set the new timeout
	cancelEvent(timeSlotTimer);
	scheduleAt(simTime() + timeSlotSize/2, timeSlotTimer);
}

void YMMOG::sendEventList(int serviceType, int subSpaceId, int timeSlot, std::vector<Event> eventList, IPAddress destip) {
	PublishEventListMessage* publishMsg = new PublishEventListMessage("EventList");
	publishMsg->setCommand(PUBLISHEVENTLIST);
	publishMsg->setServiceType( serviceType );
	publishMsg->setSubSpaceId( subSpaceId );
	publishMsg->setSent( simTime() );
	publishMsg->setTimeSlot( timeSlot );
	publishMsg->setEventListArraySize( eventList.size() );
	publishMsg->setSender( thisNode.ip.get4() );
	for( uint i =0; i < eventList.size(); i++) {
		publishMsg->setEventList(i, eventList.at(i));
	}
	sendMsg(publishMsg, destip);	
}


void YMMOG::sendPlayerList(int action, int subSpaceId) {
	ResponsibleNodeData* respNode = findResponsibleNodeDataById(subSpaceId);
	if( respNode != NULL && ( respNode->getSubSpace().getBackupIp() != (IPAddress) "127.0.0.1")  ) {
		int failed = 0;
		std::list<YMMOGMessage*>::iterator i = dupedMessages.begin();
		while( i != dupedMessages.end() ) {
                     	if( (*i)->getCommand() == PLAYERLIST) {
                     		PlayerListMessage* playerListMsg = dynamic_cast<PlayerListMessage*>(*i);
                     		if( playerListMsg != NULL && playerListMsg->getSubSpaceId() == subSpaceId ) {
                     			playerListMsg->setFailed( playerListMsg->getFailed() + 1);
                     			if( playerListMsg->getFailed() > maxFailed ) {
                     				RequestNodeMessage* replaceMsg = new RequestNodeMessage("ReplaceBackupNodeData");
                     				replaceMsg->setCommand(REQUESTNODE);
               	      				replaceMsg->setServiceType(REQUEST);
                     				replaceMsg->setNodeType(BACKUPNODE);
                     				replaceMsg->setSubSpaceId( subSpaceId );
                     				sendMsg(replaceMsg, lobbyIp);
                     				if( logLevel > 0 ) {
                     					nodeLog.push_back( NodeLogEntry(simTime(), 22, playerListMsg->getSubSpaceId(), playerListMsg->getDestip() ) );
                     				}
                     				delete (*i);
                     				i = dupedMessages.erase( i );
                     				return;
                     			}
                     			else {
                     				failed = playerListMsg->getFailed() ;
                     				delete (*i);
                     				i = dupedMessages.erase( i );
                     			}
                     		}
                     		else { ++i; }
                     	}
                     	else {
                     		++i;
                     	}
                }
		PlayerListMessage* playerListMsg = new PlayerListMessage("PlayerListMessage");
		playerListMsg->setCommand(PLAYERLIST);
		playerListMsg->setServiceType(REQUEST);
		playerListMsg->setTimeStamp( simTime() );
		playerListMsg->setFailed( failed );		
		playerListMsg->setSubSpaceId( subSpaceId );
		std::list<TreeNode> tmp = respNode->getPlayerList();
		playerListMsg->setTreeArraySize( tmp.size() );
		int t = 0;
		while ( tmp.size() > 0 ) {
			playerListMsg->setTree( t, tmp.front() );
			tmp.pop_front();
			t++;
		}
		/*
		if( action == COMPLETE ) {
			std::list<PlayerNode> playerList = respNode->getPlayerList();
			std::list<PlayerNode>::iterator pL;
			playerListMsg->setPlayerListArraySize(playerList.size());
			int i=0;
			for(pL=playerList.begin(); pL != playerList.end(); ++pL ) {
				playerListMsg->setPlayerList(i, *pL);
				i++;
			}
			playerListMsg->setAction(COMPLETE);
		}*/
		playerListMsg->setAction(COMPLETE);
		sendMsg(playerListMsg, respNode->getBackupIp());
	}
}


	
void YMMOG::processSubscribe(SubscribeMessage* subscribeMsg) {
/** 
	This function is called every time a SubScribeMessage is recieved.
	It handels both for the playernode and for the responsibleNodes.
	SubscribeMsg is for a responsibleNode if ServiceType REQUEST or CONFIRMATION
	It is for the playerNode if it is a RESPONSE.
**/
	receivedSubscribes++;
	if( subscribeMsg->getServiceType() == REQUEST ) {
		// Are we holding the subSpace the player trys to subscribe? 
		int subSpaceId = subscribeMsg->getSubSpaceId();
		ResponsibleNodeData* respNode = findResponsibleNodeDataById( subSpaceId );
		if (respNode != NULL ) {
			//yes we do, so subscribe him and send a response.
			maxPlayerOnline++;
			PlayerNode* player = new PlayerNode( subscribeMsg->getPlayerId(), subscribeMsg->getSrcip(), 0,0);
			// Send eventually intermediateNodes messages 
			sendTreeMessages( respNode, respNode->subscribe( *player ) );
			magicLog->record(subSpaceId, 3, respNode->getPlayerListSize() );
			magicLog->record(subSpaceId, 6, respNode->getAllIntNodes().size() );
			// Respond
			subscribeMsg->setGameState( respNode->getSubSpace().getGameState() );	
			subscribeMsg->setServiceType(RESPONSE);
			sendMsg(subscribeMsg, subscribeMsg->getSrcip());
		}
		else {
			delete subscribeMsg;
		}
	}
	else if ( subscribeMsg->getServiceType() == RESPONSE ) {
		subscribed_SubSpaces.push_back( SubSpace( subscribeMsg->getSubSpaceId(), subscribeMsg->getSrcip() ) );
		// Delete Subscribe from OpenSubscribes
		deleteFromDupedMessages( subscribeMsg );
		delete subscribeMsg;
	}
	else if ( subscribeMsg->getServiceType() == FORWARD ) {
		IntermediateNodeData* inode = findIntermediateNodeDataById(subscribeMsg->getSubSpaceId());
		if( inode != NULL ) {
			inode->addIp( subscribeMsg->getPlayerIp() );
		}
		subscribeMsg->setServiceType(CONFIRMATION);
		sendMsg( subscribeMsg, subscribeMsg->getSrcip() );
	}
	else { // Confirmation
		deleteFromDupedMessages( subscribeMsg );
		delete subscribeMsg;
	}
	
}

void YMMOG::processUnSubscribe(UnSubscribeMessage* unsubscribeMsg) {
/**
	This function is called every time a unsubscribe Message arrives.
	The message is for the playernode if action = response or for one responsibleNode we're holding if it is request/confirmation
**/
	receivedUnSubscribes++;
	if( unsubscribeMsg->getServiceType() == REQUEST) {
		int subSpaceId = unsubscribeMsg->getSubSpaceId();
		ResponsibleNodeData* respNode = findResponsibleNodeDataById( subSpaceId );
		if( respNode != NULL) {
			// delete the player from subspace
			magicLog->record(subSpaceId, 3, respNode->getPlayerListSize() );
			magicLog->record(subSpaceId, 6, respNode->getAllIntNodes().size() );
			// inform Intermediate Nodes
			sendTreeMessages( respNode, respNode->unsubscribe( unsubscribeMsg->getPlayerId() ) );
			// and send a response		
			unsubscribeMsg->setServiceType( RESPONSE );
			sendMsg(unsubscribeMsg, unsubscribeMsg->getSrcip());
		}
		else { 
			delete unsubscribeMsg; 
		}
	}
	else if ( unsubscribeMsg->getServiceType() == RESPONSE) {
		int subSpaceId = unsubscribeMsg->getSubSpaceId();
		if( findSubSpace( subSpaceId ) ) {
			if( getSubSpace( subSpaceId )->getRespIp() == unsubscribeMsg->getSrcip() ) {
				// we're successfull unsubscribed from the subspace so delete it
				removeSubSpace( subSpaceId  );
				if( debug) std::cout<< " Unsubcribed from SubSpace: " << unsubscribeMsg->getSubSpaceId() << "\n";
				deleteFromDupedMessages( unsubscribeMsg );
			}
		}
		delete unsubscribeMsg;
	}
	else if ( unsubscribeMsg->getServiceType() == FORWARD ) {
		IntermediateNodeData* inode = findIntermediateNodeDataById(unsubscribeMsg->getSubSpaceId() );
		if( inode != NULL ){
			inode->removeIp( unsubscribeMsg->getPlayerIp() );
		}
		unsubscribeMsg->setServiceType(CONFIRMATION);
		sendMsg( unsubscribeMsg, unsubscribeMsg->getSrcip() );
	}
	else { //Confirmation
		deleteFromDupedMessages( unsubscribeMsg );
		delete unsubscribeMsg;
	}
}

void YMMOG::processPublish(PublishMessage* publishMsg) {
/**
	This function is one way
**/
	receivedEvents++;
	if ( publishMsg->getServiceType() == REQUEST ) {
		ResponsibleNodeData* responsibleNode = findResponsibleNodeDataById( publishMsg->getSubSpaceId() );
		if( responsibleNode != NULL ) {
			responsibleNode->addEvent( publishMsg->getPublishedEvent() ) ;
			delete publishMsg;
		}
		else { delete publishMsg; } // some error handling missing
	}
}



void YMMOG::processPublishEventList(PublishEventListMessage* publishMsg) {
/** 
	This function is called everytime a PublishEventList message arrives
	If we're holding the subspace its for gameclient
	if we're holding the intermediateNode we have to forward it.
	Both is possible, too.

	We only use 1-Way ServiceType's to decrease the messages.
	if servicetyp = request then message comes from responsibleNode directly
	forward = at least one intermediate Node was on the way to us
	If a Node miss one it has to ask for it by itself.
**/
	receivedEventLists++;

	IPAddress backupIp;
	int subSpaceId = publishMsg->getSubSpaceId();
	int timeSlot = publishMsg->getTimeSlot();
	std::vector<Event> eventList;
	for(unsigned int i=0; i < publishMsg->getEventListArraySize(); ++i) {
		eventList.push_back( publishMsg->getEventList(i) );
	}

	if( findSubSpace( subSpaceId ) ) {
		/** We are holding the subSpace as Player
			forward it to the GameClient **/
		if( (timeSlot == (getSubSpace(subSpaceId)->getTimeSlot() + 1) || (getSubSpace(subSpaceId)->getTimeSlot() == -1) ) && publishMsg->getSender() == getSubSpace(subSpaceId)->getRespIp() ) {
			// We received the correct next timeStamp
			getSubSpace(subSpaceId)->setTimeSlot( timeSlot );
			ClientEventList* clientEventList = new ClientEventList("ClientEventList");
			clientEventList->setCommand(PUBLISHEVENTLIST);
			clientEventList->setEventListArraySize(publishMsg->getEventListArraySize());
			magicLog->record(subSpaceId,10, simTime() - publishMsg->getTimeStamp() );
			for(unsigned int i=0; i < eventList.size(); ++i) {
				magicLog->record(-1,9, simTime() - eventList.at(i).getTimeStamp() );
				clientEventList->setEventList(i, eventList.at(i));
			}
			send(clientEventList, "to_app");
		}
		else if( publishMsg->getTimeSlot() > (getSubSpace(subSpaceId)->getTimeSlot() + 1 )) {
			// We missed at least one TimeSlot request it
			/** function missing **/
		}
	}
	
	IntermediateNodeData* intermediateNode = findIntermediateNodeDataById( subSpaceId );
	if( intermediateNode != NULL ) {
		// We seem to have an Intermediate Node and we have to forward the Eventlists
		if( timeSlot > intermediateNode->getTimeSlot() || intermediateNode->getTimeSlot() == -1) {
			intermediateNode->setTimeSlot( timeSlot );
			backupIp = intermediateNode->getBackupIp();
			std::list<IPAddress> ips = intermediateNode->getIps();
			std::list<IPAddress>::iterator forwardTo;
			for( forwardTo = ips.begin() ; forwardTo != ips.end() ; ++forwardTo) {
				PublishEventListMessage* clone =  new PublishEventListMessage(*publishMsg);
				clone->setServiceType(FORWARD);
				sendMsg(clone, *forwardTo);
			}
			if( backupIp != thisNode.ip.get4() && publishMsg->getServiceType() == FORWARD) {
				// Compute the Delay between the last Node and us [as long lastNode is not ResponsibleNodeData]:
				if ( intermediateNode->addDelay( simTime() - publishMsg->getSent() ) ) {
					sendDelay( subSpaceId, intermediateNode->getDelay(), publishMsg->getSrcip(), backupIp);
				}
			}
		}
	}
	if( !findSubSpace( subSpaceId ) && intermediateNode == NULL ) {
		if( !findOpenUnSubscribe( subSpaceId, publishMsg->getSender() ) ) {
			UnSubscribeMessage* unsubscribeMsg = new UnSubscribeMessage("Unsubscribe");
			unsubscribeMsg->setCommand(UNSUBSCRIBE);
			unsubscribeMsg->setServiceType(REQUEST);
			unsubscribeMsg->setSubSpaceId( subSpaceId );
			unsubscribeMsg->setPlayerId( player_id) ;
			unsubscribeMsg->setPlayerIp( thisNode.ip.get4() );
			sendMsg(unsubscribeMsg, publishMsg->getSender() );
		}
		if( publishMsg->getSender() != publishMsg->getSrcip() && !findOpenUnSubscribe( subSpaceId, publishMsg->getSrcip() ) ) {
			// Send the intermediate Node one unsubscribe too
			UnSubscribeMessage* unsubscribeMsg2 = new UnSubscribeMessage("Unsubscribe");
			unsubscribeMsg2->setCommand(UNSUBSCRIBE);
			unsubscribeMsg2->setServiceType(FORWARD);
			unsubscribeMsg2->setSubSpaceId( subSpaceId );
			unsubscribeMsg2->setPlayerIp( thisNode.ip.get4() );
			sendMsg(unsubscribeMsg2, publishMsg->getSrcip() );
		}
	}
		 
	delete publishMsg;
}
void YMMOG::sendDelay(int subSpaceId, double lastDelay, IPAddress fromIp, IPAddress destip) {
	DelayMessage* delayMsg = new DelayMessage("delayMsg"); 
	delayMsg->setCommand(DELAY);
	delayMsg->setDelay( lastDelay );
	delayMsg->setSubSpaceId( subSpaceId );
	delayMsg->setFromIp( fromIp );
	sendMsg(delayMsg, destip);
}
	
void YMMOG::processCreate(CreateNodeMessage* createMsg) {
/**
	This function is called everytime a create message arrives.
	Response is handled by LobbyServer
**/
	receivedCreates++;
	//Welcher Type von Node soll erstellt werden?
	if ( createMsg->getServiceType() == REQUEST ) {
		if( createMsg->getJobType() == DELETENODE) {
			if( createMsg->getNodeType() == RESPONSIBLENODE ) {
				int subSpaceId = createMsg->getSubSpaceToCreate().getId();
				createMsg->setServiceType(RESPONSE);
				createMsg->setServiceType(false);
				for( i = responsible_nodes.begin(); i != responsible_nodes.end(); ++i ) {
					if( (*i)->getSubSpace().getId() == subSpaceId ) {
						delete *i;
						responsible_nodes.erase(i);
						createMsg->setCreationSuccessfull(true);
						break;
					}
				}
				sendMsg( createMsg, createMsg->getSrcip() );
				return;
			}
			if( createMsg->getNodeType() == BACKUPNODE ) {
				int subSpaceId = createMsg->getSubSpaceToCreate().getId();
				BackupNodeData* backupNode = findBackupNodeDataById( subSpaceId );
				if ( backupNode != NULL ) {
					if( logLevel > 0 ) nodeLog.push_back( NodeLogEntry( simTime(), (char)12, subSpaceId, createMsg->getSrcip(), backupNode->getSubSpace().getRespIp(), backupNode->lastMessage, true) );
					for( b = backup_nodes.begin(); b != backup_nodes.end(); ++b) {
						if( b->getId() == subSpaceId) {
							//Clean up OpenPings
							std::list<BackupNodeData>::iterator b2;
							int count=-1;
							//check if we can delete the ping (only this BackupNodeData is pinging this ip)
							for( b2= backup_nodes.begin(); b2 != backup_nodes.end(); ++b2) {
								if( b2->getSubSpace().getRespIp() == b->getSubSpace().getRespIp() ) count++;
							}
							if( count < 1 ) {
								deleteFromOpenPings(b->getSubSpace().getRespIp() );
							}
							//Delete BackupNodeData
							backup_nodes.erase(b);
							break;
						}
					}
					//delete backupNode;
					createMsg->setServiceType(RESPONSE);
					sendMsg(createMsg, createMsg->getSrcip());
					return;
				} else {
					if( logLevel > 0 ) nodeLog.push_back( NodeLogEntry(simTime(), (char)12, subSpaceId, createMsg->getSrcip(), thisNode.ip.get4(), -1, false) );
					delete createMsg;
					return;
				}
			}
			else if ( createMsg->getNodeType() == INTERMEDIATENODE ) {
				for(in = intermediate_nodes.begin(); in != intermediate_nodes.end(); ++in) {
					if( in->getSubSpaceId() == createMsg->getSubSpaceToCreate().getId() ) {
						intermediate_nodes.erase( in);
						break;
					}
				}
				createMsg->setServiceType(RESPONSE);
				createMsg->setCreationSuccessfull(true);
				sendMsg(createMsg, lobbyIp);
				return;
			}
		}
		switch ( createMsg->getNodeType() ) {
			case RESPONSIBLENODE : { // Responsible Node
				//Create new ResponsibleNodeData
				if ( findResponsibleNodeDataById( createMsg->getSubSpaceToCreate().getId() ) != NULL ) {
					//We already have created the SubSpace... our Response seems to be dropped, resend it
					createMsg->setServiceType(RESPONSE);
					createMsg->setCreationSuccessfull(true);
					if( logLevel > 0 ) nodeLog.push_back( NodeLogEntry(simTime(), (char)1, createMsg->getSubSpaceToCreate().getId(), createMsg->getSrcip(), thisNode.ip.get4(), false) );
					sendMsg(createMsg, createMsg->getSrcip());
					break;
				}
				SubSpace toCreate = createMsg->getSubSpaceToCreate();
				toCreate.setRespIp(thisNode.ip.get4() );
				responsible_nodes.push_back(new ResponsibleNodeData( toCreate , maxPlayerLeafs, treeWidth, (int) simTime() ));
				ResponsibleNodeData* respNode = findResponsibleNodeDataById(createMsg->getSubSpaceToCreate().getId() );
				if( respNode != NULL ) {
					createMsg->setCreationSuccessfull(true);
				}
				// Send Response				
				createMsg->setServiceType(RESPONSE);
				sendMsg(createMsg, createMsg->getSrcip());
				if( logLevel > 0 ) nodeLog.push_back( NodeLogEntry( simTime(), (char)1, createMsg->getSubSpaceToCreate().getId(), createMsg->getSrcip(), thisNode.ip.get4(), true) );
				break;
			}
			case BACKUPNODE: {
				int subSpaceId = createMsg->getSubSpaceToCreate().getId();
				if( findBackupNodeDataById( subSpaceId ) != NULL ) {
					// we already are the Backup Node for that SubSpace
					createMsg->setServiceType(RESPONSE);
					createMsg->setCreationSuccessfull(true);
					sendMsg( createMsg, createMsg->getSrcip() );
					if( logLevel > 0 ) nodeLog.push_back( NodeLogEntry(simTime(), (char)2, createMsg->getSubSpaceToCreate().getId(), createMsg->getSubSpaceToCreate().getRespIp(), thisNode.ip.get4(), false ) );
					break;
				}
				if( backup_nodes.size() == 0) {
					cancelEvent(delayTimer);
					scheduleAt(simTime() + delayDelay, delayTimer);
				}
				backup_nodes.push_back( BackupNodeData(createMsg->getSubSpaceToCreate() ) );
				if( logLevel > 0 ) nodeLog.push_back( NodeLogEntry( simTime(), (char)2, createMsg->getSubSpaceToCreate().getId(), createMsg->getSubSpaceToCreate().getRespIp(), thisNode.ip.get4(), true ) );
				// Request PlayerList from ResponsibleNodeData
				PlayerListMessage* playerListMsg = new PlayerListMessage("InitialPlayerListMessage");
				playerListMsg->setCommand(PLAYERLIST);
				playerListMsg->setServiceType(FORWARD);
				playerListMsg->setSubSpaceId( subSpaceId );
				sendMsg(playerListMsg,  createMsg->getSubSpaceToCreate().getRespIp() );
				// Begin pinging the responsibleNodes
				createMsg->setServiceType(RESPONSE);
				createMsg->setCreationSuccessfull(true);
				sendMsg(createMsg, createMsg->getSrcip());
				break;	
			}
			case INTERMEDIATENODE: {
				int subSpaceId = createMsg->getSubSpaceToCreate().getId();
				if( findIntermediateNodeDataById(subSpaceId) != NULL ) {
					createMsg->setServiceType(RESPONSE);
					createMsg->setCreationSuccessfull(false);
					sendMsg( createMsg, createMsg->getSrcip() );
					if( logLevel > 0 ) nodeLog.push_back( NodeLogEntry(simTime(), (char)3, createMsg->getSubSpaceToCreate().getId(), createMsg->getSubSpaceToCreate().getRespIp(), thisNode.ip.get4() , false) );
					break;
				}
				if( logLevel > 0 ) nodeLog.push_back( NodeLogEntry(simTime(), (char)1, createMsg->getSubSpaceToCreate().getId(), createMsg->getSubSpaceToCreate().getRespIp(), thisNode.ip.get4(), true) );
				intermediate_nodes.push_back( IntermediateNodeData( createMsg->getSubSpaceToCreate() ) );
				// Tell Responsible Node that we're ready 
				RequestIpMessage* requestMsg = new RequestIpMessage("FinishIntermediateNodeDataCreation");
				requestMsg->setCommand(REQUESTIP);
				requestMsg->setServiceType(REQUEST);
				requestMsg->setSubSpaceId(subSpaceId);
				requestMsg->setNodeType(INTERMEDIATENODE);
				requestMsg->setRequestedIp(thisNode.ip.get4());
				sendMsg( requestMsg,  createMsg->getSubSpaceToCreate().getRespIp() );
				
				// respond to lobby Server
				createMsg->setServiceType(RESPONSE);
				createMsg->setCreationSuccessfull(true);
				sendMsg(createMsg, createMsg->getSrcip());
				break;
			}
			default: {
				delete createMsg;
				break;
			}	
		}
	}
	else  {
		deleteFromDupedMessages(createMsg);
		delete createMsg;
	} 
}

void YMMOG::processRequestIp(RequestIpMessage* requestIpMsg) {
	receivedRequestIps++;
	if (requestIpMsg->getServiceType() == REQUEST ) {
		if( requestIpMsg->getNodeType() == INTERMEDIATENODE ) {	
			ResponsibleNodeData* respNode = findResponsibleNodeDataById(requestIpMsg->getSubSpaceId() );
			if( respNode != NULL ) {
				sendTreeMessages( respNode, respNode->addIntermediateNode(requestIpMsg->getSrcip() ) );
			}
			requestIpMsg->setServiceType(RESPONSE);
			sendMsg(requestIpMsg, requestIpMsg->getSrcip() );
			if( debug ) std::cout<< "Got a ready Intermediate Node.. IP: " << requestIpMsg->getSrcip() << " \n";
		}
		if ( requestIpMsg->getNodeType() == RESPONSIBLENODE) {
			IPAddress responseIp = findIpThroughNeighbor(requestIpMsg->getSubSpaceId());
			if( responseIp != (IPAddress) "127.0.0.1") {
				requestIpMsg->setRequestedIp(responseIp);
				requestIpMsg->setServiceType(RESPONSE);
				sendMsg(requestIpMsg, requestIpMsg->getSrcip() );
			}
			else {
				requestIpMsg->setServiceType(RESPONSE);
				notReadyMessages.push_back(requestIpMsg);
				bool alreadyWaiting = false;
				std::list<YMMOGMessage*>::iterator d= dupedMessages.begin();
				while( d != dupedMessages.end() ) {
					if( (*d)->getCommand() == REQUESTIP ) {
						RequestIpMessage* rmsg = dynamic_cast<RequestIpMessage*>(*d);
						if( rmsg != NULL && rmsg->getSubSpaceId() == requestIpMsg->getSubSpaceId() && rmsg->getDestip() == lobbyIp ) {
							alreadyWaiting=true;
							break;
						}
					}
					++d;
				}
				if( ! alreadyWaiting ) {
 					requestIp(requestIpMsg->getSubSpaceId(), RESPONSIBLENODE);
				}
			}
		}
	}
	else if ( requestIpMsg->getServiceType() == RESPONSE ) {
		if( requestIpMsg->getNodeType() == INTERMEDIATENODE ) {
			deleteFromDupedMessages(requestIpMsg);
			delete requestIpMsg;
			return;
		}
		//check if we was respNode and have to forward it:
		std::list<YMMOGMessage*>::iterator w= notReadyMessages.begin();
		while( w != notReadyMessages.end() ) {
			if( (*w)->getCommand() == REQUESTIP ) { //check if we have requests waiting because we were a responsibleNode
				RequestIpMessage* rmsg = dynamic_cast<RequestIpMessage*>(*w);
				if( rmsg != NULL && rmsg->getSubSpaceId() == requestIpMsg->getSubSpaceId() && rmsg->getServiceType() == RESPONSE ) {
					rmsg->setRequestedIp(requestIpMsg->getRequestedIp());
					w = notReadyMessages.erase(w);
					sendMsg(rmsg, rmsg->getSrcip());
					
				}
				else { ++w; }
			}
			else if( (*w)->getCommand() == SUBSCRIBE ) { // check if we have own subscribes waiting
				SubscribeMessage* subscribeMsg = dynamic_cast<SubscribeMessage*>(*w);
				if( subscribeMsg != NULL && subscribeMsg->getSubSpaceId() == requestIpMsg->getSubSpaceId() ) {
					w = notReadyMessages.erase(w);
					sendMsg( subscribeMsg, requestIpMsg->getRequestedIp() );	
				}
				else { ++w; }
			}
			else { ++w; }
		}
		deleteFromDupedMessages( requestIpMsg );
		// Check if there are some old Messages
		std::list<YMMOGMessage*>::iterator i = dupedMessages.begin(); 
		while( i != dupedMessages.end() ) {
			if( (*i)->getCommand() == SUBSCRIBE) {
				SubscribeMessage* subscribeMsg = dynamic_cast<SubscribeMessage*>(*i);
				if( subscribeMsg != NULL && subscribeMsg->getSubSpaceId() == requestIpMsg->getSubSpaceId() && subscribeMsg->getDestip() != requestIpMsg->getRequestedIp() ) {
					i = dupedMessages.erase(i);
					sendMsg( subscribeMsg, requestIpMsg->getRequestedIp() );
				}
				else { ++i; }
			}
			else if( (*i)->getCommand() == UNSUBSCRIBE) {
				UnSubscribeMessage* unsubscribeMsg = dynamic_cast<UnSubscribeMessage*>(*i);
				if( unsubscribeMsg != NULL && unsubscribeMsg->getSubSpaceId() == requestIpMsg->getSubSpaceId() && unsubscribeMsg->getDestip() != requestIpMsg->getRequestedIp() ) {
					i = dupedMessages.erase(i);
					sendMsg( unsubscribeMsg, requestIpMsg->getRequestedIp() );
				}
				else { ++i; }
			}
			else { ++i; }
		}
		if( findResponsibleNodeDataById( requestIpMsg->getSubSpaceId() - gameSpaceSize - 1) != NULL) {
			findResponsibleNodeDataById( requestIpMsg->getSubSpaceId() - gameSpaceSize - 1)->setNeighbor(0,requestIpMsg->getRequestedIp());
		}
		else if( findResponsibleNodeDataById( requestIpMsg->getSubSpaceId() - gameSpaceSize) != NULL) {
			findResponsibleNodeDataById( requestIpMsg->getSubSpaceId() - gameSpaceSize)->setNeighbor(1,requestIpMsg->getRequestedIp());
		}		
		else if( findResponsibleNodeDataById( requestIpMsg->getSubSpaceId() - gameSpaceSize + 1) != NULL) {
			findResponsibleNodeDataById( requestIpMsg->getSubSpaceId() - gameSpaceSize +1)->setNeighbor(2,requestIpMsg->getRequestedIp());
		}
		else if( findResponsibleNodeDataById( requestIpMsg->getSubSpaceId() - 1) != NULL) {
			findResponsibleNodeDataById( requestIpMsg->getSubSpaceId() - 1)->setNeighbor(3,requestIpMsg->getRequestedIp());
		}
		else if( findResponsibleNodeDataById( requestIpMsg->getSubSpaceId() + 1) != NULL) {
			findResponsibleNodeDataById( requestIpMsg->getSubSpaceId() + 1)->setNeighbor(4,requestIpMsg->getRequestedIp());
		}
		else if( findResponsibleNodeDataById( requestIpMsg->getSubSpaceId() + gameSpaceSize - 1) != NULL) {
			findResponsibleNodeDataById( requestIpMsg->getSubSpaceId() + gameSpaceSize -1)->setNeighbor(5,requestIpMsg->getRequestedIp());
		}
		else if( findResponsibleNodeDataById( requestIpMsg->getSubSpaceId() + gameSpaceSize ) != NULL) {
			findResponsibleNodeDataById( requestIpMsg->getSubSpaceId() + gameSpaceSize)->setNeighbor(6,requestIpMsg->getRequestedIp());
		}
		else if( findResponsibleNodeDataById( requestIpMsg->getSubSpaceId() + gameSpaceSize + 1) != NULL) {
			findResponsibleNodeDataById( requestIpMsg->getSubSpaceId() + gameSpaceSize +1)->setNeighbor(7,requestIpMsg->getRequestedIp());
		}
		delete requestIpMsg;
	}

	else { 
		delete requestIpMsg;
	}
}

void YMMOG::processUpdateRespIp(UpdateRespIpMessage* updateMsg) {
/**
	This function is called everytime an updaterespip message arrives.
	-request -> we have to update our subSpace we're holding as responsibleNode and inform all players currently subscribed
	-response is handled by lobbyserver
	-forward -> we have to update our subSpace we're holding as playerNode
**/
	receivedUpdateRespIps++;
	if( updateMsg->getServiceType() == REQUEST ) {
		int subspaceid = updateMsg->getSubSpaceId();
		ResponsibleNodeData* respNode = findResponsibleNodeDataById(subspaceid);
		if( respNode != NULL ) {
			IPAddress updatedip = updateMsg->getUpdatedIp();
			int direction = updateMsg->getDirection();
			respNode->setNeighbor(direction, updatedip);
			if( debug) std::cout<< "Updating "<< respNode->getId() <<" "<< subspaceid << " with IPAddress: "<< updatedip << "\n";
			if( direction != -1 ) {
				/*//Inform all players
				std::list<PlayerNode> playerlist = respNode->getPlayerList();
				std::list<PlayerNode>::iterator pl;
				for(pl= playerlist.begin(); pl != playerlist.end(); ++pl) {
					if( pl->getIp() != thisNode.ip.get4() ) {
						UpdateRespIpMessage* forwardMsg = new UpdateRespIpMessage("Update Resp IP");
						forwardMsg->setCommand(UPDATERESPIP);
						forwardMsg->setServiceType(FORWARD);
						forwardMsg->setSubSpaceId(subspaceid);
						forwardMsg->setDirection(direction);
						forwardMsg->setUpdatedIp(updatedip);
						if(debug) std::cout<< "Sending Updated SubSpace "<< subspaceid << " with new IP: "<< updatedip << " at "<< direction << " to: " << pl->getIp() << "\n";
						sendMsg(forwardMsg, pl->getIp());
					}
				}*/
			}
			else {
				// Inform intermediateNodes from new BackupNodeData
				std::list<IPAddress> ips = respNode->getAllIntNodes();
				std::list<IPAddress>::iterator i;
				for(i = ips.begin(); i != ips.end(); ++i ) {
					UpdateRespIpMessage* forwardMsg = new UpdateRespIpMessage("Update Resp Backup IP");
					forwardMsg->setCommand(UPDATERESPIP);
					forwardMsg->setServiceType(FORWARD);
					forwardMsg->setSubSpaceId(subspaceid);
					forwardMsg->setDirection(direction);
					forwardMsg->setUpdatedIp(updatedip);
					if(debug) std::cout<< "Sending Updated SubSpace "<< subspaceid << " with new Backup IP: "<< updatedip << " at "<< direction << " to: " << (*i) << "\n";
					sendMsg(forwardMsg, (*i));
				}
			}
			updateMsg->setServiceType(CONFIRMATION);
			sendMsg(updateMsg, updateMsg->getSrcip() );
		}
		else {
			// we're not holding requestet subspace
			delete updateMsg;
		} 
	}
	else if( updateMsg->getServiceType() == CONFIRMATION ) {
		deleteFromDupedMessages( updateMsg );
		delete updateMsg;
	}
	else if( updateMsg->getServiceType() == FORWARD ) {
		SubSpace* subSpace = getSubSpace(updateMsg->getSubSpaceId());
		if( subSpace != NULL ) {
			subSpace->setNeighbor(updateMsg->getDirection(), updateMsg->getUpdatedIp());
		}
		if( updateMsg->getDirection() == -1) {
			IntermediateNodeData* intNode = findIntermediateNodeDataById( updateMsg->getSubSpaceId() );
			if( intNode != NULL ) {
				intNode->setBackupIp( updateMsg->getUpdatedIp() );
			}
		}
		updateMsg->setServiceType(CONFIRMATION);
		sendMsg(updateMsg, updateMsg->getSrcip());
	}
}

void YMMOG::processLogout(LogoutMessage* logoutMsg) {
	/** missing **/
}

void YMMOG::subscribe(int subSpaceId, IPAddress destip) {
/**
	This function is called whenever we want to subscribe to a new SubSpace.
	It creates an OpenSubscribe and sends it.
**/	
	if( !findOpenSubScribe( subSpaceId, destip) ) {
		SubscribeMessage* subscribeMsg = new SubscribeMessage("SubscribeToSubSpace");
		subscribeMsg->setCommand(SUBSCRIBE);
		subscribeMsg->setServiceType(REQUEST);
		subscribeMsg->setPlayerId( player_id );
		subscribeMsg->setSubSpaceId( subSpaceId );
		sendMsg(subscribeMsg, destip);
	}
}


void YMMOG::unsubscribe(int subSpaceId, IPAddress destip) {
/**
	This function is called whenever we want to unsubscribe from a SubSpace.
	It creates an OpenSubscribe and sends it.
**/	
	if ( !findOpenUnSubscribe(subSpaceId,destip) ) {
		UnSubscribeMessage* unsubscribeMsg = new UnSubscribeMessage("UnsubscribeFromSubSpace");
		unsubscribeMsg->setCommand(UNSUBSCRIBE);
		unsubscribeMsg->setServiceType(REQUEST);
		unsubscribeMsg->setPlayerId( player_id);
		unsubscribeMsg->setSubSpaceId( subSpaceId );
		sendMsg( unsubscribeMsg, destip );
	}
}

void YMMOG::unsubscribeAll() {
/**
	It simply calls unsubscribe for all subscribed SubSpaces
**/
	std::vector<SubSpace>::iterator i;
	for(i=subscribed_SubSpaces.begin() ; i != subscribed_SubSpaces.end(); ++i) {
		unsubscribe( (*i).getId() , (*i).getRespIp() );
	}
}

void YMMOG::removeSubSpace( int subSpaceId ) {
	std::vector<SubSpace>::iterator i;
	for(i=subscribed_SubSpaces.begin() ; i != subscribed_SubSpaces.end(); ++i) {
		if( i->getId() == subSpaceId) { 
			subscribed_SubSpaces.erase(i);
			break;
		}
	}
}

bool YMMOG::findSubSpace(int id) {
	std::vector<SubSpace>::iterator i;
	for(i=subscribed_SubSpaces.begin(); i != subscribed_SubSpaces.end() ; ++i ) {
		if( i->getId() == id ) return true;
	}
	return false;
}

SubSpace* YMMOG::getSubSpace(int id) {
	std::vector<SubSpace>::iterator i;
	for(i=subscribed_SubSpaces.begin(); i != subscribed_SubSpaces.end() ; ++i ) {
		if( i->getId() == id ) return &(*i);
	}
	return NULL;
}

void YMMOG::informNeighbors(SubSpace subSpace) {
	for( int i=0; i < 8; i++) {
		if( subSpace.getNeighbor(i) != (IPAddress) "127.0.0.1" ) {
			UpdateRespIpMessage* updateMsg = new UpdateRespIpMessage("Update Responsible Ip");
			updateMsg->setCommand(UPDATERESPIP);
			updateMsg->setServiceType(REQUEST);
			if( i < 3 ) { updateMsg->setSubSpaceId(subSpace.getId() - gameSpaceSize + i - 1 ); }
			else if( i == 3) { updateMsg->setSubSpaceId(subSpace.getId() - 1 ); }
			else if( i == 4) { updateMsg->setSubSpaceId(subSpace.getId() + 1 ); }
			else if (i >= 5) { updateMsg->setSubSpaceId(subSpace.getId() + gameSpaceSize + i - 6 ); }
			updateMsg->setDirection(7-i);
			updateMsg->setUpdatedIp(subSpace.getRespIp());
			sendMsg(updateMsg, subSpace.getNeighbor(i));
		}
	}
}
IPAddress YMMOG::findIpThroughNeighbor(int subSpaceId) {
	//
	IPAddress toSubscribe;
	ResponsibleNodeData* respNode = findResponsibleNodeDataById( subSpaceId );
	if ( respNode != NULL ) { return respNode->getSubSpace().getRespIp(); }
	if ( findSubSpace(subSpaceId - gameSpaceSize -1) ) {
		toSubscribe = getSubSpace(subSpaceId - gameSpaceSize -1)->getNeighbor(SE);
		if( toSubscribe != (IPAddress) "127.0.0.1" ) return toSubscribe;	
	}
	if ( findSubSpace(subSpaceId - gameSpaceSize) ) {
		toSubscribe = getSubSpace(subSpaceId - gameSpaceSize)->getNeighbor(S);
		if( toSubscribe != (IPAddress) "127.0.0.1" ) return toSubscribe;	
	}
	if ( findSubSpace(subSpaceId - gameSpaceSize + 1) ) {
		toSubscribe = getSubSpace(subSpaceId - gameSpaceSize + 1)->getNeighbor(SW);
		if( toSubscribe != (IPAddress) "127.0.0.1" ) return toSubscribe;	
	}
	if ( findSubSpace(subSpaceId - 1) ) {
		toSubscribe = getSubSpace(subSpaceId - 1)->getNeighbor(E);
		if( toSubscribe != (IPAddress) "127.0.0.1" ) return toSubscribe;	
	}
	if ( findSubSpace(subSpaceId + 1) ) {
		toSubscribe = getSubSpace(subSpaceId + 1)->getNeighbor(W);
		if( toSubscribe != (IPAddress) "127.0.0.1" ) return toSubscribe;	
	}
	if ( findSubSpace(subSpaceId + gameSpaceSize - 1) ) {
		toSubscribe = getSubSpace(subSpaceId + gameSpaceSize - 1)->getNeighbor(NE);
		if( toSubscribe != (IPAddress) "127.0.0.1" ) return toSubscribe;	
	}
	if ( findSubSpace(subSpaceId + gameSpaceSize ) ) {
		toSubscribe = getSubSpace(subSpaceId + gameSpaceSize)->getNeighbor(N);
		if( toSubscribe != (IPAddress) "127.0.0.1" ) return toSubscribe;	
	}
	if ( findSubSpace(subSpaceId + gameSpaceSize + 1) ) {
		toSubscribe = getSubSpace(subSpaceId + gameSpaceSize + 1)->getNeighbor(NW);
		if( toSubscribe != (IPAddress) "127.0.0.1" ) return toSubscribe;	
	}
	// We haven't found the ip in our neighbors so return dummy.
	return (IPAddress) "127.0.0.1";
}

void YMMOG::requestIp(int subSpaceId, int nodeType) {
	std::list<YMMOGMessage*>::iterator i=dupedMessages.begin();
	while( i != dupedMessages.end() ) {
		if( (*i)->getCommand() == REQUESTIP ) {
			RequestIpMessage* rmsg = dynamic_cast<RequestIpMessage*>(*i);
			if( rmsg != NULL && rmsg->getSubSpaceId() == subSpaceId && rmsg->getDestip() == lobbyIp ) {
				return;
			}
			else { ++i; }
		}
		else { ++i; }
	}	
	RequestIpMessage* requestMsg = new RequestIpMessage("RequestIpForNode");
	requestMsg->setCommand(REQUESTIP);
	requestMsg->setServiceType(REQUEST);
	requestMsg->setSubSpaceId( subSpaceId );
	requestMsg->setNodeType( nodeType );
	sendMsg( requestMsg, lobbyIp );
}

void YMMOG::requestIpFromNeighbor(int neighborId, int requestedSubSpace) {
	std::list<YMMOGMessage*>::iterator i=dupedMessages.begin();
	while( i != dupedMessages.end() ) {
		if( (*i)->getCommand() == REQUESTIP ) {
			RequestIpMessage* rmsg = dynamic_cast<RequestIpMessage*>(*i);
			if( rmsg != NULL && rmsg->getSubSpaceId() == requestedSubSpace && rmsg->getDestip() ==  getSubSpace(neighborId)->getRespIp() ) {
				return;
			}
			else { ++i; }
		}
		else { ++i; }
	}
	RequestIpMessage* requestMsg = new RequestIpMessage("RequestIpThroughNeighbor");
	requestMsg->setCommand(REQUESTIP);
	requestMsg->setServiceType(REQUEST);
	requestMsg->setSubSpaceId( requestedSubSpace );
	requestMsg->setNodeType( RESPONSIBLENODE );
	sendMsg( requestMsg, getSubSpace(neighborId)->getRespIp() );
}

int YMMOG::findNeighborSubSpace(int subSpaceId) {
	if( findSubSpace(subSpaceId - gameSpaceSize -1) ) return subSpaceId - gameSpaceSize -1;
	if( findSubSpace(subSpaceId - gameSpaceSize ) ) return subSpaceId - gameSpaceSize;
	if( findSubSpace(subSpaceId - gameSpaceSize +1) ) return subSpaceId - gameSpaceSize +1;
	if( findSubSpace(subSpaceId - 1) ) return subSpaceId -1;
	if( findSubSpace(subSpaceId + 1) ) return subSpaceId +1;
	if( findSubSpace(subSpaceId + gameSpaceSize -1) ) return subSpaceId + gameSpaceSize -1;
	if( findSubSpace(subSpaceId + gameSpaceSize ) ) return subSpaceId + gameSpaceSize;
	if( findSubSpace(subSpaceId + gameSpaceSize +1) ) return subSpaceId + gameSpaceSize +1;
	return -1;	
}
	

void YMMOG::checkSubscriptions() {
	lastCheckedX = xpos;
	lastCheckedY = ypos;
	// Find all SubSpaces in Subscription Range that are not subscribed yet and subscribe to them
	if ( debug ) {
		std::cout<< "We're subscribed to " << subscribed_SubSpaces.size() << " SubSpaces yet\n";
	}
	int xbegin = (xpos - subscriptionRange) / subSpaceSize;
	if ( xbegin < 0 ) xbegin = 0;
	int xend  = (xpos + subscriptionRange) / subSpaceSize;
	if ( xend >= gameSpaceSize ) xend = gameSpaceSize - 1;
	
	int ybegin = (ypos - subscriptionRange) / subSpaceSize;
	if ( ybegin < 0 ) ybegin = 0;
	int yend = (ypos + subscriptionRange) / subSpaceSize;
	if ( yend >= gameSpaceSize ) yend = gameSpaceSize - 1;
	for( int i= xbegin; i <= xend; i++) {
		for (int k= ybegin; k <= yend; k++) {
			// Is SubSpace already subscribed?
			if ( ! findSubSpace( k*gameSpaceSize + i ) ) {
				//No then lets Subscribe..
				//check if we already have an open Subscribe we're waiting for:
				bool alreadyWaitingForSubscribe=false;
				std::list<YMMOGMessage*>::iterator w= notReadyMessages.begin();
				while( w != notReadyMessages.end() ) {
					if( (*w)->getCommand() == SUBSCRIBE ) {
						SubscribeMessage* subscribeMsg = dynamic_cast<SubscribeMessage*>(*w);
						if( subscribeMsg != NULL && subscribeMsg->getSubSpaceId() == k*gameSpaceSize + i ) {
							alreadyWaitingForSubscribe = true;
							break;
						}
					}
					++w;
				}
				w = dupedMessages.begin();
				while( w != dupedMessages.end() ) {
					if( (*w)->getCommand() == SUBSCRIBE ) {
						SubscribeMessage* subscribeMsg = dynamic_cast<SubscribeMessage*>(*w);
						if( subscribeMsg != NULL && subscribeMsg->getSubSpaceId() == k*gameSpaceSize + i ) {
							alreadyWaitingForSubscribe = true;
							break;
						}
					}
					++w;
				}
				if( !alreadyWaitingForSubscribe ) {
					SubscribeMessage* subscribeMsg = new SubscribeMessage("Subscribe");
					subscribeMsg->setCommand(SUBSCRIBE);
					subscribeMsg->setServiceType(REQUEST);
					subscribeMsg->setPlayerId(player_id);
					subscribeMsg->setPlayerIp(thisNode.ip.get4());
					subscribeMsg->setSubSpaceId(k*gameSpaceSize + i);
					if( findResponsibleNodeDataById(k*gameSpaceSize+i) != NULL ) {
						sendMsg(subscribeMsg, thisNode.ip.get4() );
					}
					else {
						notReadyMessages.push_back(subscribeMsg);
					}
					int sid = findNeighborSubSpace( k*gameSpaceSize + i );
					if( sid != -1 ) {
						requestIpFromNeighbor(sid, k*gameSpaceSize + i);
					}
					else {
							requestIp( k*gameSpaceSize + i, RESPONSIBLENODE);
					}
				}
			}
		}
	}
	// Do we have to unsubscribe some?
	std::vector<SubSpace>::iterator i;
	for(i=subscribed_SubSpaces.begin(); i != subscribed_SubSpaces.end() ; ++i ) {
		if(  i->getId() < ((ypos - unsubscriptionRange) / subSpaceSize)*gameSpaceSize  ) {
			unsubscribe(i->getId(), i->getRespIp() );
		}
		else if ( i->getId() % gameSpaceSize < ((xpos - unsubscriptionRange) / subSpaceSize) ) {
			unsubscribe(i->getId(), i->getRespIp() );
		}
		else if( i->getId() % gameSpaceSize > ((xpos + unsubscriptionRange) / subSpaceSize) ) {
			unsubscribe(i->getId(), i->getRespIp() );
		}
		else if( i->getId() > ( (( (ypos + unsubscriptionRange) / subSpaceSize) +1) )*gameSpaceSize ) {
			unsubscribe(i->getId(), i->getRespIp() );
		}
	}
}

void YMMOG::processPlayerList(PlayerListMessage* playerListMsg) {
	receivedPlayerLists++;
	if(playerListMsg->getServiceType() == REQUEST ) {
		BackupNodeData* backupNode = findBackupNodeDataById( playerListMsg->getSubSpaceId() );
		if( backupNode != NULL) {
			backupNode->setBackupDelay( simTime() - playerListMsg->getTimeStamp() );
			
			if( playerListMsg->getAction() == COMPLETE ) {
				std::list<TreeNode> tmp;
				uint i = 0;
				while ( i < playerListMsg->getTreeArraySize() ) {
					tmp.push_back(  playerListMsg->getTree(i) );
					i++;
				}
				backupNode->setTree( tmp );
			}
			
			if( pings.size() == 0) {
				cancelEvent(pingTimer);
				scheduleAt(simTime() + pingDelay, pingTimer);
			}
			if( !findOpenPing( backupNode->getSubSpace().getRespIp()) ) {
				PingMessage* pingMsg = new PingMessage("BackupNodeDataPing");
				pingMsg->setCommand(PING);
				pingMsg->setServiceType(REQUEST);
				pingMsg->setDestip( backupNode->getSubSpace().getRespIp() );
				pingMsg->setSrcip(thisNode.ip.get4() );
				pingMsg->setFailed(0);
				pingMsg->setTimeStamp(simTime());
				pings.push_back(pingMsg);
				sendPing( pingMsg );
			}

			deleteFromDupedMessages( playerListMsg );
			backupNode->lastMessage=simTime();
			playerListMsg->setServiceType(CONFIRMATION);
			sendMsg(playerListMsg, playerListMsg->getSrcip() );
		}
		else {
			if (debug) {
				std::cout<< "WARNING! We got a playerList for subSpace: " << playerListMsg->getSubSpaceId() << " from " << playerListMsg->getSrcip() << " but we're not holding the backupNode. I'm " << thisNode.ip.get4() << "\n";
			}
			//delete playerListMsg;
			playerListMsg->setServiceType(CONFIRMATION);
			sendMsg(playerListMsg, playerListMsg->getSrcip() );

		}
	}
	else if ( playerListMsg->getServiceType() == CONFIRMATION ) {
		deleteFromDupedMessages( playerListMsg );
		delete playerListMsg;
	}
	else if ( playerListMsg->getServiceType() == FORWARD ) {
		if( debug) std::cout<< "Received playerListMsg FORWARD from: " << playerListMsg->getSrcip() << " SubSpaceId: " << playerListMsg->getSubSpaceId() << " . I'm " << thisNode.ip.get4() << "\n";
		
		ResponsibleNodeData* respNode = findResponsibleNodeDataById(playerListMsg->getSubSpaceId());
		deleteFromDupedMessages( playerListMsg );
		if( respNode != NULL ) {
			int subSpaceId = playerListMsg->getSubSpaceId();
			respNode->setBackupIp(playerListMsg->getSrcip());
			sendPlayerList(COMPLETE, subSpaceId);			
			//Inform all IntermediateNodeDatas about the new Backup node
			std::list<IPAddress> playerlist = respNode->getAllIntNodes();
			std::list<IPAddress>::iterator pl;
			for(pl= playerlist.begin(); pl != playerlist.end(); ++pl) {
				UpdateRespIpMessage* forwardMsg = new UpdateRespIpMessage("Update Resp IP");
				forwardMsg->setCommand(UPDATERESPIP);
				forwardMsg->setServiceType(FORWARD);
				forwardMsg->setSubSpaceId(subSpaceId);
				forwardMsg->setDirection(-1);
				forwardMsg->setUpdatedIp(playerListMsg->getSrcip());
				sendMsg(forwardMsg, *pl);
			}
		}
		else {
			if ( debug) std::cout<< "WARNING! We got request to send PlayerList to backupNode " << playerListMsg->getSrcip() << " but we 're not holding RespNode Id: " << playerListMsg->getSubSpaceId() << " I'm : " << thisNode.ip.get4() << "\n";
		}
		delete playerListMsg;
	}
}



void YMMOG::processReplacement(ReplacementMessage* replaceMsg) {
	receivedReplacements++;
	if( replaceMsg->getServiceType() == REQUEST) {
		SubSpace* subSpace =  getSubSpace(replaceMsg->getSubSpaceId());
		if( subSpace != NULL ) {
			IPAddress newIp = replaceMsg->getSrcip();
			int subSpaceId = replaceMsg->getSubSpaceId();
			IPAddress oldRespIp = subSpace->getRespIp();
			subSpace->setRespIp( newIp );
			//check if there are some openUnsubscribes
			std::list<YMMOGMessage*>::iterator i= dupedMessages.begin();
			while( i != dupedMessages.end() ) {
				if( (*i)->getCommand() == UNSUBSCRIBE ) {
					UnSubscribeMessage* unsubscribeMsg = dynamic_cast<UnSubscribeMessage*>(*i);
					if( unsubscribeMsg != NULL && unsubscribeMsg->getSubSpaceId() == subSpaceId && unsubscribeMsg->getDestip() == oldRespIp && oldRespIp != newIp ) {
						unsubscribeMsg->setFailed(0);
						i = dupedMessages.erase(i);
						sendMsg(unsubscribeMsg, newIp);
					}
					else { ++i; }
				}
				else { ++i; }
			}
			if( replaceMsg->getSendGameState() == 1) {
				replaceMsg->setGameState( getSubSpace(replaceMsg->getSubSpaceId() )->getGameState() );
			}
			replaceMsg->setServiceType(CONFIRMATION);
			sendMsg(replaceMsg, replaceMsg->getSrcip() );
		}
		else {
			delete replaceMsg;
		}
	}
	else { // Confirmation
		if( replaceMsg->getSendGameState() == 1) {
			// Load GameState to responsibleNodes
			/** missing **/
		}
		delete replaceMsg;
	}
}

void YMMOG::processPlayerPingTimer() {
	if( responsible_nodes.size() > 0 ) {
		// All Pings still in playerPings are failed. So unsubscribe them from all ResponsibleNodeDatas 
		std::list<PingMessage*>::iterator p = playerPings.begin();
		while( playerPings.size() > 0 ) {
			IPAddress failedIp = playerPings.front()->getDestip();
			for( i = responsible_nodes.begin(); i != responsible_nodes.end() ; ++i) {
				sendTreeMessages( *i, (*i)->unsubscribe( failedIp ) );
			}
			delete playerPings.front();
			playerPings.pop_front();
		}
		// Fill playerPings with new Ips
		for( i = responsible_nodes.begin(); i != responsible_nodes.end(); ++i ) {
			std::list<IPAddress> ipList = (*i)->getNodeList();
			while( ipList.size() > 0 ) {
				IPAddress toInsert = ipList.front();
				bool alreadyInserted = false;
				for( p = playerPings.begin(); p != playerPings.end(); ++p) {
					if( (*p)->getDestip() == toInsert ) {
						alreadyInserted = true;
						break;
					}
				}
				if( !alreadyInserted) {
					PingMessage* pingMsg = new PingMessage("PlayerPing");
					pingMsg->setCommand(PING);
					pingMsg->setServiceType(REQUEST);
					pingMsg->setDestip(toInsert);
					pingMsg->setFailed(0);
					pingMsg->setTimeStamp(simTime());
					playerPings.push_back(pingMsg);
				}
				ipList.pop_front();
			}
		}
		//send the pings:
		for( p = playerPings.begin(); p != playerPings.end(); ++p ) {
			PingMessage* cloneMsg = new PingMessage(*(*p));
			sendMsg(cloneMsg, cloneMsg->getDestip() );
		}
	}
	cancelEvent(playerPingTimer);
	scheduleAt(simTime() + 20, playerPingTimer);
}

void YMMOG::processPingTimer() {
	selfMessages++;
	cancelEvent(pingTimer);
	std::list<PingMessage*>::iterator p= pings.begin();
	double next=simTime();
	while( p != pings.end() ) {
		if( ( (*p)->getTimeStamp() + pingDelay) < simTime() ) {
			(*p)->setTimeStamp(simTime());
			(*p)->setFailed( (*p)->getFailed() + 1 );
			if( (*p)->getFailed() > maxFailedPings ) {
				// A Responsible Node failed. Lets replace it
				b = backup_nodes.begin();
				while( b != backup_nodes.end() ) {
					//There could be more then one ResponsibleNodeData with that ip
					if( b->getSubSpace().getRespIp() == (*p)->getDestip() ) {
						// We are the responsible Node now for that subspace
						if( logLevel > 0 ) { nodeLog.push_back( NodeLogEntry( simTime(), 27,  b->getSubSpace().getId(), b->getSubSpace().getRespIp(), (*p)->getDestip(), (int) b->getTree().size(), true ) ); }
						
						b->getSubSpace().setRespIp( thisNode.ip.get4() );
						
						ResponsibleNodeData* respNode = new ResponsibleNodeData( b->getSubSpace() , maxPlayerLeafs, treeWidth , simTime() - maxFailedPings * pingDelay ,  b->getTree() );
						int sendGameState = 1;
						std::list<IPAddress> toInform = respNode->getNodeList();
						std::list<IPAddress>::iterator tI;
						respNode->setBackupIp( (IPAddress) "127.0.0.1" );
						int t=0;
						for(tI = toInform.begin(); tI != toInform.end(); ++tI) {
							// Inform All Players and subscribe them to the new Resp Node	
							ReplacementMessage* replaceMsg = new ReplacementMessage("ReplaceResponsibleNodeData");
							replaceMsg->setCommand(REPLACEMENT);
							replaceMsg->setServiceType(REQUEST);
							replaceMsg->setSubSpaceId( b->getSubSpace().getId() );
							replaceMsg->setSendGameState( sendGameState );
							if( sendGameState == 1) {			
								sendGameState = 0;
							}
							sendMsg(replaceMsg, *tI );t++;
						}
						responsible_nodes.push_back( respNode );
						//Inform LobbyServer
						ReplacementMessage* replaceMsg = new ReplacementMessage("ReplaceResponsibleNodeData");
						replaceMsg->setCommand(REPLACEMENT);
						replaceMsg->setServiceType(REQUEST);
						replaceMsg->setSubSpaceId( b->getSubSpace().getId() );
						replaceMsg->setSendGameState( sendGameState );
						sendMsg(replaceMsg, lobbyIp);
						//Inform Neighbors
						informNeighbors(respNode->getSubSpace() );
						// delete BackupNodeData 
						b = backup_nodes.erase( b);
					}
					else { ++b; }
				}
				delete *p;
				p = pings.erase(p);
			} else {
				if( next > (*p)->getTimeStamp() ) next = (*p)->getTimeStamp();
				sendPing( (*p) );
				++p;
			}
		}
		else { ++p; }
	}
	
	//resend playerPings if there are some
	p = playerPings.begin();
	while( p != playerPings.end() ) {
		if( ( (*p)->getTimeStamp() + pingDelay) < simTime() ) {
			(*p)->setTimeStamp(simTime());
			(*p)->setFailed( (*p)->getFailed() + 1 );
			if( (*p)->getFailed() > maxFailedPings ) {
				IPAddress failedIp = (*p)->getDestip();
				for( i = responsible_nodes.begin(); i != responsible_nodes.end() ; ++i) {
					sendTreeMessages( *i, (*i)->unsubscribe( failedIp ) );
				}
				delete *p;
				p = playerPings.erase(p);
			}
			else {
				PingMessage* cloneMsg = new PingMessage(*(*p));
				sendMsg(cloneMsg, cloneMsg->getDestip());
				++p;
			}
		}
		else {
			++p;
		}
	}
	scheduleAt( next + pingDelay, pingTimer);
}

void YMMOG::processRequestNode(RequestNodeMessage* requestMsg) {
	receivedRequestNodes++;
	if( requestMsg->getServiceType() == CONFIRMATION ) {
		deleteFromDupedMessages( requestMsg );
		delete requestMsg;
	}
}

void YMMOG::requestIntermediateNode(int subSpaceId) {
	RequestNodeMessage* requestMsg = new RequestNodeMessage("RequestingIntermediateNodeData");
	requestMsg->setCommand(REQUESTNODE);
	requestMsg->setServiceType(REQUEST);
	requestMsg->setSubSpaceId( subSpaceId);
	requestMsg->setNodeType( INTERMEDIATENODE );
	sendMsg( requestMsg, lobbyIp );
}


void YMMOG::sendTreeMessages(ResponsibleNodeData* respNode, std::list<TreeMessage> treeMessages ) {
	std::list<TreeMessage> msgList = treeMessages;
	std::list<TreeMessage>::iterator m ;
	for ( m = msgList.begin(); m != msgList.end() ; ++m) {
		if( m->type == 0 ) {
			// Forward Subscribe to IntermediateNodeData
			SubscribeMessage* subscribeMsg = new SubscribeMessage("ForwardSubScribeToIntermediateNodeData");
			subscribeMsg->setCommand(SUBSCRIBE);
			subscribeMsg->setServiceType(FORWARD);
			subscribeMsg->setPlayerId(-1);
			subscribeMsg->setPlayerIp( m->playerip );
			subscribeMsg->setSubSpaceId( respNode->getId()  );
			sendMsg( subscribeMsg, m->destip );
		}
		else if( m->type == 1) {
			// Forward UnSubscribe to IntermediateNodeData
			UnSubscribeMessage* unsubscribeMsg = new UnSubscribeMessage("ForwardUnSubscribeToIntermediateNodeData");
			unsubscribeMsg->setCommand(UNSUBSCRIBE);
			unsubscribeMsg->setServiceType(FORWARD);
			unsubscribeMsg->setPlayerId(-1);
			unsubscribeMsg->setPlayerIp( m->playerip );
			unsubscribeMsg->setSubSpaceId( respNode->getId() );
			sendMsg( unsubscribeMsg, m->destip);
		}
		else if ( m->type == 2 ) {
			// request a new IntermediateNode
			requestIntermediateNode(respNode->getId() );
		}
		else if ( m->type == 3) {
			// delete IntermediateNodeData
			CreateNodeMessage* createMsg = new CreateNodeMessage("DeleteIntermediateNodeData");
			createMsg->setCommand(CREATE);
			createMsg->setServiceType(FORWARD);
			createMsg->setNodeType(INTERMEDIATENODE);
			createMsg->setJobType(DELETENODE);
			createMsg->setSubSpaceToCreate( SubSpace(respNode->getId()) );
			sendMsg(createMsg, m->destip);
			if( logLevel > 0 ) { nodeLog.push_back( NodeLogEntry(simTime(),26,respNode->getId(),m->destip) ); }	
		}
	}
	
}

void YMMOG::processDelayTimer()  {
	cancelEvent(delayTimer);
	for( b = backup_nodes.begin(); b != backup_nodes.end(); ++b) {
		Delay delay = b->getMaxDelay();
		if( b->getBackupDelay() < delay.delay ) {
			// Send Responsible Node Replace for delay.ip 
			if( logLevel > 0 ) nodeLog.push_back( NodeLogEntry(simTime(), 20, b->getSubSpace().getId(), delay.ip, b->getSubSpace().getRespIp(), b->getBackupDelay(),delay.delay) );
			ReplaceIntermediateNodeMessage* replaceMsg = new ReplaceIntermediateNodeMessage("ReplaceIntermediateNodeData");
			replaceMsg->setCommand(REPLACEINT);
			replaceMsg->setServiceType(REQUEST);
			replaceMsg->setSubSpaceId(b->getSubSpace().getId() );
			replaceMsg->setToReplace( delay.ip );
			sendMsg(replaceMsg, b->getSubSpace().getRespIp() ); 
		}
	}
	scheduleAt(simTime() + delayDelay , delayTimer);
}

void YMMOG::processDelay(DelayMessage* delayMsg) {
	BackupNodeData* backupNode = findBackupNodeDataById(delayMsg->getSubSpaceId() );
	if( backupNode != NULL ) {
		backupNode->addDelay( delayMsg->getDelay(), delayMsg->getFromIp() );	
	}
	delete delayMsg;
}
void YMMOG::processReplaceIntermediateNode(ReplaceIntermediateNodeMessage* replaceMsg) {
	if( replaceMsg->getServiceType() == REQUEST) {
		ResponsibleNodeData* respNode = findResponsibleNodeDataById( replaceMsg->getSubSpaceId() );
		if( respNode != NULL ) {
			
			std::list<IPAddress> ips = respNode->replaceIntermediateNode(replaceMsg->getToReplace(), replaceMsg->getSrcip());
			replaceMsg->setIpListArraySize(ips.size());
			if( logLevel > 0 ) nodeLog.push_back( NodeLogEntry(simTime(), 25, replaceMsg->getSubSpaceId(), replaceMsg->getToReplace(), replaceMsg->getSrcip(), (double) ips.size() ) );
			// copy Ips to replaceMsg
			for(unsigned int i=0; i < replaceMsg->getIpListArraySize(); i++) {
				replaceMsg->setIpList( i, ips.front());
				ips.pop_front();
			}
			magicLog->record(replaceMsg->getSubSpaceId(),5,respNode->getReplacedIntermediateNodeCount());
			replaceMsg->setSubSpace( respNode->getSubSpace() );
			replaceMsg->setServiceType( RESPONSE );
			// Request new Backup Node
			RequestNodeMessage* requestMsg = new RequestNodeMessage("RequestBackupNodeData");
			requestMsg->setCommand(REQUESTNODE);
			requestMsg->setServiceType(REQUEST);
			requestMsg->setNodeType(BACKUPNODE);
			requestMsg->setSubSpaceId(replaceMsg->getSubSpaceId() );
			if( logLevel > 0 ) nodeLog.push_back( NodeLogEntry( simTime(), 22, replaceMsg->getSubSpaceId(), respNode->getBackupIp() ) );
			
			sendMsg(replaceMsg, replaceMsg->getSrcip() );
			sendMsg(requestMsg, lobbyIp);
		}
	}
	else if ( replaceMsg->getServiceType() == RESPONSE ) {	
		if ( findIntermediateNodeDataById( replaceMsg->getSubSpaceId() ) == NULL ) {
			intermediate_nodes.push_back( IntermediateNodeData(replaceMsg->getSubSpace() ) );
		}
		IntermediateNodeData* intNode = findIntermediateNodeDataById( replaceMsg->getSubSpaceId() );
		if( intNode != NULL ) {
			for(unsigned int i=0; i < replaceMsg->getIpListArraySize() ; ++i ){
				intNode->addIp( replaceMsg->getIpList(i));
			}
		}
		deleteFromDupedMessages(replaceMsg);
		replaceMsg->setIpListArraySize(0);
		replaceMsg->setServiceType(FORWARD);
		sendMsg(replaceMsg, replaceMsg->getToReplace() );
	}
	else if( replaceMsg->getServiceType() == FORWARD ) {
		deleteIntermediateNodeDataById( replaceMsg->getSubSpaceId() );
		delete replaceMsg;
	}
	else {
		delete replaceMsg;
	}
}



void YMMOG::processStabilizeTimer() {
	cancelEvent(stabilizeTimer);
	for( i = responsible_nodes.begin(); i != responsible_nodes.end(); ++i ) {
		sendStabilize((*i)->getSubSpace().getId(), RESPONSIBLENODE, (*i)->getSubSpace() );
	}
	for( b= backup_nodes.begin(); b != backup_nodes.end(); ++b) {
		sendStabilize(b->getSubSpace().getId(), BACKUPNODE, b->getSubSpace() );
	}
	scheduleAt(simTime() + stabilizeDelay, stabilizeTimer);
}

void YMMOG::sendStabilize(int subSpaceId, int nodeType, SubSpace subSpace) {
	StabilizeMessage* stabilizeMsg = new StabilizeMessage("Stabilize: Am i?");
	stabilizeMsg->setCommand(STABILIZE);
	stabilizeMsg->setServiceType(REQUEST);
	stabilizeMsg->setNodeType(nodeType);
	stabilizeMsg->setSubSpaceId(subSpaceId);
	stabilizeMsg->setCorrect(false);
	stabilizeMsg->setSubSpace(subSpace);
	sendMsg(stabilizeMsg, lobbyIp);
}

void YMMOG::processStabilize(StabilizeMessage* stabilizeMsg) {
	if( stabilizeMsg->getServiceType() == REQUEST ) {
		if( findResponsibleNodeDataById(stabilizeMsg->getSubSpaceId()) == NULL ) {
			responsible_nodes.push_back(new ResponsibleNodeData( stabilizeMsg->getSubSpace() , maxPlayerLeafs, treeWidth, (int) simTime() ));
		}
	}
	else if( stabilizeMsg->getServiceType() == RESPONSE ) {
		if ( ! stabilizeMsg->getCorrect() ) {
			if( stabilizeMsg->getNodeType() == RESPONSIBLENODE) {
				for( i = responsible_nodes.begin(); i != responsible_nodes.end(); ++i) {
					if( (*i)->getSubSpace().getId() == stabilizeMsg->getSubSpaceId() ) {
						responsible_nodes.erase(i);
						break;
					}
				}
			}
			else if ( stabilizeMsg->getNodeType() == BACKUPNODE) {
				for( b = backup_nodes.begin(); b != backup_nodes.end(); ++b) {
					if( b->getSubSpace().getId() == stabilizeMsg->getSubSpaceId() ) {
						backup_nodes.erase(b);
						break;
					}
				}
			}
		}
		else { 
			if( stabilizeMsg->getNodeType() == RESPONSIBLENODE ) {
				ResponsibleNodeData* respNode = findResponsibleNodeDataById(stabilizeMsg->getSubSpaceId() ) ;
				if( respNode != NULL ) {
					respNode->setSubSpace( stabilizeMsg->getSubSpace() );
				}
			}
			else if( stabilizeMsg->getNodeType() == BACKUPNODE) {
				BackupNodeData* backupNode = findBackupNodeDataById( stabilizeMsg->getSubSpaceId() );
				if( backupNode != NULL ) {
					backupNode->setSubSpace( stabilizeMsg->getSubSpace() );
				}
			}
		}
	}
	delete stabilizeMsg;
}

void YMMOG::processBandwithTimer() {
	cancelEvent(bandwithTimer);
	upstream->record( up / bandwithDelay );
	downstream->record( down / bandwithDelay );
	if( responsible_nodes.size() == 1 && backup_nodes.size() == 0 && intermediate_nodes.size() == 0) {
		magicLog->record(responsible_nodes.front()->getPlayerListSize(),11, up / bandwithDelay);
		magicLog->record(responsible_nodes.front()->getPlayerListSize(),12, down / bandwithDelay);
	}
	else if( responsible_nodes.size() == 0 && backup_nodes.size() == 1 && intermediate_nodes.size() == 0) {
		magicLog->record(backup_nodes.front().getSize(),13, up / bandwithDelay);
		magicLog->record(backup_nodes.front().getSize(),14, down / bandwithDelay);
	}
	else if( responsible_nodes.size() == 0 && backup_nodes.size() == 0 && intermediate_nodes.size() == 1) {
		magicLog->record(intermediate_nodes.front().getSize(),15, up / bandwithDelay);
		magicLog->record(intermediate_nodes.front().getSize(),16, down / bandwithDelay);
	}		
	up = 0;
	down = 0;	
	scheduleAt(simTime() + bandwithDelay, bandwithTimer);
}

std::ostream& operator<<(std::ostream& Stream, const YMMOG y)
{
    return Stream << "IP: "  << "Holding "<< y.responsible_nodes.size() << " Responsible Nodes";
}

