//
// Copyright (C) 2007 Institut fuer Telematik, Universitaet Karlsruhe (TH)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//

/**
 * @file DHTTestApp.cc
 * @author Ingmar Baumgart
 */

#include <IPAddressResolver.h>
#include <BootstrapOracleAccess.h>
#include <GlobalStatisticsAccess.h>
#include <UnderlayConfiguratorAccess.h>
#include "../Common/CommonMessages_m.h"

#include "DHTTestApp.h"

Define_Module(DHTTestApp);

const int DEFAULT_TTL = 30000;

DHTTestApp::~DHTTestApp()
{
    //...
}

void DHTTestApp::initializeApp(int stage)
{
    if (stage != MIN_STAGE_APP)
        return;

    // fetch parameters
    debugOutput = par("debugOutput");
    activeNetwInitPhase = par("activeNetwInitPhase");

    mean = par("messageDelay");
    deviation = mean / 10;

    bootstrapOracle = BootstrapOracleAccess().get();
    underlayConfigurator = UnderlayConfiguratorAccess().get();
    globalStatistics = GlobalStatisticsAccess().get();

    // statistics
    numSent = 0;
    numGetSent = 0;
    numGetError = 0;
    numGetSuccess = 0;
    numPutSent = 0;
    numPutError = 0;
    numPutSuccess = 0;

    initRpcs();
    WATCH(numSent);
    WATCH(numGetSent);
    WATCH(numGetError);
    WATCH(numGetSuccess);
    WATCH(numPutSent);
    WATCH(numPutError);
    WATCH(numPutSuccess);

    // initiate test message emision
    cMessage* dhttestput_timer = new cMessage("dhttest_put_timer");
    cMessage* dhttestget_timer = new cMessage("dhttest_get_timer");
    cMessage* dhttestmod_timer = new cMessage("dhttest_mod_timer");

    if (mean > 0) {
	scheduleAt(simulation.simTime() + truncnormal(mean, deviation),
		dhttestput_timer);
	scheduleAt(simulation.simTime() + truncnormal(mean+mean/3, deviation),
		dhttestget_timer);
	scheduleAt(simulation.simTime() + truncnormal(mean+2*mean/3, deviation),
		dhttestmod_timer);
    }
}

void DHTTestApp::handleRpcResponse(BaseResponseMessage* msg, int rpcId,
                                   simtime_t rtt)
{
    RPC_SWITCH_START(msg)
    RPC_ON_RESPONSE( DHTputCAPI ) {
        handlePutResponse(_DHTputCAPIResponse);
        EV << "DHT Put RPC Response received: id=" << rpcId
           << " msg=" << *_DHTputCAPIResponse << " rtt=" << rtt << endl;
        break;
    }
    RPC_ON_RESPONSE( DHTgetCAPI ) {
        handleGetResponse(_DHTgetCAPIResponse);
        EV << "DHT Get RPC Response received: id=" << rpcId
           << " msg=" << *_DHTgetCAPIResponse << " rtt=" << rtt << endl;
        break;
    }
    RPC_SWITCH_END( )
}

void DHTTestApp::handlePutResponse(DHTputCAPIResponse* msg)
{
    OverlayKey key = msg->getKey();
    globalStatistics->addStdDev("DHTTest: Put length (s)", simTime()
            - lastPutTime);
    if (!(msg->getIsSuccess())) {
        numPutError++;
        return;
    }
    numPutSuccess++;
    DHTEntry entry;
    entry.value = new BinaryValue(msg->getValue());
    entry.endtime = simulation.simTime()+DEFAULT_TTL;
    bootstrapOracle->getDataMap()->erase(msg->getKey());
    bootstrapOracle->getDataMap()->insert(make_pair(msg->getKey(), entry));
}

void DHTTestApp::handleGetResponse(DHTgetCAPIResponse* msg)
{
    OverlayKey key = msg->getKey();
    globalStatistics->addStdDev("DHTTest: Get length (s)",
                                simTime() - lastGetTime);

    if (!(msg->getIsSuccess())) {
        numGetError++;
        return;
    }
    std::map<OverlayKey, DHTEntry>::iterator it = bootstrapOracle->getDataMap()->find(key);

    if (it == bootstrapOracle->getDataMap()->end()) {
        //unexpected key
        numGetError++;
        return;
    }

    DHTEntry entry = it->second;
    if (simulation.simTime() > entry.endtime) {
        //this key doesn't exist anymore in the DHT, delete it in our hashtable
        bootstrapOracle->getDataMap()->erase(key);
        if (msg->getValue() != BinaryValue::UNSPECIFIED_VALUE) {
            numGetError++;
            return;
        } else {
            numGetSuccess++;
            return;
        }
    } else {
        if (dynamic_cast< vector<char>& >((msg->getValue()))
                != dynamic_cast< vector<char>& >(*(entry.value))) {
            numGetError++;
            return;
        } else {
            numGetSuccess++;
            return;
        }
    }

}

void DHTTestApp::handleTraceMessage(cMessage* msg)
{
    char* cmd = new char[strlen(msg->name())];
    strcpy(cmd, msg->name());
    if (strncmp(cmd, "PUT ", 4) == 0) {
	// Generate key
	char* buf = cmd + 4;
	while (!isspace(buf[0])) {
	    if ( buf[0] == '\0' ) throw new cException("Error parsing PUT command");
	    buf++;
	}
	buf[0] = '\0';
	BinaryValue b(cmd+4);
	OverlayKey destKey(OverlayKey::sha1(b));

	// get value
	buf++;

	// build putMsg
        DHTputCAPICall* dhtPutMsg = new DHTputCAPICall();
        dhtPutMsg->setKey(destKey);
        dhtPutMsg->setValue(buf);
        dhtPutMsg->setTtl(DEFAULT_TTL);
        dhtPutMsg->setIsModifiable(true);

        numSent++;
        numPutSent++;
        sendInternalRpcCall(TIER1_COMP, dhtPutMsg);
        delete cmd;
    } else if (strncmp(cmd, "GET ", 4) == 0) {
	// Get key
        BinaryValue b(cmd+4);
        OverlayKey key(OverlayKey::sha1(b));

        DHTgetCAPICall* dhtGetMsg = new DHTgetCAPICall();
        dhtGetMsg->setKey(key);
        lastGetTime = simTime();
        numSent++;
        numGetSent++;
        sendInternalRpcCall(TIER1_COMP, dhtGetMsg);
    } else {
        throw new cRuntimeError("Unknown trace command; "
                                "only GET and PUT are allowed");
    }
}

void DHTTestApp::handleTimerEvent(cMessage* msg)
{
    if (msg->isName("dhttest_put_timer")) {
        // schedule next timer event
        scheduleAt(simulation.simTime() + truncnormal(mean, deviation), msg);

        // do nothing if the network is still in the initialization phase
        if ((!activeNetwInitPhase) && (underlayConfigurator->isInit()))
            return;

        // create a put test message with random destination key
        OverlayKey destKey = OverlayKey::random();
        DHTputCAPICall* dhtPutMsg = new DHTputCAPICall();
        char * value = generateRandomValue();
        dhtPutMsg->setKey(destKey);
        dhtPutMsg->setValue(value);
        dhtPutMsg->setTtl(DEFAULT_TTL);
        dhtPutMsg->setIsModifiable(true);
        free(value);

        lastPutTime = simTime();
        numSent++;
        numPutSent++;
        sendInternalRpcCall(TIER1_COMP, dhtPutMsg);
    } else  if (msg->isName("dhttest_get_timer")) {
        scheduleAt(simulation.simTime() + truncnormal(mean, deviation), msg);
        OverlayKey key = getRandomKey();
        
        if (key.isUnspecified())
            return;
        
        DHTgetCAPICall* dhtGetMsg = new DHTgetCAPICall();
        dhtGetMsg->setKey(key);
        lastGetTime = simTime();
        numSent++;
        numGetSent++;
        sendInternalRpcCall(TIER1_COMP, dhtGetMsg);
    } else if (msg->isName("dhttest_mod_timer")) {
        scheduleAt(simulation.simTime() + truncnormal(mean, deviation), msg);
        OverlayKey key = getRandomKey();

        if (key.isUnspecified())
            return;

        DHTputCAPICall* dhtPutMsg = new DHTputCAPICall();
        char * value = generateRandomValue();
        dhtPutMsg->setKey(key);
        dhtPutMsg->setValue(value);
        dhtPutMsg->setTtl(DEFAULT_TTL);
        dhtPutMsg->setIsModifiable(true);
        free(value);
        lastPutTime = simTime();
        numSent++;
        numPutSent++;
        sendInternalRpcCall(TIER1_COMP, dhtPutMsg);
    }
}

OverlayKey DHTTestApp::getRandomKey()
{
    BootstrapOracle::DataMap* dataMap = bootstrapOracle->getDataMap();

    if (dataMap->size() == 0)
        return OverlayKey::UNSPECIFIED_KEY;

    // return random OverlayKey in O(log n) (taken from BootstrapOracle)
    std::map<OverlayKey, DHTEntry>::iterator it = dataMap->end();
    DHTEntry tempEntry = { NULL, 0 };

    while (it == dataMap->end()) {
        OverlayKey randomKey = OverlayKey::random();

        it = dataMap->find(randomKey);

        if (it == dataMap->end()) {
            it = dataMap->insert(std::make_pair(randomKey,tempEntry)).first;
            dataMap->erase(it++);
        }

        if (it == dataMap->end())
            it = dataMap->begin();
    }
    return it->first;
}

char* DHTTestApp::generateRandomValue()
{
    char* value;
    value = (char *)malloc(sizeof(char)*21);
    int i, randvalue;

    for (i=0; i < 20; i++) {
        randvalue = (int)intuniform(0, 25);
        value[i] = randvalue+'a';
    }

    value[20] = '\0';
    return value;
}

void DHTTestApp::finishApp()
{
    // record scalar data
    globalStatistics->addStdDev("DHTTest: Sent Messages", numSent);
    globalStatistics->addStdDev("DHTTest: Get Messages sent", numGetSent);
    globalStatistics->addStdDev("DHTTest: Get Errors", numGetError);
    globalStatistics->addStdDev("DHTTest: Get Success", numGetSuccess);
    if (numGetSent > 0) {
        globalStatistics->addStdDev("DHTTest: Get Success Ratio",
                                        (double)numGetSuccess/(double)numGetSent);
    } else { 
        globalStatistics->addStdDev("DHTTest: Get Success Ratio", 0);
    }
    globalStatistics->addStdDev("DHTTest: Put Messages sent", numPutSent);
    globalStatistics->addStdDev("DHTTest: Put Errors", numPutError);
    globalStatistics->addStdDev("DHTTest: Put Success", numPutSuccess);
}

