//
// Copyright (C) 2007 Institut fuer Telematik, Universitaet Karlsruhe (TH)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//

/**
 * @file P2pns.cc
 * @author Ingmar Baumgart
 */

#include <IPAddressResolver.h>

#include "P2pns.h"

Define_Module(P2pns);

P2pns::P2pns()
{
    p2pnsCache = NULL;
}

P2pns::~P2pns()
{
    std::map<unsigned int, BaseCallMessage*>::iterator it = rpcIdMap.begin();
    while (it != rpcIdMap.end()) {
        cancelAndDelete(it->second);
        it++;
    }
    rpcIdMap.clear();
}

void P2pns::initializeApp(int stage)
{
    if (stage != MIN_STAGE_APP)
        return;
    
    twoStageResolution = par("twoStageResolution");
    
    p2pnsCache = check_and_cast<P2pnsCache*>
                     (parentModule()->submodule("p2pnsCache"));
}

void P2pns::handleTimerEvent(cMessage* msg)
{
}

bool P2pns::handleRpc(BaseCallMessage* msg)
{
    // delegate messages
    RPC_SWITCH_START(msg)
    // RPC_DELEGATE( <messageName>[Call|Response], <methodToCall> )
    RPC_DELEGATE( P2pnsRegister, p2pnsRegisterRpc );
    RPC_DELEGATE( P2pnsResolve, p2pnsResolveRpc );
    RPC_SWITCH_END()

    return RPC_HANDLED;
}


void P2pns::handleRpcResponse(BaseResponseMessage* msg, int rpcId, simtime_t rtt)
{
    RPC_SWITCH_START(msg)
    RPC_ON_RESPONSE( DHTputCAPI ) {
        EV << "DHTputCAPI RPC Response received: id=" << rpcId
        << " msg=" << *_DHTputCAPIResponse << " rtt=" << rtt << endl;
        handleDHTputCAPIResponse(_DHTputCAPIResponse);
        break;
    }
    RPC_ON_RESPONSE( DHTgetCAPI ) {
        EV << "DHTgetCAPI RPC Response received: id=" << rpcId
           << " msg=" << *_DHTgetCAPIResponse << " rtt=" << rtt << endl;
        handleDHTgetCAPIResponse(_DHTgetCAPIResponse);
        break;
    }
    RPC_ON_RESPONSE( Lookup ) {
            EV << "Lookup RPC Response received: id=" << rpcId
               << " msg=" << *_LookupResponse << " rtt=" << rtt << endl;
            handleLookupResponse(_LookupResponse);
            break;
    }
    RPC_SWITCH_END()
}

void P2pns::handleRpcTimeout(BaseCallMessage* msg,
                             const TransportAddress& dest, int rpcId,
                             const OverlayKey& destKey)
{
    opp_error("P2PNS::handleRpcTimeout(): Not implemented!");
}

void P2pns::p2pnsRegisterRpc(P2pnsRegisterCall* registerCall)
{
    p2pnsCache->addData(registerCall->getName(),
                        registerCall->getAddress());
    
    DHTputCAPICall* dhtPutMsg = new DHTputCAPICall();

    dhtPutMsg->setKey(OverlayKey::sha1(registerCall->getName()));
    if (twoStageResolution) {
        dhtPutMsg->setValue(overlay->getThisNode().key.toString());
    } else {
        dhtPutMsg->setValue(registerCall->getAddress());
    }
    
    dhtPutMsg->setTtl(registerCall->getTtl());
    dhtPutMsg->setIsModifiable(true);

    int nonce = sendInternalRpcCall(TIER1_COMP, dhtPutMsg); 

    rpcIdMap.insert(make_pair(nonce, registerCall));
}

void P2pns::p2pnsResolveRpc(P2pnsResolveCall* resolveCall)
{
    DHTgetCAPICall* dhtGetMsg = new DHTgetCAPICall();

    dhtGetMsg->setKey(OverlayKey::sha1(resolveCall->getName()));
        
    int nonce = sendInternalRpcCall(TIER1_COMP, dhtGetMsg);
    
    rpcIdMap.insert(make_pair(nonce, resolveCall));
}

void P2pns::handleDHTputCAPIResponse(DHTputCAPIResponse* putResponse)
{
    std::map<unsigned int, BaseCallMessage*>::iterator it =
            rpcIdMap.find(putResponse->getNonce());
    if (it == rpcIdMap.end() || it->second == NULL) {
        opp_error("P2PNS::handleDHTputCAPIResponse(): Unknown nonce!");
        return;
    }

    P2pnsRegisterCall* registerCall =
            check_and_cast<P2pnsRegisterCall*>(it->second);

    rpcIdMap.erase(putResponse->getNonce());

    P2pnsRegisterResponse* registerResponse = new P2pnsRegisterResponse();
    registerResponse->setName(registerCall->getName());
    registerResponse->setAddress(registerCall->getAddress());
    registerResponse->setIsSuccess(putResponse->getIsSuccess());
    sendRpcResponse(registerCall, registerResponse);
}

void P2pns::handleDHTgetCAPIResponse(DHTgetCAPIResponse* getResponse)
{
    std::map<unsigned int, BaseCallMessage*>::iterator it =
            rpcIdMap.find(getResponse->getNonce());
    if (it == rpcIdMap.end() || it->second == NULL) {
        opp_error("P2PNS::handleDHTgetCAPIResponse(): Unknown nonce!");
        return;
    }

    P2pnsResolveCall* resolveCall =
            check_and_cast<P2pnsResolveCall*>(it->second);

    rpcIdMap.erase(getResponse->getNonce());
    
    std::stringstream valueStream;
    valueStream << getResponse->getValue();
            
    if ((!getResponse->getIsSuccess())
            || (valueStream.str().size() == 0)) {
        P2pnsResolveResponse* resolveResponse = new P2pnsResolveResponse();
        resolveResponse->setName(resolveCall->getName());
        resolveResponse->setAddress(BinaryValue(""));
        resolveResponse->setIsSuccess(false);
        sendRpcResponse(resolveCall, resolveResponse);
        return;
    }

    p2pnsCache->addData(resolveCall->getName(),
                        getResponse->getValue());

    if (twoStageResolution) {
        OverlayKey key(valueStream.str(), 16);
        
        LookupCall* lookupCall = new LookupCall();
        
        lookupCall->setKey(key);
        lookupCall->setNumSiblings(1);
            
        int nonce = sendInternalRpcCall(OVERLAY_COMP, lookupCall);
        
        rpcIdMap.insert(make_pair(nonce, resolveCall));
        
        return;
    }
    
    P2pnsResolveResponse* resolveResponse = new P2pnsResolveResponse();
    resolveResponse->setName(resolveCall->getName());
    resolveResponse->setAddress(getResponse->getValue());
    resolveResponse->setIsSuccess(getResponse->getIsSuccess());
    sendRpcResponse(resolveCall, resolveResponse);
}

void P2pns::handleLookupResponse(LookupResponse* lookupResponse)
{
    std::map<unsigned int, BaseCallMessage*>::iterator it =
            rpcIdMap.find(lookupResponse->getNonce());
    if (it == rpcIdMap.end() || it->second == NULL) {
        opp_error("P2PNS::handleLookupResponse(): Unknown nonce!");
        return;
    }

    P2pnsResolveCall* resolveCall =
            check_and_cast<P2pnsResolveCall*>(it->second);

    rpcIdMap.erase(lookupResponse->getNonce());
    
    stringstream sstream;
    sstream << lookupResponse->getSiblings(0);
    
    P2pnsResolveResponse* resolveResponse = new P2pnsResolveResponse();
    resolveResponse->setName(resolveCall->getName());
    resolveResponse->setAddress(sstream.str());
    resolveResponse->setIsSuccess(lookupResponse->getIsValid());
    sendRpcResponse(resolveCall, resolveResponse);
}


void P2pns::finishApp()
{
}

