//
// Copyright (C) 2006 Institut fuer Telematik, Universitaet Karlsruhe (TH)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//

/**
 * @file SimMud.cc
 * @author Stephan Krause
 */

#include "SimMud.h"
#include "ScribeMessage_m.h"
#include "limits.h"

Define_Module(SimMud);

SimMud::SimMud()
{
    currentRegionX = currentRegionY = INT_MIN;
    currentRegionID = OverlayKey::UNSPECIFIED_KEY;
    playerTimer = new cMessage("playerTimeout");
}

SimMud::~SimMud()
{
    cancelAndDelete(playerTimer);
}

void SimMud::finishApp()
{
}

void SimMud::initializeApp(int stage)
{
    if( stage != (numInitStages()-1)) {
        return;
    }

    // FIXME: areaDimension is not used consistently between all overlays!
    fieldSize = par("areaDimension");
    regionSize = par("regionDimension");
    playerTimeout = par("playerTimeout");

    WATCH(currentRegionX);
    WATCH(currentRegionY);
    WATCH(currentRegionID);
//    WATCH_MAP( playerMap );
}

bool SimMud::handleRpc( BaseCallMessage* msg )
{
//    RPC_SWITCH_START(msg);
//    RPC_SWITCH_END( );
//    return RPC_HANDLED;
      return false;
}

void SimMud::handleRpcResponse( BaseResponseMessage* msg, int rpcId, simtime_t rtt )
{
//    RPC_SWITCH_START(msg);
//    RPC_SWITCH_END( );
}

void SimMud::handleTimerEvent( cMessage *msg )
{
    if( msg == playerTimer ) {
        // Check if any player didn't send any updates since last check
        map<NodeHandle, PlayerInfo>::iterator it;
        list<NodeHandle> erasePlayers;
        for( it = playerMap.begin(); it != playerMap.end(); ++it ) {
            if( it->second.update == false ) {
                erasePlayers.push_back( it->first );
            }
            it->second.update = false;
        }
        for( list<NodeHandle>::iterator it = erasePlayers.begin(); it != erasePlayers.end(); ++it) {
            // Delete all neighbors that didn't update (and inform app)
            GameAPIListMessage *scMsg = new GameAPIListMessage("NEIGHBOR_UPDATE");
            scMsg->setCommand(NEIGHBOR_UPDATE);
            scMsg->setRemoveNeighborArraySize(1);
            scMsg->setRemoveNeighbor(0, *it);
            send(scMsg, "to_upperTier");

            playerMap.erase( *it );
        }
        erasePlayers.clear();

        scheduleAt( simulation.simTime() + playerTimeout, msg );
    }
}

void SimMud::handleLowerMessage( cMessage *msg )
{
    if (ALMMulticastMessage* mcastMsg =
            dynamic_cast<ALMMulticastMessage*>(msg) ){

        cMessage* innerMsg = mcastMsg->decapsulate();
        SimMudMoveMessage* moveMsg = NULL;
        if( innerMsg ) {
                moveMsg = dynamic_cast<SimMudMoveMessage*>(innerMsg);
        }
        if( moveMsg ) {
            handleOtherPlayerMove( moveMsg );
        }
        delete innerMsg;
        delete mcastMsg;
    }
}

void SimMud::handleReadyMessage( TierReadyMessage *msg )
{
    if( msg->getReady() ) {
        // FIXME/TODO: Should be done by BaseApp
        thisNode = msg->getThisNode();
        // TODO/FIXME: maybe ugly? send TierReadyMessage directly form BaseApp to SimpleGameClient?
        send(msg, "to_upperTier");

        if( playerTimeout > 0 ) {
            scheduleAt( simulation.simTime() + playerTimeout, playerTimer );
        }
    }
    //delete msg;
}

void SimMud::handleUpperMessage( cMessage *msg )
{
    if( GameAPIPositionMessage* moveMsg =
        dynamic_cast<GameAPIPositionMessage*>(msg) ) {
        handleMove( moveMsg );
        delete msg;
    }
}

void SimMud::handleMove( GameAPIPositionMessage* msg )
{
    if( (int) (msg->getPosition().x/regionSize) != currentRegionX ||
            (int) (msg->getPosition().y/regionSize) != currentRegionY ) {
        // New position is outside of current region; change regions

        if( !currentRegionID.isUnspecified() ){
            // leave old region's multicastGroup
            ALMLeaveMessage* leaveMsg = new ALMLeaveMessage("LEAVE_REGION_GROUP");
            leaveMsg->setGroupId(currentRegionID);
            send(leaveMsg, "to_lowerTier");
            // TODO: leave old simMud region

            // Inform other players about region leave
            SimMudMoveMessage* moveMsg = new SimMudMoveMessage("MOVE/LEAVE_REGION");
            moveMsg->setSrc( thisNode );
            moveMsg->setPosX( msg->getPosition().x );
            moveMsg->setPosY( msg->getPosition().y );
            moveMsg->setLeaveRegion( true );
            ALMMulticastMessage* mcastMsg = new ALMMulticastMessage("MOVE/LEAVE_REGION");
            mcastMsg->setGroupId(currentRegionID);
            mcastMsg->encapsulate( moveMsg );

            send(mcastMsg, "to_lowerTier");
        }

        // get region ID
        currentRegionX = (int) (msg->getPosition().x/regionSize);
        currentRegionY = (int) (msg->getPosition().y/regionSize);
        std::stringstream regionstr;
        regionstr << currentRegionX << ":" << currentRegionY;
        OverlayKey region = OverlayKey::sha1( BinaryValue(regionstr.str() ));
        currentRegionID = region;

        // join region's multicast group
        ALMSubscribeMessage* subMsg = new ALMSubscribeMessage;
        subMsg->setGroupId(region);
        send( subMsg, "to_lowerTier" );

        // TODO: join simMud region
   }

    // publish movement
    SimMudMoveMessage* moveMsg = new SimMudMoveMessage("MOVE");
    moveMsg->setSrc( thisNode );
    moveMsg->setPosX( msg->getPosition().x );
    moveMsg->setPosY( msg->getPosition().y );
    ALMMulticastMessage* mcastMsg = new ALMMulticastMessage("MOVE");
    mcastMsg->setGroupId(currentRegionID);
    mcastMsg->encapsulate( moveMsg );

    send(mcastMsg, "to_lowerTier");
}

void SimMud::handleOtherPlayerMove( SimMudMoveMessage* msg )
{
    GameAPIListMessage *scMsg = new GameAPIListMessage("NEIGHBOR_UPDATE");
    scMsg->setCommand(NEIGHBOR_UPDATE);

    NodeHandle& src = msg->getSrc();

    if( msg->getLeaveRegion() ) {
        // Player leaves region
        scMsg->setRemoveNeighborArraySize(1);
        scMsg->setRemoveNeighbor(0, src);
        playerMap.erase( src );

    } else {
        PlayerInfo player;
        player.pos = Vector2D( msg->getPosX(), msg->getPosY() );
        player.update = true;

        pair< map<NodeHandle, PlayerInfo>::iterator, bool> inserter =
            playerMap.insert( make_pair(src, player) );

        /*    if( inserter.second ) {
            // new player

        } else {
            // move player
        }*/

        // Ordinary move
        scMsg->setAddNeighborArraySize(1);
        scMsg->setNeighborPositionArraySize(1);
        scMsg->setAddNeighbor(0, src);
        scMsg->setNeighborPosition(0, Vector2D(msg->getPosX(), msg->getPosY()) );
    }
    send(scMsg, "to_upperTier");

}

