//
// Copyright (C) 2007 Institut fuer Telematik, Universitaet Karlsruhe (TH)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//

/**
 * @file XmlRpcInterface.cc
 * @author Ingmar Baumgart
 */

#include <NodeVector.h>
#include "XmlRpcInterface.h"

using namespace XmlRpc;

Define_Module(XmlRpcInterface);

// Register a name with P2PNS
class P2pnsRegister : public XmlRpcServerMethod
{
public:
    P2pnsRegister(XmlRpcServer* s) :
        XmlRpcServerMethod("register", s)
    {
    }

    void execute(XmlRpcValue& params, XmlRpcValue& result)
    {
        (dynamic_cast<XmlRpcInterface*>(_server))->p2pnsRegister(params, result);
    }

    std::string help()
    {
        return std::string("Register a name with P2PNS");
    }
};

// Register a name with P2PNS
class P2pnsResolve : public XmlRpcServerMethod
{
public:
    P2pnsResolve(XmlRpcServer* s) :
        XmlRpcServerMethod("resolve", s)
    {
    }

    void execute(XmlRpcValue& params, XmlRpcValue& result)
    {
        (dynamic_cast<XmlRpcInterface*>(_server))->p2pnsResolve(params, result);
    }

    std::string help()
    {
        return std::string("Resolve a name with P2PNS");
    }
};

// Common API: local_lookup()
class LocalLookup : public XmlRpcServerMethod
{
public:
    LocalLookup(XmlRpcServer* s) :
        XmlRpcServerMethod("local_lookup", s)
    {
    }

    void execute(XmlRpcValue& params, XmlRpcValue& result)
    {
        (dynamic_cast<XmlRpcInterface*>(_server))->localLookup(params, result);
    }

    std::string help()
    {
        return std::string("Lookup key in local "
            "routing table");
    }
};

// Common API: lookup()
class Lookup : public XmlRpcServerMethod
{
public:
    Lookup(XmlRpcServer* s) :
        XmlRpcServerMethod("lookup", s)
    {
    }

    void execute(XmlRpcValue& params, XmlRpcValue& result)
    {
        (dynamic_cast<XmlRpcInterface*>(_server))->lookup(params, result);
    }

    std::string help()
    {
        return std::string("Lookup key with KBR layer");
    }
};

// Store a value in the DHT
class Put : public XmlRpcServerMethod
{
public:
    Put(XmlRpcServer* s) :
        XmlRpcServerMethod("put", s)
    {
    }

    void execute(XmlRpcValue& params, XmlRpcValue& result)
    {
        (dynamic_cast<XmlRpcInterface*>(_server))->put(params, result);
    }

    std::string help()
    {
        return std::string("Store a value in the DHT");
    }
};

// Get a value from the DHT
class Get : public XmlRpcServerMethod
{
public:
    Get(XmlRpcServer* s) :
        XmlRpcServerMethod("get", s)
    {
    }

    void execute(XmlRpcValue& params, XmlRpcValue& result)
    {
        (dynamic_cast<XmlRpcInterface*>(_server))->get(params, result);
    }

    std::string help()
    {
        return std::string("Get a value from the DHT");
    }
};

// Join the overlay with a specific nodeID
class JoinOverlay : public XmlRpcServerMethod
{
public:
    JoinOverlay(XmlRpcServer* s) :
        XmlRpcServerMethod("join", s)
    {
    }

    void execute(XmlRpcValue& params, XmlRpcValue& result)
    {
        (dynamic_cast<XmlRpcInterface*>(_server))->joinOverlay(params, result);
    }

    std::string help()
    {
        return std::string("Join the overlay with a specific nodeID");
    }
};

void XmlRpcInterface::p2pnsRegister(XmlRpcValue& params, XmlRpcValue& result)
{
    if ((params.size() != 3) ||
            (params[0].getType() != XmlRpcValue::TypeBase64) ||
            (params[1].getType() != XmlRpcValue::TypeBase64) ||
            (params[2].getType() != XmlRpcValue::TypeInt))
        throw XmlRpcException("register(base64 name, base64 address, int ttl): "
                              "Invalid argument type");

    if (overlay->getCompModule(TIER2_COMP) == NULL)
        throw XmlRpcException("register(base64 name, base64 address, int ttl): "
                              "No P2PNS service");

    _connectionState = EXECUTE_REQUEST;

    P2pnsRegisterCall* registerCall = new P2pnsRegisterCall();

    registerCall->setName(((const XmlRpcValue::BinaryData&)params[0]));
    registerCall->setAddress(((const XmlRpcValue::BinaryData&)params[1]));
    registerCall->setTtl(params[2]);

    sendInternalRpcCall(TIER2_COMP, registerCall);

    cancelEvent(timeout_msg);
    scheduleAt(simulation.simTime() + XMLRPC_TIMEOUT, timeout_msg);
}

void XmlRpcInterface::p2pnsResolve(XmlRpcValue& params, XmlRpcValue& result)
{
    if ((params.size() != 1) ||
            (params[0].getType() != XmlRpcValue::TypeBase64))
        throw XmlRpcException("resolve(base64 name): Invalid argument type");

    if (overlay->getCompModule(TIER2_COMP) == NULL)
        throw XmlRpcException("resolve(base64 name): No P2PNS service");

    _connectionState = EXECUTE_REQUEST;

    P2pnsResolveCall* resolveCall = new P2pnsResolveCall();

    resolveCall->setName(((const XmlRpcValue::BinaryData&)params[0]));
    
    sendInternalRpcCall(TIER2_COMP, resolveCall);
    
    cancelEvent(timeout_msg);
    scheduleAt(simulation.simTime() + XMLRPC_TIMEOUT, timeout_msg);
}

void XmlRpcInterface::localLookup(XmlRpcValue& params, XmlRpcValue& result)
{
    if ((params.size() != 3)
            || (params[0].getType() != XmlRpcValue::TypeBase64)
            || (params[1].getType() != XmlRpcValue::TypeInt)
            || (params[2].getType() != XmlRpcValue::TypeBoolean))
        throw XmlRpcException("local_lookup(base64 key, int num, "
                "boolean safe): Invalid argument type");

    BinaryValue keyString = (const XmlRpcValue::BinaryData&)params[0];

    NodeVector* nextHops = overlay->local_lookup(OverlayKey::sha1(keyString),
                                                 params[1], params[2]);

    for (uint i=0; i < nextHops->size(); i++) {
        result[i][0] = (*nextHops)[0].ip.str();
        result[i][1] = (*nextHops)[0].port;
        result[i][2] = (*nextHops)[0].key.toString(16);
    }

    delete nextHops;
}

void XmlRpcInterface::lookup(XmlRpcValue& params, XmlRpcValue& result)
{
    if ((params.size() != 2)
            || (params[0].getType() != XmlRpcValue::TypeBase64)
            || (params[1].getType() != XmlRpcValue::TypeInt))
        throw XmlRpcException("lookup(base64 key, int numSiblings): "
                "Invalid argument type");

    if ((int)params[1] > overlay->getMaxNumSiblings())
        throw XmlRpcException("lookup(base64 key, int numSiblings): "
                "numSibling to big");

    _connectionState = EXECUTE_REQUEST;

    LookupCall* lookupCall = new LookupCall();

    BinaryValue keyString = (const XmlRpcValue::BinaryData&)params[0];
    lookupCall->setKey(OverlayKey::sha1(keyString));
    lookupCall->setNumSiblings(params[1]);

    sendInternalRpcCall(OVERLAY_COMP, lookupCall);

    cancelEvent(timeout_msg);
    scheduleAt(simulation.simTime() + XMLRPC_TIMEOUT, timeout_msg);
}

void XmlRpcInterface::joinOverlay(XmlRpcValue& params, XmlRpcValue& result)
{
    if ((params.size() != 1)
            || (params[0].getType() != XmlRpcValue::TypeBase64))
        throw XmlRpcException("join(base64 nodeID): Invalid argument type");

    BinaryValue nodeID = (const XmlRpcValue::BinaryData&)params[0];

    overlay->join(OverlayKey::sha1(nodeID));

    result[0] = 0;
}

void XmlRpcInterface::put(XmlRpcValue& params, XmlRpcValue& result)
{
    if ((params.size() != 4)
            || (params[0].getType() != XmlRpcValue::TypeBase64)
            || (params[1].getType() != XmlRpcValue::TypeBase64)
            || (params[2].getType() != XmlRpcValue::TypeInt)
            || (params[3].getType() != XmlRpcValue::TypeString))
        throw XmlRpcException("put(base64 key, base64 value, int ttl "
                ", string application): Invalid argument type");

    if (overlay->getCompModule(TIER1_COMP) == NULL)
        throw XmlRpcException("put(base64 key, base64 value, int ttl "
                ", string application): No DHT service");

    _connectionState = EXECUTE_REQUEST;

    DHTputCAPICall* dhtPutMsg = new DHTputCAPICall();

    BinaryValue keyString = (const XmlRpcValue::BinaryData&)params[0];

    dhtPutMsg->setKey(OverlayKey::sha1(keyString));
    dhtPutMsg->setValue(((const XmlRpcValue::BinaryData&)params[1]));
    dhtPutMsg->setTtl(params[2]);
    dhtPutMsg->setIsModifiable(true);

    sendInternalRpcCall(TIER1_COMP, dhtPutMsg);

    cancelEvent(timeout_msg);
    scheduleAt(simulation.simTime() + XMLRPC_TIMEOUT, timeout_msg);
}

void XmlRpcInterface::get(XmlRpcValue& params, XmlRpcValue& result)
{
    if ((params.size() != 4)
            || (params[0].getType() != XmlRpcValue::TypeBase64)
            || (params[1].getType() != XmlRpcValue::TypeInt)
            || (params[2].getType() != XmlRpcValue::TypeBase64)
            || (params[3].getType() != XmlRpcValue::TypeString))
        throw XmlRpcException("get(base64 key, int num, base64 placemark "
                ", string application): Invalid argument type");

    if (overlay->getCompModule(TIER1_COMP) == NULL)
        throw XmlRpcException("get(base64 key, int num, base64 placemark "
                ", string application): No DHT service");

    _connectionState = EXECUTE_REQUEST;

    DHTgetCAPICall* dhtGetMsg = new DHTgetCAPICall();

    BinaryValue keyString = (const XmlRpcValue::BinaryData&)params[0];
    dhtGetMsg->setKey(OverlayKey::sha1(keyString));

    sendInternalRpcCall(TIER1_COMP, dhtGetMsg);

    cancelEvent(timeout_msg);
    scheduleAt(simulation.simTime() + XMLRPC_TIMEOUT, timeout_msg);
}

void XmlRpcInterface::initializeApp(int stage)
{
    // all initialization is done in the first stage
    if (stage != MIN_STAGE_APP)
        return;

    packetNotification = new cMessage("packetNotification");
    mtu = par("mtu");

    scheduler = check_and_cast<RealtimeScheduler *>(simulation.scheduler());
    scheduler->setInterfaceModule(this, packetNotification, &packetBuffer, mtu,
                                   true);

    XmlRpc::setVerbosity(1);

    _localLookup = new LocalLookup(this);
    _lookup = new Lookup(this);
    _register = new P2pnsRegister(this);
    _resolve = new P2pnsResolve(this);
    _put = new Put(this);
    _get = new Get(this);
    _joinOverlay = new JoinOverlay(this);

    enableIntrospection(true);
    
    appFd = -1;

    _connectionState = READ_HEADER;
    _keepAlive = true;

    timeout_msg = new cMessage("timeout_msg");
}

void XmlRpcInterface::handleMessage(cMessage *msg)
{
    // Packet from the application...
    if (msg==packetNotification) {
        EV << "[DHTXMLRealworldApp::handleMessage() @ " << thisNode.ip
        << " (" << thisNode.key.toString(16) << ")]\n"
        << "    Message from application. Queue length = " << packetBuffer.size()
        << endl;
        while (packetBuffer.size() > 0) {
            // get packet from buffer and parse it
            RealtimeScheduler::PacketBufferEntry packet =
                    *(packetBuffer.begin());
            packetBuffer.pop_front();
            switch (packet.func) {
            
            case RealtimeScheduler::PacketBufferEntry::DATA:
           
                if (packet.fd != appFd) {
                    error("XmlRpcInterface::handleMessage(): Received packet "
                          "from unknown socket!"); 
                }
                handleRealworldPacket(packet.data, packet.length);
                break;
            
            case RealtimeScheduler::PacketBufferEntry::FD_NEW:
                if (appFd != -1) {
                    error("XmlRpcInterface::handleMessage(): Multiple connections "
                          "not supported yet!"); 
                }
                appFd = packet.fd;
                break;

            case RealtimeScheduler::PacketBufferEntry::FD_CLOSE:
                if (packet.fd != appFd) {
                    error("XmlRpcInterface::handleMessage(): Trying to close unknown "
                          "connection!"); 
                }
             
                appFd = -1;
                _header = "";
                _request = "";
                _response = "";
                _connectionState = READ_HEADER;
            }
            if (packet.data) {
                delete[] packet.data;
            }
        }
    } else if (msg->isSelfMessage()) {
        // process rpc self-messages
        BaseRpcMessage* rpcMessage = dynamic_cast<BaseRpcMessage*>(msg);
        if (rpcMessage!=NULL) {
            internalHandleRpcMessage(rpcMessage);
            return;
        }
        // process all other self-messages
        if (msg == timeout_msg) {
            std::cout << "DHTXMLRealworldApp(): XML-RPC failed!" << endl;
            _response = generateFaultResponse("XML-RPC timeout", 22);
            _connectionState = WRITE_RESPONSE;
            if (!writeResponse() ) {
                scheduler->closeAppSocket(appFd);
            }
            return;
        }
    } else {
        // RPCs
        BaseRpcMessage* rpcMessage = dynamic_cast<BaseRpcMessage*>(msg);
        if (rpcMessage!=NULL) {
            internalHandleRpcMessage(rpcMessage);
            return;
        }
        // common API
        CommonAPIMessage* commonAPIMsg = dynamic_cast<CommonAPIMessage*>(msg);
        if (commonAPIMsg != NULL)
            handleCommonAPIPacket(commonAPIMsg);

        delete msg;
    }

}

void XmlRpcInterface::handleRealworldPacket(char *buf, uint32_t length)
{
    if (_connectionState == READ_HEADER) {
        if (!readHeader(buf, length)) {
            // discard data, if the header is invalid
            _header = "";
            _request = "";
            _response = "";
            _connectionState = READ_HEADER;
            return;
        }
    }

    if (_connectionState == READ_REQUEST)
        if (!readRequest(buf, length))
            return;

    if (_connectionState == WRITE_RESPONSE)
        if (!writeResponse() ) {
            scheduler->closeAppSocket(appFd);
            return;
        }

    return;
}

void XmlRpcInterface::handleRpcResponse(BaseResponseMessage* msg, int rpcId,
                                        simtime_t rtt)
{
    RPC_SWITCH_START(msg)
    RPC_ON_RESPONSE( Lookup ) {
        if (_connectionState != EXECUTE_REQUEST) break;

        XmlRpcValue resultValue;
        resultValue.setSize(_LookupResponse->getSiblingsArraySize());

        if (_LookupResponse->getIsValid() == true) {
            for (uint i=0; i < _LookupResponse->getSiblingsArraySize();
                    i++) {
                resultValue[i].setSize(3);
                resultValue[i][0] =
                _LookupResponse->getSiblings(i).ip.str();
                resultValue[i][1] =
                _LookupResponse->getSiblings(i).port;
                resultValue[i][2] =
                _LookupResponse->getSiblings(i).key.toString(16);
            }
            _response = generateResponse(resultValue.toXml());
        } else {
            std::cout << "XmlRpcInterface(): lookup() failed!" << endl;
            _response = generateFaultResponse("lookup() failed", 22);
        }

        _connectionState = WRITE_RESPONSE;
        if (! writeResponse() ) {
            scheduler->closeAppSocket(appFd);
        }
        break;
    }
    RPC_ON_RESPONSE(P2pnsRegister) {
        if (_connectionState != EXECUTE_REQUEST)
            break;

        XmlRpcValue resultValue;

        if (_P2pnsRegisterResponse->getIsSuccess() == true) {
            resultValue = 0;
            _response = generateResponse(resultValue.toXml());
        } else {
            std::cout << "XmlRpcInterface(): register() failed!" << endl;
            _response = generateFaultResponse("register() failed", 22);
        }

        _connectionState = WRITE_RESPONSE;
        if (!writeResponse() ) {
            scheduler->closeAppSocket(appFd);
        }
        break;
    }
    RPC_ON_RESPONSE(P2pnsResolve) {
        if (_connectionState != EXECUTE_REQUEST)
            break;

        XmlRpcValue resultValue;
        resultValue.setSize(1);
        
        if (_P2pnsResolveResponse->getIsSuccess() == true) {
            resultValue[0] = XmlRpcValue(
                                         &(*(_P2pnsResolveResponse->getAddress().begin())),
                                         _P2pnsResolveResponse->getAddress().size());
            _response = generateResponse(resultValue.toXml());
        } else {
            std::cout << "XmlRpcInterface(): resolve() failed!" << endl;
            _response = generateFaultResponse("resolve() failed", 22);
        }

        _connectionState = WRITE_RESPONSE;
        if (!writeResponse() ) {
            scheduler->closeAppSocket(appFd);
        }
        break;
    }
    RPC_ON_RESPONSE(DHTputCAPI) {
        if (_connectionState != EXECUTE_REQUEST)
            break;

        XmlRpcValue resultValue;

        if (_DHTputCAPIResponse->getIsSuccess() == true) {
            resultValue = 0;
            _response = generateResponse(resultValue.toXml());
        } else {
            std::cout << "XmlRpcInterface(): put() failed!" << endl;
            _response = generateFaultResponse("put() failed", 22);
        }

        _connectionState = WRITE_RESPONSE;
        if (!writeResponse() ) {
            scheduler->closeAppSocket(appFd);
        }
        break;
    }
    RPC_ON_RESPONSE(DHTgetCAPI) {
        if (_connectionState != EXECUTE_REQUEST)
            break;

        XmlRpcValue resultValue;
        resultValue.setSize(2);
        resultValue[0].setSize(1);

        if (_DHTgetCAPIResponse->getIsSuccess() == true) {
            resultValue[0][0] = XmlRpcValue(
                      &(*(_DHTgetCAPIResponse->getValue().begin())),
                      _DHTgetCAPIResponse->getValue().size());
            resultValue[1] = std::string();
            _response = generateResponse(resultValue.toXml());
        } else {
            std::cout << "XmlRpcInterface(): get() failed!" << endl;
            _response = generateFaultResponse("get() failed", 22);
        }

        _connectionState = WRITE_RESPONSE;
        if (!writeResponse() ) {
            scheduler->closeAppSocket(appFd);
        }
        break;
    }
    RPC_SWITCH_END( )
}

void XmlRpcInterface::handleCommonAPIPacket(cMessage *msg)
{
    error("DHTXMLRealworldApp::handleCommonAPIPacket(): Unknown Packet!");
}

bool XmlRpcInterface::readHeader(char* buf, uint32_t length)
{
    // Read available data
    bool eof = false;

    _header.append(std::string(buf, length));

    if (length <= 0) {
        // Its only an error if we already have read some data
        if (_header.length() > 0)
            XmlRpcUtil::error("XmlRpcServerConnection::readHeader: error "
                "while reading header.");
        return false;
    }

    XmlRpcUtil::log(4, "XmlRpcServerConnection::readHeader: read %d bytes.",
                    _header.length());
    char *hp = (char*)_header.c_str(); // Start of header
    char *ep = hp + _header.length(); // End of string
    char *bp = 0; // Start of body
    char *lp = 0; // Start of content-length value
    char *kp = 0; // Start of connection value

    for (char *cp = hp; (bp == 0) && (cp < ep); ++cp) {
        if ((ep - cp > 16) && (strncasecmp(cp, "Content-length: ", 16) == 0))
            lp = cp + 16;
        else if ((ep - cp > 12) && (strncasecmp(cp, "Connection: ", 12) == 0))
            kp = cp + 12;
        else if ((ep - cp > 4) && (strncmp(cp, "\r\n\r\n", 4) == 0))
            bp = cp + 4;
        else if ((ep - cp > 2) && (strncmp(cp, "\n\n", 2) == 0))
            bp = cp + 2;
    }

    // If we haven't gotten the entire header yet, return (keep reading)
    if (bp == 0) {
        // EOF in the middle of a request is an error, otherwise its ok
        if (eof) {
            XmlRpcUtil::log(4, "XmlRpcServerConnection::readHeader: EOF");
            if (_header.length() > 0)
                XmlRpcUtil::error("XmlRpcServerConnection::readHeader: EOF while reading header");
            return false; // Either way we close the connection
        }

        return true; // Keep reading
    }

    // Decode content length
    if (lp == 0) {
        XmlRpcUtil::error("XmlRpcServerConnection::readHeader: No Content-length specified");
        return false; // We could try to figure it out by parsing as we read, but for now...
    }

    _contentLength = atoi(lp);
    if (_contentLength <= 0) {
        XmlRpcUtil::error(
                          "XmlRpcServerConnection::readHeader: Invalid Content-length specified (%d).",
                          _contentLength);
        return false;
    }

    XmlRpcUtil::log(
                     3,
                    "XmlRpcServerConnection::readHeader: specified content length is %d.",
                    _contentLength);

    // Otherwise copy non-header data to request buffer and set state to read request.
    _request = bp;

    // Parse out any interesting bits from the header (HTTP version, connection)
    _keepAlive = true;
    if (_header.find("HTTP/1.0") != std::string::npos) {
        if (kp == 0 || strncasecmp(kp, "keep-alive", 10) != 0)
            _keepAlive = false; // Default for HTTP 1.0 is to close the connection
    } else {
        if (kp != 0 && strncasecmp(kp, "close", 5) == 0)
            _keepAlive = false;
    }
    XmlRpcUtil::log(3, "KeepAlive: %d", _keepAlive);

    _header = "";
    _connectionState = READ_REQUEST;
    return true; // Continue monitoring this source
}

bool XmlRpcInterface::readRequest(char *buf, uint32_t length)
{
    // If we dont have the entire request yet, read available data
    if (int(_request.length()) < _contentLength) {
        bool eof = false;

        _request.append(std::string(buf, length));

        if (length <= 0) {
            XmlRpcUtil::error("XmlRpcServerConnection::readRequest: read error.");
            return false;
        }

        // If we haven't gotten the entire request yet, return (keep reading)
        if (int(_request.length()) < _contentLength) {
            if (eof) {
                XmlRpcUtil::error("XmlRpcServerConnection::readRequest: EOF while reading request");
                return false; // Either way we close the connection
            }
            return true;
        }
    }

    // Otherwise, parse and dispatch the request
    XmlRpcUtil::log(3, "XmlRpcServerConnection::readRequest read %d bytes.",
                    _request.length());
    //XmlRpcUtil::log(5, "XmlRpcServerConnection::readRequest:\n%s\n", _request.c_str());

    _connectionState = WRITE_RESPONSE;

    return true; // Continue monitoring this source
}

bool XmlRpcInterface::writeResponse()
{
    cancelEvent(timeout_msg);

    if (_response.length() == 0) {
        _response = executeRequest(_request);
        _bytesWritten = 0;

        if (_connectionState == EXECUTE_REQUEST)
            return true;

        if (_response.length() == 0) {
            XmlRpcUtil::error("XmlRpcServerConnection::writeResponse: empty response.");
            return false;
        }
    }

    // Try to write the response
    int curBytesWritten = scheduler->sendBytes(_response.c_str(),
                                               _response.length(), 0, 0, true, appFd);

    if (curBytesWritten <= 0) {
        XmlRpcUtil::error("XmlRpcServerConnection::writeResponse: write error.");
        return false;
    } else {
        _bytesWritten += curBytesWritten;
    }

    XmlRpcUtil::log(
                     3,
                    "XmlRpcServerConnection::writeResponse: wrote %d of %d bytes.",
                    _bytesWritten, _response.length());

    // Prepare to read the next request
    if (_bytesWritten == int(_response.length())) {
        _header = "";
        _request = "";
        _response = "";
        _connectionState = READ_HEADER;
    }

    return _keepAlive; // Continue monitoring this source if true
}
