//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//

/**
 * @file SimpleNetConfigurator.cc
 * @author Stephan Krause
 * @author Bernhard Heep (migrateRandomNode)
 */

#include <omnetpp.h>
#include <vector>
#include <map>
#include <iostream>
#include <fstream>
#include <sstream>

#include "InterfaceTable.h"
#include "InterfaceEntry.h"
#include "IPv4InterfaceData.h"
#include "IPAddressResolver.h"
#include "cenvir.h"
#include "cxmlelement.h"

#include "SimpleUDP.h"

#include "SimpleNetConfigurator.h"


NodeRecord::NodeRecord(int dim)
{
    coords = new double[dim];
}

NodeRecord::~NodeRecord()
{
    //delete[] coords;
}

Define_Module(SimpleNetConfigurator);

void SimpleNetConfigurator::initializeUnderlay(int stage)
{
    if (stage != MAX_STAGE_UNDERLAY)
        return;

    // fetch some parameters
    moduleType = findModuleType(par("overlayTerminalType")); // not used anymore!
    fixedNodePositions = par("fixedNodePositions");

    // set maximum coordinates and send queue length
    fieldSize = par("fieldSize");
    //fieldDimension = 2;
    sendQueueLength = par("sendQueueLength");

    SimpleNodeEntry::setFieldSize(fieldSize); //deprecated
    SimpleNodeEntry::setSendQueueLength(sendQueueLength); //deprecated

    // get parameter of sourcefile's name
    nodeCoordinateSource = par("nodeCoordinateSource");

    // check if exists and process xml-file containing coordinates
    std::ifstream check_for_xml_file(nodeCoordinateSource);
    if (check_for_xml_file) {
        useXmlCoords = 1;
        uint max_coord = parseCoordFile(nodeCoordinateSource);
        SimpleNodeEntry::setFieldSize(max_coord);

        EV << "[SimpleNetConfigurator::initializeUnderlay()]\n"
           << "    Using '" << nodeCoordinateSource
           << "' as Coordinate-Sourcefile" << endl;
    } else {
        useXmlCoords = 0;
        dimensions = 2;
        EV << "[SimpleNetConfigurator::initializeUnderlay()]\n"
           << "    Using conventional (random) coordinates for placing nodes!\n"
           << "    (An XML-Sourcefile could not be opened)" << endl;
    }
    check_for_xml_file.close();

    // FIXME get address from parameter
    nextFreeAddress = 0x1000001;

    // count the overlay clients
    overlayTerminalCount = 0;
    // Now done in base class   churnGenerator->setOverlayTerminalCount(overlayTerminalCount);

    numCreated = 0;
    numKilled = 0;

    //Now done in base class    churnGenerator->initializeChurn();
}

TransportAddress* SimpleNetConfigurator::createNode(NodeType type,
                                                    bool initialize)
{
    Enter_Method_Silent();
    // derive overlay node from ned
    cModuleType* moduleType = findModuleType(type.terminalType.c_str());

    cModule* node = moduleType->create("overlayTerminal", parentModule());

    node->par("overlayType").setStringValue(type.overlayType.c_str());
    node->par("tier1Type").setStringValue(type.tier1Type.c_str());
    node->par("tier2Type").setStringValue(type.tier2Type.c_str());
    node->par("tier3Type").setStringValue(type.tier3Type.c_str());

    node->setDisplayString("i=device/wifilaptop_l;i2=block/circle_s");
    node->buildInside();
    node->scheduleStart(simulation.simTime());

    for (int i = 0; i < MAX_STAGE_UNDERLAY + 1; i++) {
        node->callInitialize(i);
    }

    // FIXME use only IPv4?
    IPvXAddress addr = IPAddress(nextFreeAddress++);

    cChannelType* chan = findChannelType(type.channelTypes[intuniform(0,
            type.channelTypes.size()
            - 1)].c_str());
    if (!chan)
        opp_error("Could not find Channel Type. Most likely parameter "
                  "channelTypes does not match the channels defined in "
                  "channels.ned");
    SimpleNodeEntry* entry;

    if (!useXmlCoords) {
        entry = new SimpleNodeEntry(node, chan, fieldSize, sendQueueLength);
    } else {
        // stop with errormessage if no more unused nodes available
        if (unusedNodeRecords.size() < 1) {
            opp_error("Cannot create any more Nodes. "
                      "Provide %s-file with more Nodes!\n",
                      nodeCoordinateSource);
        }

        //get random node from set of unused        
        int volunteer = intuniform(0, unusedNodeRecords.size()-1);
        entry = new SimpleNodeEntry(node, chan, dimensions,
                unusedNodeRecords[volunteer].coords);

        //insert IP-address into noderecord used
        unusedNodeRecords[volunteer].ip = addr;

        // add noderecord to assigned...
        assignedNodeRecords.insert(make_pair(addr, unusedNodeRecords[volunteer]));
        // ... and erase from unused
        unusedNodeRecords.erase(unusedNodeRecords.begin()+volunteer);

    }

    node->gate("out",0)->setChannel(chan->create("outChan"));

    SimpleUDP* simple = check_and_cast<SimpleUDP*>(node->submodule("udp"));
    simple->setNodeEntry(*entry);

    // Add pseudo-Interface to node's interfaceTable
    IPv4InterfaceData* ifdata = new IPv4InterfaceData;
    ifdata->setInetAddress(addr.get4());
    ifdata->setNetmask(IPAddress("255.255.255.255"));
    InterfaceEntry* e = new InterfaceEntry;
    e->setName("dummy interface");
    e->setIPv4Data(ifdata);

    IPAddressResolver().interfaceTableOf(node)->addInterface(e, NULL);

    // create meta information
    SimpleInfo* info = new SimpleInfo(type.typeID, node->id());
    info->setEntry(entry);

    // append index to module name
    char buf[80];
    sprintf(buf, "overlayTerminal[%i]", numCreated);
    node->setName(buf);

    //add node to bootstrap oracle
    bootstrapOracle->addPeer(addr, info);

    // if the node was not created during startup we have to
    // finish the initialization process manually
    if ( !initialize) {
        for (int i = MAX_STAGE_UNDERLAY + 1; i < NUM_STAGES_ALL; i++) {
            node->callInitialize(i);
        }
    }

    //show ip... todo: migrate   
    if (fixedNodePositions && ev.isGUI()) {
        node->displayString().insertTag("p");
        node->displayString().setTagArg("p", 0, (long int)entry->getX());
        node->displayString().setTagArg("p", 1, (long int)entry->getY());
        node->displayString().insertTag("t", 0);
        node->displayString().setTagArg("t", 0, addr.str().c_str());
        node->displayString().setTagArg("t", 1, "l");
    }

    overlayTerminalCount++;
    numCreated++;

    churnGenerator[type.typeID - 1]->terminalCount++;

    TransportAddress *address = new TransportAddress(addr);

    // update display
    setDisplayString();

    return address;
}

uint SimpleNetConfigurator::parseCoordFile(const char* nodeCoordinateSource)
{
    cXMLElement * rootElement = ev.getXMLDocument(nodeCoordinateSource);

    // get number of dimensions from attribute of xml rootelement
    dimensions = atoi(rootElement->getAttribute("dimensions"));
    EV << "[SimpleNetConfigurator::parseCoordFile()]\n"
       << "    using " << dimensions << " dimensions" << endl;

    double max_coord = 0;
    for (cXMLElement *tmpElement = rootElement->getFirstChild(); tmpElement;
         tmpElement = tmpElement->getNextSibling()) {

        // get "ip" and "isRoot" from Attributes (not needed yet)
        /*
         const char* str_ip = tmpElement->getAttribute("ip");        
         int tmpIP = 0;
         if (str_ip) sscanf(str_ip, "%x", &tmpIP);
         bool tmpIsRoot = atoi(tmpElement->getAttribute("isroot"));
         */

        // create tmpNode to be added to vector
        NodeRecord tmpNode = NodeRecord(dimensions);

        // get coords from childEntries and fill tmpNodes Array
        int i = 0;
        for (cXMLElement *coord = tmpElement->getFirstChild(); coord;
             coord = coord->getNextSibling()) {
            tmpNode.coords[i] = atof(coord->getNodeValue());
            if (fabs(tmpNode.coords[i]) > max_coord)
                max_coord = fabs(tmpNode.coords[i]);
            i++;
        }

        // add to vector
        unusedNodeRecords.push_back(tmpNode);
    } // now all nodes from xml-file are in vector unusedRootNodes

    EV << "[SimpleNetConfigurator::parseCoordFile()]\n    "
       << unusedNodeRecords.size()
       << " nodes added to vector \"unusedNodeRecords\"." << endl;

    return (uint)ceil(max_coord);
}

void SimpleNetConfigurator::preKillNode(NodeType type, TransportAddress* addr)
{
    Enter_Method_Silent();

    SimpleNodeEntry* entry= NULL;
    SimpleInfo* info;

    if (addr != NULL) {
        info = dynamic_cast<SimpleInfo*>(bootstrapOracle->getPeerInfo(*addr));
        if (info != NULL) {
            entry = info->getEntry();
        } else
            opp_error("SimpleNetConfigurator: Trying to pre kill node "
                "with nonexistant TransportAddress!");
    } else {
        info = dynamic_cast<SimpleInfo*>(bootstrapOracle
            ->getRandomPeerInfo(type.typeID));
        entry = info->getEntry();
    }

    uint32_t effectiveType = info->getTypeID();
    cGate* gate = entry->getGate();

    cModule* node = gate->ownerModule()->parentModule();

    if (scheduledID.count(node->id()))
        return;

    //remove node from bootstrap oracle
    bootstrapOracle->removePeer(IPAddressResolver().addressOf(node));

    // put node into the kill list and schedule a message for final removal
    // of the node
    killList.push_front(IPAddressResolver().addressOf(node));
    scheduledID.insert(node->id());

    overlayTerminalCount--;
    numKilled++;

    churnGenerator[effectiveType - 1]->terminalCount--;

    // update display
    setDisplayString();

    // inform the notification board about the removal
    NotificationBoard* nb = check_and_cast<NotificationBoard*>(node->
    submodule("notificationBoard"));
    nb->fireChangeNotification(NF_OVERLAY_NODE_LEAVE);

    double random = uniform(0, 1);
    if (random < gracefulLeavePropability) {
        nb->fireChangeNotification(NF_OVERLAY_NODE_GRACEFUL_LEAVE);
    }

    cMessage* msg = new cMessage();
    scheduleAt(simulation.simTime() + gracefulLeaveDelay, msg);
}

void SimpleNetConfigurator::migrateNode(NodeType type, TransportAddress* addr)
{
    Enter_Method_Silent();

    SimpleNodeEntry* entry= NULL;

    if (addr != NULL) {
        SimpleInfo* info =
                dynamic_cast<SimpleInfo*>(bootstrapOracle->getPeerInfo(*addr));
        if (info != NULL) {
            entry = info->getEntry();
        } else
            opp_error("SimpleNetConfigurator: Trying to migrate node with "
                      "nonexistant TransportAddress!");
    } else {
        SimpleInfo* info = dynamic_cast<SimpleInfo*>(bootstrapOracle->
        getRandomPeerInfo(type.typeID));
        entry = info->getEntry();
    }

    cGate* gate = entry->getGate();
    cModule* node = gate->ownerModule()->parentModule();

    // do not migrate node that is already scheduled
    if (scheduledID.count(node->id()))
        return;

    IPvXAddress tmp_ip = IPAddressResolver().addressOf(node);

    //remove node from bootstrap oracle
    bootstrapOracle->killPeer(tmp_ip);

    node->bubble("I am migrating!");

    // FIXME use only IPv4?
    IPvXAddress address = IPAddress(nextFreeAddress++);

    //std::cout << "migration @ " << tmp_ip << " --> " << address << std::endl;

    SimpleNodeEntry* newentry;

    if (useXmlCoords) {
        // save coords before removal from assignedNodeRecords
        double * tmp_coords = new double[dimensions];
        map<IPvXAddress, NodeRecord>::iterator iter =
            assignedNodeRecords.find(tmp_ip);
        if (iter != assignedNodeRecords.end()) {
            for (int i=0; i<dimensions; i++) {
                tmp_coords[i] = (iter->second).coords[i];
            }
        } else {
            opp_error("SimpleNetConfigurator: couldn't find ip in "
                "assignedNodeRecords!");
        }

        // remove NodeRecord from map with used ones    
        assignedNodeRecords.erase(tmp_ip);

        // create new NR to re-insert with new address
        NodeRecord tmpNR = NodeRecord(dimensions);
        for (int i=0; i<dimensions; i++) {
            tmpNR.coords[i] = tmp_coords[i];
        }
        tmpNR.ip = address;

        // insert NodeRecord again with new address
        assignedNodeRecords.insert(make_pair(address, tmpNR) );

        newentry = new SimpleNodeEntry(node,
            findChannelType(channelTypes[intuniform(0, channelTypes.size() - 1)].c_str()),
            dimensions, tmp_coords);
        
        delete[] tmp_coords;
    } else {
        newentry = new SimpleNodeEntry(node,
            findChannelType(channelTypes[intuniform(0, channelTypes.size() - 1)].c_str()),
            fieldSize, sendQueueLength);
    }

    SimpleUDP* simple = check_and_cast<SimpleUDP*>(gate->ownerModule());
    simple->setNodeEntry(*newentry);

    InterfaceEntry* ie = IPAddressResolver().interfaceTableOf(node)->
    interfaceByName("dummy interface");

    delete ie->ipv4();

    // Add pseudo-Interface to node's interfaceTable
    IPv4InterfaceData* ifdata = new IPv4InterfaceData;
    ifdata->setInetAddress(address.get4());
    ifdata->setNetmask(IPAddress("255.255.255.255"));
    ie->setIPv4Data(ifdata);

    // create meta information
    SimpleInfo* newinfo = new SimpleInfo(type.typeID, node->id());
    newinfo->setEntry(newentry);

    //add node to bootstrap oracle
    bootstrapOracle->addPeer(address, newinfo);

    // inform the notofication board about the migration
    NotificationBoard* nb =
        check_and_cast<NotificationBoard*>(node->submodule("notificationBoard"));
    nb->fireChangeNotification(NF_OVERLAY_TRANSPORTADDRESS_CHANGED);
}

void SimpleNetConfigurator::handleTimerEvent(cMessage* msg)
{
    Enter_Method_Silent();

    // get next scheduled node and remove it from the kill list
    IPvXAddress addr = killList.back();
    killList.pop_back();

    SimpleNodeEntry* entry= NULL;

    SimpleInfo* info =
            dynamic_cast<SimpleInfo*>(bootstrapOracle->getPeerInfo(addr));
    if (info != NULL) {
        entry = info->getEntry();
    } else
        opp_error("SimpleNetConfigurator: Trying to kill node with "
                  "nonexistant TransportAddress!");

    cGate* gate = entry->getGate();
    cModule* node = gate->ownerModule()->parentModule();

    scheduledID.erase(node->id());
    bootstrapOracle->killPeer(addr);

    InterfaceEntry* ie = IPAddressResolver().interfaceTableOf(node)->
    interfaceByName("dummy interface");
    delete ie->ipv4();

    if (useXmlCoords) {
        // remove node from assignedNodeRecords
        map<IPvXAddress, NodeRecord>::iterator iter =
            assignedNodeRecords.find(addr);
        if (iter != assignedNodeRecords.end()) {
            assignedNodeRecords.erase(iter);
        } else {
            opp_error("SimpleNetConfigurator::handleTimerEvent(): "
                      "couldn't find ip in assignedNodeRecords!");
        }
    }

    node->callFinish();
    node->deleteModule();

    delete msg;
}

void SimpleNetConfigurator::setDisplayString()
{
    //
    // Updates the statistics display string.
    //
    char buf[80];
    sprintf(buf, "%i overlay clients", overlayTerminalCount);
    displayString().setTagArg("t", 0, buf);
}

void SimpleNetConfigurator::finishUnderlay()
{
    // statistics
    recordScalar("Terminals added", numCreated);
    recordScalar("Terminals removed", numKilled);

    struct timeval now, diff;
    gettimeofday(&now, NULL);
    timersub(&now, &initFinishedTime, &diff);
    printf("Simulation time: %li.%06li\n", diff.tv_sec, diff.tv_usec);
}
