//
// Copyright (C) 2006 Institut fuer Telematik, Universitaet Karlsruhe (TH)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//

/**
 * @file SimpleNetConfigurator.h
 * @author Stephan Krause
 */

#ifndef __SIMPLENETCONFIGURATOR_H__
#define __SIMPLENETCONFIGURATOR_H__


#include <omnetpp.h>
#include <BasicModule.h>
#include <deque>
#include <set>

#include <UnderlayConfigurator.h>
#include <InitStages.h>
#include <SimpleInfo.h>

class NodeRecord
{
public:
    NodeRecord(int dim);
    ~NodeRecord();
    void debugOutput(int dim);

    double * coords;
    IPvXAddress ip;
};


/**
 * Sets up a SimpleNetwork
 */

/**
 * Sets up a SimpleNetwork.
 * Adds overlay nodes to the network in init phase
 * and adds/removes/migrates overlay nodes after init phase.
 */
class SimpleNetConfigurator : public UnderlayConfigurator
{
public:
    /**
     * Creates an overlay node
     *
     * @param type NodeType of the node to create
     * @param initialize are we in init phase?
     */
    virtual TransportAddress* createNode(NodeType type, bool initialize=false);

    /**
     * Notifies and schedules overlay nodes for removal.
     *
     * @param type NodeType of the node to remove
     * @param addr NULL for random node
     */
    virtual void preKillNode(NodeType type, TransportAddress* addr=NULL);

    /**
     * Migrates overlay nodes from one access net to another.
     *
     * @param type NodeType of the node to migrate
     * @param addr NULL for random node
     */
    virtual void migrateNode(NodeType type, TransportAddress* addr=NULL);

    uint getFieldSize() { return fieldSize; };
    uint getFieldDimension() { return dimensions; };
    uint getSendQueueLenghth() { return sendQueueLength; };

protected:

    /**
     * Enables access to the globalHashMap, sets some parameters and adds the
     * initial number of nodes to the network
     *
     * @param stage the phase of the initialisation
     */
    void initializeUnderlay(int stage);

    void handleTimerEvent(cMessage* msg);

    /**
     * Saves statistics, prints simulation time
     */
    void finishUnderlay();

    /**
     * Prints statistics
     */
    void setDisplayString();
    uint parseCoordFile(const char * nodeCoordinateSource);
    
    uint32 nextFreeAddress; /**< adress of the node that will be created next */
    cModuleType* moduleType; /**< type of overlay terminals */
    std::deque<IPvXAddress> killList; //!< stores nodes scheduled to be killed
    std::set<int> scheduledID; //!< stores nodeIds to prevent migration of prekilled nodes

    uint sendQueueLength; /**< send queue length of overlay terminals */
    uint fieldSize;
    int dimensions;
    bool fixedNodePositions;

    bool useXmlCoords;
    const char* nodeCoordinateSource;
    
    // each node is either in unused- or in assignedNodeRecords
    vector<NodeRecord> unusedNodeRecords;
    map<IPvXAddress, NodeRecord> assignedNodeRecords;
    
    // statistics
    int numCreated; /**< number of overall created overlay terminals */
    int numKilled; /**< number of overall killed overlay terminals */
};

#endif
