//
// Copyright (C) 2006 Institut fuer Telematik, Universitaet Karlsruhe (TH)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//

/**
 * @file SimpleNodeEntry.cc
 * @author Bernhard Heep
 */

#include <sstream>

#include "SimpleNodeEntry.h"
#include "IPAddressResolver.h"


uint SimpleNodeEntry::fieldSize;
uint SimpleNodeEntry::sendQueueLength;

SimpleNodeEntry::SimpleNodeEntry(cModule* node,
                                 cChannelType* type,
                                 uint fieldSize,
                                 uint sendQueueLength)
{
    //uint fieldSize = size;
    //uint sendQueueLength = queueLength;

    ingate = node->submodule("udp")->gate("network_in");
	
    dimensions = 2;
    coordinates = new double[2];
    
    //use random values as coordinates
    coordinates[0] = uniform(0, fieldSize) - fieldSize / 2;
    coordinates[1] = uniform(0, fieldSize) - fieldSize / 2;

    cBasicChannel* temp = dynamic_cast<cBasicChannel*>(type->create("temp"));

    bandwidth = temp->datarate();
    errorRate = temp->error();
    accessDelay = temp->delay();

    txMaxQueueTime = ((float)sendQueueLength * 8) / bandwidth;
    txFinished = simulation.simTime(); // 0?

    delete temp;
}

SimpleNodeEntry::SimpleNodeEntry(cModule* node,
                                 cChannelType* type,
                                 int my_dim,
                                 double * my_coords)
{
    ingate = node->submodule("udp")->gate("network_in");

    dimensions = my_dim;
    coordinates = new double[dimensions];
    
    for (int i=0; i<dimensions; i++) {
        coordinates[i] = my_coords[i];
    }

    cBasicChannel* temp = dynamic_cast<cBasicChannel*>(type->create("temp"));

    bandwidth = temp->datarate();
    errorRate = temp->error();
    accessDelay = temp->delay();

    txMaxQueueTime = ((float)sendQueueLength * 8) / bandwidth;
    txFinished = simulation.simTime(); // 0?

    delete temp;
}

float SimpleNodeEntry::operator-(const SimpleNodeEntry& entry) const
{
    double sum_of_squares = 0;
    for (int i = 0; i < dimensions; i++) {
        sum_of_squares += pow(coordinates[i] - entry.coordinates[i], 2);
    }
    return sqrt(sum_of_squares);
}

SimpleNodeEntry::SimpleDelay SimpleNodeEntry::calcDelay(const SimpleUDPPacket& msg, const SimpleNodeEntry& dest)
{
    if (uniform(0, 1) < errorRate || uniform(0, 1) < dest.errorRate)
        return SimpleDelay(0, false);

    simtime_t now = simulation.simTime();
    simtime_t bandwidthDelay= ((msg.byteLength() * 8) / bandwidth);
    simtime_t newTxFinished = fmax(txFinished, now) + bandwidthDelay;

    // send queue
    if ((newTxFinished > now + txMaxQueueTime) && (txMaxQueueTime != 0)) {
        EV << "[SimpleNodeEntry::calcDelay()]\n"
           << "    Send queue overrun"
           << "\n    newTxFinished = fmax(txFinished, now) + bandwidthDelay"
           << "\n    newTxFinished = " << newTxFinished
           << "\n    txFinished = " << txFinished
           << "\n    now = " << now
           << "\n    bandwidthDelay = " << bandwidthDelay
           << "\n    (newTxFinished > now + txMaxQueueTime) == true"
           << "\n    txMaxQueueTime = " << txMaxQueueTime
           << endl;
        return SimpleDelay(0, false);
    }

    txFinished = newTxFinished;

    simtime_t destBandwidthDelay = (msg.byteLength() * 8) / dest.bandwidth;
    simtime_t coordDelay = 0.001 * (*this - dest);

    return SimpleDelay(txFinished - now
		       + accessDelay
		       + coordDelay
		       + destBandwidthDelay + dest.accessDelay, true);
}

std::string SimpleNodeEntry::info() const
{
    std::ostringstream str;
    str << *this;
    return str.str();
}

std::ostream& operator<<(std::ostream& out, const SimpleNodeEntry& entry)
{
    out << "(x:" << entry.coordinates[0] << ", y:" << entry.coordinates[1]
        << ")\nbandwidth = " << entry.bandwidth
        << ",\ndelay = " << entry.accessDelay
        << ",\ntxMaxQueueTime = " << entry.txMaxQueueTime
        << ",\ntxFinished = " << entry.txFinished;

    return out;
}
