//
// Copyright (C) 2006 Institut fuer Telematik, Universitaet Karlsruhe (TH)
//
// This program is free software; you can redistribute it and/or
// modify it under the terms of the GNU General Public License
// as published by the Free Software Foundation; either version 2
// of the License, or (at your option) any later version.
//
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
//

/**
 * @file realtimescheduler.cc
 * @author Stephan Krause
 */

#include "realtimescheduler.h"

inline std::ostream& operator<<(std::ostream& ev, const timeval& tv)
{
    return ev << (unsigned long)tv.tv_sec << "s" << tv.tv_usec << "us";
}

//---

RealtimeScheduler::RealtimeScheduler() : cScheduler()
{
    FD_ZERO(&all_fds);
    maxfd = -1;
    netw_fd = -1;
    additional_fd = -1;
}

RealtimeScheduler::~RealtimeScheduler()
{ }

void RealtimeScheduler::startRun()
{
    if (initsocketlibonce()!=0)
        throw new cRuntimeError("RealtimeScheduler: Cannot initialize socket library");

    gettimeofday(&baseTime, NULL);

    appModule = NULL;
    appNotificationMsg = NULL;
    module = NULL;
    notificationMsg = NULL;

    appConnectionLimit = ev.config()->getAsInt("ExternalApp", "connectionLimit");

    if (initializeNetwork()) {
        opp_error("realtimeScheduler error: initializeNetwork failed\n");
    }
}

void RealtimeScheduler::endRun()
{}

void RealtimeScheduler::executionResumed()
{
    gettimeofday(&baseTime, NULL);
    baseTime = timeval_substract(baseTime, sim->simTime());
}

void RealtimeScheduler::setInterfaceModule(cModule *mod, cMessage *notifMsg,  PacketBuffer* buffer, int mtu, bool isApp)
{
    if (!mod || !notifMsg || !buffer)
        throw new cRuntimeError("RealtimeScheduler: setInterfaceModule(): arguments must be non-NULL");

    if (!isApp) {
        if (module)
            throw new cRuntimeError("RealtimeScheduler: setInterfaceModule() already called");

        module = mod;
        notificationMsg = notifMsg;
        packetBuffer = buffer;
        buffersize = mtu;
    } else {
        if (appModule)
            throw new cRuntimeError("RealtimeScheduler: setInterfaceModule() already called");

        appModule = mod;
        std::cout << "RealtimeScheduler::setInferfaceModule: " << appModule->fullPath() << std::endl;
        appNotificationMsg = notifMsg;
        appPacketBuffer = buffer;
        appBuffersize = mtu;
    }
}

bool RealtimeScheduler::receiveWithTimeout(long usec)
{
    bool newEvent = false;
    // prepare sets for select()
    fd_set readFD;
    readFD = all_fds;

    timeval timeout;
    timeout.tv_sec = 0;
    timeout.tv_usec = usec;

    if (select(FD_SETSIZE, &readFD, NULL, NULL, &timeout) > 0) {
        // Read on all sockets with data
        for( int fd = 0; fd <= maxfd; fd++) {
            if( FD_ISSET(fd, &readFD)) {
                // Incoming data on netw_fd
                if( fd == netw_fd ) {
                    char* buf = new char[buffersize];
                    int nBytes;

                    // FIXME: Ugly. But we want to support IPv4 and IPv6 here, so we
                    // reserve enough space for the "bigger" address.
                    sockaddr* from = (sockaddr*) new sockaddr_in6;
                    socklen_t addrlen = sizeof(sockaddr_in6);

                    // FIXME: Ugly...
                    getsockname(netw_fd, from, &addrlen);
                    if ( from->sa_family != SOCK_DGRAM ) {
                        delete from;
                        from = 0;
                        addrlen = 0;
                        nBytes = read(netw_fd, buf, buffersize);
                    } else {
                        addrlen = sizeof(sockaddr_in6);
                        nBytes = recvfrom(netw_fd, buf, buffersize, 0, from, &addrlen);
                    }

                    if (nBytes < 0) {
                        ev << "[RealtimeScheduler::receiveWithTimeout()]\n"
                            << "    Error reading from network: " << strerror(errno)
                            << endl;
                        opp_error("Read from network device returned an error");
                    } else if (nBytes == 0) {
                        opp_error("network device closed...");
                    } else {
                        // write data to buffer
                        ev << "[RealtimeScheduler::receiveWithTimeout()]\n"
                            << "    Received " << nBytes << " bytes"
                            << endl;
                        packetBuffer->push_back(PacketBufferEntry(buf, nBytes, from, addrlen));
                        // schedule notificationMsg for the interface module
                        sendNotificationMsg(notificationMsg, module);
                        newEvent = true;
                    }
                } else if ( fd == additional_fd ) {
                    // Data on additional FD
                    additionalFD();
                    newEvent = true;
                } else {
                    // Data on app FD
                    char* buf = new char[appBuffersize];
                    int nBytes = read(fd, buf, appBuffersize);
                    if (nBytes < 0) {
                        opp_error("Read from network device returned an error (App)");
                    } else if (nBytes == 0) {
                        // Application closed Socket
                        ev << "[RealtimeScheduler::receiveWithTimeout()]\n"
                            << "    Application closed socket"
                            << endl;

                        closeAppSocket(fd);
                        newEvent = true;
                    } else {
                        // write data to buffer
                        ev << "[RealtimeScheduler::receiveWithTimeout()]\n"
                            << "    Received " << nBytes << " bytes"
                            << endl;
                        appPacketBuffer->push_back(PacketBufferEntry(buf, nBytes, PacketBufferEntry::DATA, fd));
                        // schedule notificationMsg for the interface module
                        sendNotificationMsg(appNotificationMsg, appModule);
                        newEvent = true;
                    }
                }
            }
        }
    }
    return newEvent;
}

int RealtimeScheduler::receiveUntil(const timeval& targetTime)
{
    // if there's more than 200ms to wait, wait in 100ms chunks
    // in order to keep UI responsiveness by invoking ev.idle()
    timeval curTime;
    gettimeofday(&curTime, NULL);
    while (targetTime.tv_sec-curTime.tv_sec >=2 ||
            timeval_diff_usec(targetTime, curTime) >= 200000) {
        if (receiveWithTimeout(100000)) { // 100ms
            if (ev.idle()) return -1;
            return 1;
        }
        if (ev.idle()) return -1;
        gettimeofday(&curTime, NULL);
    }

    // difference is now at most 100ms, do it at once
    long usec = timeval_diff_usec(targetTime, curTime);
    if (usec>0)
        if (receiveWithTimeout(usec)) {
            if (ev.idle()) return -1;
            return 1;
        }
    if (ev.idle()) return -1;
    return 0;
}

cMessage *RealtimeScheduler::getNextEvent()
{
    // assert that we've been configured
    if (!module)
        throw new cRuntimeError("RealtimeScheduler: setInterfaceModule() not called: it must be called from a module's initialize() function");
    // FIXME: reimplement sanity check
//    if (app_fd >= 0 && !appModule)
//        throw new cRuntimeError("RealtimeScheduler: setInterfaceModule() not called from application: it must be called from a module's initialize() function");

    // calculate target time
    timeval targetTime;
    cMessage *msg = sim->msgQueue.peekFirst();
    if (!msg) {
        // if there are no events, wait until something comes from outside
        // TBD: obey simtimelimit, cpu-time-limit
        targetTime.tv_sec = LONG_MAX;
        targetTime.tv_usec = 0;
    } else {
        // use time of next event
        simtime_t eventSimtime = msg->arrivalTime();
        targetTime = timeval_add(baseTime, eventSimtime);
    }

    // if needed, wait until that time arrives
    timeval curTime;
    gettimeofday(&curTime, NULL);
    if (timeval_greater(targetTime, curTime)) {
        int status = receiveUntil(targetTime);
        if (status == -1) {
            printf("WARNING: receiveUntil returned -1 (user interrupt)\n");
            return NULL; // interrupted by user
        } else if (status == 1) {
            msg = sim->msgQueue.peekFirst(); // received something
        }
    } else {
        //    printf("WARNING: Lagging behind realtime!\n");
        // we're behind -- customized versions of this class may
        // alert if we're too much behind, whatever that means
    }
    // ok, return the message
    return msg;
}

void RealtimeScheduler::closeAppSocket(int fd)
{
    close(fd);
    FD_CLR(fd, &all_fds);

    appPacketBuffer->push_back(PacketBufferEntry(0, 0, PacketBufferEntry::FD_CLOSE, fd));
    sendNotificationMsg(appNotificationMsg, appModule);
}

void RealtimeScheduler::sendNotificationMsg(cMessage* msg, cModule* mod)
{
    if( msg->isScheduled() ) return; // Notification already scheduled
    timeval curTime;
    gettimeofday(&curTime, NULL);
    curTime = timeval_substract(curTime, baseTime);
    simtime_t t = curTime.tv_sec + curTime.tv_usec*1e-6;
    
    msg->setSentFrom(mod, -1, simulation.simTime());
    msg->setArrival(mod,-1,t);
    simulation.msgQueue.insert(msg);
}

ssize_t RealtimeScheduler::sendBytes(const char *buf,
                                     size_t numBytes,
                                     sockaddr* addr,
                                     socklen_t addrlen,
                                     bool isApp,
                                     int fd)
{
    if (!buf) {
        ev << "[RealtimeScheduler::sendBytes()]\n"
        << "    Error sending packet: buf = NULL"
        << endl;
        return -1;
    }
    if (!isApp) {
        if( numBytes > buffersize ) {
            ev << "[RealtimeScheduler::sendBytes()]\n"
            << "    Trying to send oversized packet: size " << numBytes << " mtu " << buffersize
            << endl;
            opp_error("Can't send packet: too large"); //FIXME: Throw exception instead
        }

        if ( netw_fd < 0 ) {
            ev << "[RealtimeScheduler::sendBytes()]\n"
            << "    Can't send packet to network: no tun socket"
            << endl;
            return 0;
        }
        int nBytes;
        if (addr) {
            nBytes =  sendto(netw_fd, buf, numBytes, 0, addr, addrlen);
        } else {
            nBytes =  write(netw_fd, buf, numBytes);
        }
        if (nBytes < 0) {
            ev << "[RealtimeScheduler::sendBytes()]\n"
            << "    Error sending data to network: " << strerror(errno) << "\n"
            << "    FD = " << netw_fd << ", numBytes = " << numBytes <<  ", addrlen = " << addrlen
            << endl;
        }
        return nBytes;

    } else {
        if( numBytes > appBuffersize ) {
            ev << "[RealtimeScheduler::sendBytes()]\n"
            << "    Trying to send oversized packet: size " << numBytes << "\n"
            << "    mtu " << appBuffersize
            << endl;
            opp_error("Can't send packet: too large"); //FIXME: Throw exception instead
        }
        // If no fd is given, select a "random" one
        if ( fd < 0 ) {
            for ( fd = 0; fd <= maxfd; fd++ ) {
                if( fd == netw_fd ) continue;
                if( fd == additional_fd ) continue;
                if( FD_ISSET(fd, &all_fds)) break;
            }
            if ( fd > maxfd ) throw new cException("Can't send packet to Application: no socket");
        }
        return write(fd, buf, numBytes);
    }
    // TBD check for errors
}

